/* @import this colorsheet to get the default values for every property.
 * This is useful when writing special CSS tests that should not be
 * inluenced by themes - not even the default ones.
 * Keep in mind that the output will be very ugly and not look like
 * anything GTK.
 * Also, when adding new style properties, please add them here.
 */

* {
  all: unset;
}
