b {
  margin-left: calc(1px + 2s);
}

c {
  margin-left: calc(1px - 2s);
}

d {
  margin-left: calc(2 * 3);
}

e {
  margin-left: calc(2px * 3px);
}

f {
  margin-left: calc(2 / 3px);
}

g {
  margin-left: calc(2 / 3);
}

h {
  margin-left: calc(2px / 3px);
}

i {
  margin-left: calc(error);
}

j {
  margin-left: calc(1px + error);
}

k {
  margin-left: calc(1px - error);
}

l {
  margin-left: calc(1 * error);
}

m {
  margin-left: calc(1 / error);
}

n {
  margin-left: calc(1px * error);
}

o {
  margin-left: calc(1px / error);
}

p {
  margin-left: calc((((((1px))));));
}

q {
  margin-left: calc((((((1px)))))))));
}

r {
  margin-left: calc(1px + (1));
}

s {
  margin-left: calc(1px + (error));
}

t {
  margin-left: calc((3px + 4px and a lot of junk));
}
