#ifndef __GTK_SHORTCUT_MANAGER_PRIVATE_H__
#define __GTK_SHORTCUT_MANAGER_PRIVATE_H__

#include "gtkshortcutmanager.h"

G_BEGIN_DECLS

void gtk_shortcut_manager_create_controllers (GtkWidget *widget);

G_END_DECLS

#endif /* __GTK_SHORTCUT_MANAGER_PRIVATE_H__ */
