@import "reset-to-defaults.css";

box {
  border-image: url("green-20x20.png") 5000;
  background-image: none;
  background-color: #000;
  border-style: solid;
  border-width: 1px;
  padding: 0;
  border-radius: 0;
}

grid {
  background-color: #000;
}

grid * {
  background-color: lime;
  border-style: none;
  border-width: 0;
}
