@import "reset-to-defaults.css";

window {
  background-image: none;
  border-radius: 0;
  background-color: purple;
}

button {
  background-color: blue;
}
.transparent {
  background-color: transparent;
}
.reference {
  background-color: purple;
}
