a {
  background-image: radial-gradient(at left circle, red, blue);
}
b {
  background-image: radial-gradient(10px at left circle, red, blue);
}
c {
  background-image: radial-gradient(at 10px 10px circle, red, blue);
}
