@import "reset-to-defaults.css";

#reference {
  background: linear-gradient(to bottom, red, yellow);
}

window {
  background: green;
}

#box1 {
  transform: scaleY(-1);
  background: linear-gradient(to bottom, yellow, red);
}
