/* test combining multiple state changes */

box {
  color: red;
}

label:hover {
  color: green;
}

label:backdrop {
  color: blue;
}

label:disabled {
  color: blue;
}

label:focus {
  color: blue;
}
