@import "reset-to-defaults.css";

window {
  background-color: purple;
}

#red {
  background-color: red;
  min-width: 100px;
  min-height: 100px;
}

#yellow {
  background-color: yellow;
  min-width: 100px;
  min-height: 100px;
}

#green {
  background-color: green;
  min-width: 100px;
  min-height: 100px;
}
