/*
You can type here any CSS rule recognized by GTK.
You can temporarily disable this custom CSS by clicking on the “Pause” button above.

Changes are applied instantly and globally, for the whole application.
*/


modelbutton.flat:hover {
  background-color: red;
}

popover arrow {
  background-color: lime;
  border-width: 2px;
  border-color: green;
}

popover {
}
\
popover contents {
  background-color: magenta;
}