@import "reset-to-defaults.css";

.top,
#label_n {
  background-color: blue;
}

.left,
#label_w {
  background-color: green;
}

.right,
#label_e {
  background-color: yellow;
}

.bottom,
#label_s {
  background-color: red;
}
