a {
  text-decoration-color: initial;
  text-decoration-line: underline;
  text-decoration-style: initial;
}

b {
  text-decoration-color: initial;
  text-decoration-line: initial;
  text-decoration-style: wavy;
}

c {
  text-decoration-color: rgb(255,0,0);
  text-decoration-line: initial;
  text-decoration-style: initial;
}

d {
  text-decoration-color: rgb(255,0,0);
  text-decoration-line: initial;
  text-decoration-style: wavy;
}

e {
  text-decoration-color: initial;
  text-decoration-line: none;
  text-decoration-style: initial;
}

f {
  text-decoration-color: rgb(255,0,0);
  text-decoration-line: underline overline;
  text-decoration-style: wavy;
}
