@import "reset-to-defaults.css";

window {
  background: yellow;
}

#box {
  transform-origin: top left;
  transform: rotate(-90deg);
  background: linear-gradient(to bottom, red 0%, blue 100%);
}

#reference {
  background: linear-gradient(to right, red 0%, blue 100%);
}
