
#ifndef __GTK_BUILDER_TOOL_H__
#define __GTK_BUILDER_TOOL_H__

void do_simplify  (int *argc, const char ***argv);
void do_validate  (int *argc, const char ***argv);
void do_enumerate (int *argc, const char ***argv);
void do_preview   (int *argc, const char ***argv);

#endif
