window {
  background-image: linear-gradient(75deg, red, green 20%, blue 40px);
}

box {
  background-image: repeating-linear-gradient(to top right, rgba(255,100,60,0.5), magenta 100%);
}

label {
  background-image: radial-gradient(farthest-corner at 50% 30px, yellow, red 30%, blue);
}
