a {
  animation-direction: initial;
}

b {
  animation-direction: inherit;
}

c {
  animation-direction: unset;
}

d {
  animation-direction: normal;
}

e {
  animation-direction: normal, reverse, alternate, alternate-reverse;
}
