a {
  background-image: initial;
}

b {
  background-image: inherit;
}

c {
  background-image: unset;
}

d {
  background-image: none;
}

e {
  background-image: none, none;
}

f {
  background-image: none /* FIXME */;
}

g {
  background-image: linear-gradient(45deg, rgb(255,0,0) 10%, rgb(0,128,0) 20px);
}

h {
  background-image: linear-gradient(to top, rgb(0,0,255), rgb(0,0,0));
}

i {
  background-image: linear-gradient(to bottom left, rgb(0,0,255), rgb(0,0,0));
}

j {
  background-image: linear-gradient(rgb(255,255,0), rgb(0,0,255));
}

k {
  background-image: repeating-linear-gradient(45deg, rgb(255,0,0) 10%, rgb(0,128,0) 20px);
}

l {
  background-image: repeating-linear-gradient(to top, rgb(0,0,255), rgb(0,0,0));
}

m {
  background-image: repeating-linear-gradient(to bottom left, rgb(0,0,255), rgb(0,0,0));
}

n {
  background-image: repeating-linear-gradient(rgb(255,255,0), rgb(0,0,255));
}

o {
  background-image: cross-fade(40% linear-gradient(rgb(0,0,255), rgb(0,0,0)), linear-gradient(rgb(0,0,0), rgb(255,0,0)));
}

p {
  background-image: cross-fade(linear-gradient(rgb(0,0,255), rgb(0,0,0)), linear-gradient(rgb(0,0,0), rgb(255,0,0)));
}

q {
  background-image: cross-fade(40% linear-gradient(rgb(0,0,255), rgb(0,0,0)));
}

r {
  background-image: -gtk-icontheme("object-select-symbolic");
}

s {
  background-image: -gtk-scaled(-gtk-icontheme("object-select-symbolic"),1,linear-gradient(rgb(255,255,0), rgb(0,0,255)),2);
}
