a {
  text-decoration: underline;
}

b {
  text-decoration: wavy;
}

c {
  text-decoration: red;
}

d {
  text-decoration: red wavy;
}

e {
  text-decoration: none;
}

f {
  text-decoration: red underline wavy overline;
}
