a {
  box-shadow: 0 0;
}

b {
  box-shadow: 0 0 inset;
}

c {
  box-shadow: 0 0 inset red;
}

d {
  box-shadow: 0 0 red;
}

e {
  box-shadow: 0 0 red inset;
}

f {
  box-shadow: inset 0 0;
}

g {
  box-shadow: inset red 0 0;
}

h {
  box-shadow: inset 0 0 red;
}

i {
  box-shadow: red 0 0;
}

j {
  box-shadow: red inset 0 0;
}

k {
  box-shadow: red 0 0 inset;
}
