* {
  -gtk-icon-shadow: initial;
  background-clip: initial;
  background-color: initial;
  background-image: initial;
  background-origin: initial;
  background-repeat: initial;
  border-bottom-color: initial;
  border-bottom-left-radius: initial;
  border-bottom-right-radius: initial;
  border-bottom-style: initial;
  border-bottom-width: initial;
  border-image-repeat: initial;
  border-image-slice: initial;
  border-image-source: initial;
  border-image-width: initial;
  border-left-color: initial;
  border-left-style: initial;
  border-left-width: initial;
  border-right-color: initial;
  border-right-style: initial;
  border-right-width: initial;
  border-top-color: initial;
  border-top-left-radius: initial;
  border-top-right-radius: initial;
  border-top-style: initial;
  border-top-width: initial;
  box-shadow: initial;
  color: initial;
  font-family: initial;
  font-size: initial;
  font-style: initial;
  font-variant-alternates: initial;
  font-variant-caps: initial;
  font-variant-east-asian: initial;
  font-variant-ligatures: initial;
  font-variant-numeric: initial;
  font-variant-position: initial;
  font-weight: initial;
  margin-bottom: initial;
  margin-left: initial;
  margin-right: initial;
  margin-top: initial;
  padding-bottom: initial;
  padding-left: initial;
  padding-right: initial;
  padding-top: initial;
  text-shadow: initial;
}
