aa {
  background-size: initial;
}

bb {
  background-size: inherit;
}

cc {
  background-size: unset;
}

a {
  background-size: cover;
}

b {
  background-size: contain;
}

c {
  background-size: auto;
}

d {
  background-size: auto;
}

e {
  background-size: auto 100%;
}

f {
  background-size: 100%;
}

g {
  background-size: 100%;
}

h {
  background-size: auto 10px;
}

i {
  background-size: 10px;
}

j {
  background-size: 10px;
}

k {
  background-size: 10px 100%;
}

l {
  background-size: 100% 10px;
}

m {
  background-size: 10px 10px;
}

n {
  background-size: auto, cover, 10px 100%;
}
