@import "reset-to-defaults.css";

.background {
  background-color: red;
  background-image: linear-gradient(to right, blue, blue);
  background-size: 0%;
}

#reference {
  background-image: initial;
  background-size: initial;
}
