/* GTK - The GIMP Tool Kit
 * Copyright (C) 2007, 2008 GNOME Foundation
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * File auto-generated from script found at http://bugzilla.gnome.org/show_bug.cgi?id=321896
 * using the input files
 *  Input   : http://cgit.freedesktop.org/xorg/lib/libX11/plain/nls/en_US.UTF-8/Compose.pre
 *  Input   : http://www.cl.cam.ac.uk/~mgk25/ucs/keysyms.txt
 *  Input   : http://www.unicode.org/Public/UNIDATA/UnicodeData.txt
 *
 * This table is optimised for space and requires special handling to access the content.
 * This table is used solely by http://svn.gnome.org/viewcvs/gtk%2B/trunk/gtk/gtkimcontextsimple.c
 * 
 * The resulting file is placed at http://svn.gnome.org/viewcvs/gtk%2B/trunk/gtk/gtkimcontextsimpleseqs.h
 * This file is described in bug report http://bugzilla.gnome.org/show_bug.cgi?id=321896
 */

/*
 * Modified by the GTK+ Team and others 2007, 2008.  See the AUTHORS
 * file for a list of people on the GTK+ Team.  See the ChangeLog
 * files for a list of changes.  These files are distributed with
 * GTK+ at ftp://ftp.gtk.org/pub/gtk/.
 */

#ifndef __GTK_IM_CONTEXT_SIMPLE_SEQS_H__
#define __GTK_IM_CONTEXT_SIMPLE_SEQS_H__

/* === These are the original comments of the file; we keep for historical purposes ===
 *
 * The following table was generated from the X compose tables include with
 * XFree86 4.0 using a set of Perl scripts. Contact Owen Taylor <otaylor@redhat.com>
 * to obtain the relevant perl scripts.
 *
 * The following compose letter letter sequences conflicted
 *   Dstroke/dstroke and ETH/eth; resolved to Dstroke (Croatian, Vietnamese, Lappish), over
 *                                ETH (Icelandic, Faroese, old English, IPA)  [ D- -D d- -d ]
 *   Amacron/amacron and ordfeminine; resolved to ordfeminine                 [ _A A_ a_ _a ]
 *   Amacron/amacron and Atilde/atilde; resolved to atilde                    [ -A A- a- -a ]
 *   Omacron/Omacron and masculine; resolved to masculine                     [ _O O_ o_ _o ]
 *   Omacron/omacron and Otilde/atilde; resolved to otilde                    [ -O O- o- -o ]
 *
 * [ Amacron and Omacron are in Latin-4 (Baltic). ordfeminine and masculine are used for
 *   spanish. atilde and otilde are used at least for Portuguese ]
 *
 *   at and Aring; resolved to Aring                                          [ AA ]
 *   guillemotleft and caron; resolved to guillemotleft                       [ << ]
 *   ogonek and cedilla; resolved to cedilla                                  [ ,, ]
 *
 * This probably should be resolved by first checking an additional set of compose tables
 * that depend on the locale or selected input method.
 */

static const guint16 gtk_compose_seqs_compact[] = {
GDK_KEY_Greek_accentdieresis, 180, 184, 184, 184, 184, 
GDK_KEY_dead_grave, 184, 246, 333, 545, 545, 
GDK_KEY_dead_acute, 545, 609, 705, 981, 981, 
GDK_KEY_dead_circumflex, 981, 1105, 1105, 1305, 1305, 
GDK_KEY_dead_tilde, 1305, 1389, 1452, 1592, 1592, 
GDK_KEY_dead_macron, 1592, 1638, 1656, 1728, 1728, 
GDK_KEY_dead_breve, 1728, 1778, 1778, 1802, 1802, 
GDK_KEY_dead_abovedot, 1802, 1832, 1835, 1871, 1871, 
GDK_KEY_dead_diaeresis, 1871, 1959, 1971, 1995, 1995, 
GDK_KEY_dead_abovering, 1995, 2005, 2005, 2005, 2005, 
GDK_KEY_dead_doubleacute, 2005, 2015, 2015, 2015, 2015, 
GDK_KEY_dead_caron, 2015, 2057, 2057, 2065, 2065, 
GDK_KEY_dead_cedilla, 2065, 2077, 2083, 2083, 2083, 
GDK_KEY_dead_ogonek, 2083, 2093, 2093, 2093, 2093, 
GDK_KEY_dead_iota, 2093, 2115, 2214, 2646, 3306, 
GDK_KEY_dead_voiced_sound, 3306, 3352, 3352, 3352, 3352, 
GDK_KEY_dead_semivoiced_sound, 3352, 3362, 3362, 3362, 3362, 
GDK_KEY_dead_belowdot, 3362, 3378, 3378, 3394, 3394, 
GDK_KEY_dead_hook, 3394, 3472, 3475, 3531, 3531, 
GDK_KEY_dead_horn, 3531, 3541, 3541, 3541, 3541, 
GDK_KEY_dead_stroke, 3541, 3629, 3641, 3641, 3641, 
GDK_KEY_dead_psili, 3641, 3669, 3669, 3669, 3669, 
GDK_KEY_dead_dasia, 3669, 3701, 3701, 3701, 3701, 
GDK_KEY_dead_belowring, 3701, 3703, 3703, 3703, 3703, 
GDK_KEY_dead_belowtilde, 3703, 3705, 3705, 3705, 3705, 
GDK_KEY_dead_belowdiaeresis, 3705, 3705, 3708, 3708, 3708, 
GDK_KEY_dead_belowcomma, 3708, 3722, 3722, 3722, 3722, 
GDK_KEY_dead_currency, 3722, 3820, 3826, 3826, 3826, 
GDK_KEY_dead_greek, 3826, 3928, 3952, 3952, 3952, 
GDK_KEY_Multi_key, 3952, 3952, 10348, 14044, 15919, 
GDK_KEY_Greek_iota, 0x0390, 
GDK_KEY_Greek_upsilon, 0x03B0, 
GDK_KEY_space, 0x0060, 
GDK_KEY_V, 0x01DB, 
GDK_KEY_v, 0x01DC, 
GDK_KEY_nobreakspace, 0x0300, 
GDK_KEY_Abreve, 0x1EB0, 
GDK_KEY_abreve, 0x1EB1, 
GDK_KEY_Emacron, 0x1E14, 
GDK_KEY_emacron, 0x1E15, 
GDK_KEY_Omacron, 0x1E50, 
GDK_KEY_omacron, 0x1E51, 
GDK_KEY_Cyrillic_ie, 0x0450, 
GDK_KEY_Cyrillic_i, 0x045D, 
GDK_KEY_Cyrillic_IE, 0x0400, 
GDK_KEY_Cyrillic_I, 0x040D, 
GDK_KEY_Greek_iotadieresis, 0x1FD2, 
GDK_KEY_Greek_upsilondieresis, 0x1FE2, 
GDK_KEY_Greek_ALPHA, 0x1FBA, 
GDK_KEY_Greek_EPSILON, 0x1FC8, 
GDK_KEY_Greek_ETA, 0x1FCA, 
GDK_KEY_Greek_IOTA, 0x1FDA, 
GDK_KEY_Greek_OMICRON, 0x1FF8, 
GDK_KEY_Greek_UPSILON, 0x1FEA, 
GDK_KEY_Greek_OMEGA, 0x1FFA, 
GDK_KEY_Greek_alpha, 0x1F70, 
GDK_KEY_Greek_epsilon, 0x1F72, 
GDK_KEY_Greek_eta, 0x1F74, 
GDK_KEY_Greek_iota, 0x1F76, 
GDK_KEY_Greek_omicron, 0x1F78, 
GDK_KEY_Greek_upsilon, 0x1F7A, 
GDK_KEY_Greek_omega, 0x1F7C, 
GDK_KEY_dead_grave, 0x0060, 
GDK_KEY_dead_diaeresis, GDK_KEY_Greek_iota, 0x1FD2, 
GDK_KEY_dead_diaeresis, GDK_KEY_Greek_upsilon, 0x1FE2, 
GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F0A, 
GDK_KEY_dead_psili, GDK_KEY_Greek_EPSILON, 0x1F1A, 
GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F2A, 
GDK_KEY_dead_psili, GDK_KEY_Greek_IOTA, 0x1F3A, 
GDK_KEY_dead_psili, GDK_KEY_Greek_OMICRON, 0x1F4A, 
GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1F6A, 
GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F02, 
GDK_KEY_dead_psili, GDK_KEY_Greek_epsilon, 0x1F12, 
GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F22, 
GDK_KEY_dead_psili, GDK_KEY_Greek_iota, 0x1F32, 
GDK_KEY_dead_psili, GDK_KEY_Greek_omicron, 0x1F42, 
GDK_KEY_dead_psili, GDK_KEY_Greek_upsilon, 0x1F52, 
GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1F62, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F0B, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_EPSILON, 0x1F1B, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F2B, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_IOTA, 0x1F3B, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_OMICRON, 0x1F4B, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_UPSILON, 0x1F5B, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1F6B, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F03, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_epsilon, 0x1F13, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F23, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_iota, 0x1F33, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_omicron, 0x1F43, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_upsilon, 0x1F53, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1F63, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_U, 0x01DB, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_u, 0x01DC, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_Greek_iota, 0x1FD2, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_Greek_upsilon, 0x1FE2, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F0B, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_EPSILON, 0x1F1B, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F2B, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_IOTA, 0x1F3B, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_OMICRON, 0x1F4B, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_UPSILON, 0x1F5B, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1F6B, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F03, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_epsilon, 0x1F13, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F23, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_iota, 0x1F33, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_omicron, 0x1F43, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_upsilon, 0x1F53, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1F63, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F0A, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_EPSILON, 0x1F1A, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F2A, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_IOTA, 0x1F3A, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_OMICRON, 0x1F4A, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1F6A, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F02, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_epsilon, 0x1F12, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F22, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_iota, 0x1F32, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_omicron, 0x1F42, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_upsilon, 0x1F52, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1F62, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_O, 0x1EDC, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_U, 0x1EEA, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_o, 0x1EDD, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_u, 0x1EEB, 
GDK_KEY_Multi_key, GDK_KEY_U, GDK_KEY_A, 0x1EB0, 
GDK_KEY_Multi_key, GDK_KEY_U, GDK_KEY_a, 0x1EB1, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_A, 0x1EA6, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_E, 0x1EC0, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_O, 0x1ED2, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_a, 0x1EA7, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_e, 0x1EC1, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_o, 0x1ED3, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_E, 0x1E14, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_O, 0x1E50, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_e, 0x1E15, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_o, 0x1E51, 
GDK_KEY_Multi_key, GDK_KEY_b, GDK_KEY_A, 0x1EB0, 
GDK_KEY_Multi_key, GDK_KEY_b, GDK_KEY_a, 0x1EB1, 
GDK_KEY_Multi_key, GDK_KEY_macron, GDK_KEY_E, 0x1E14, 
GDK_KEY_Multi_key, GDK_KEY_macron, GDK_KEY_O, 0x1E50, 
GDK_KEY_Multi_key, GDK_KEY_macron, GDK_KEY_e, 0x1E15, 
GDK_KEY_Multi_key, GDK_KEY_macron, GDK_KEY_o, 0x1E51, 
GDK_KEY_space, 0x0027, 
GDK_KEY_V, 0x01D7, 
GDK_KEY_v, 0x01D8, 
GDK_KEY_nobreakspace, 0x0301, 
GDK_KEY_Abreve, 0x1EAE, 
GDK_KEY_abreve, 0x1EAF, 
GDK_KEY_Emacron, 0x1E16, 
GDK_KEY_emacron, 0x1E17, 
GDK_KEY_Utilde, 0x1E78, 
GDK_KEY_omacron, 0x1E53, 
GDK_KEY_utilde, 0x1E79, 
GDK_KEY_Cyrillic_ghe, 0x0453, 
GDK_KEY_Cyrillic_ka, 0x045C, 
GDK_KEY_Cyrillic_GHE, 0x0403, 
GDK_KEY_Cyrillic_KA, 0x040C, 
GDK_KEY_Greek_iotadieresis, 0x0390, 
GDK_KEY_Greek_upsilondieresis, 0x03B0, 
GDK_KEY_Greek_ALPHA, 0x0386, 
GDK_KEY_Greek_EPSILON, 0x0388, 
GDK_KEY_Greek_ETA, 0x0389, 
GDK_KEY_Greek_IOTA, 0x038A, 
GDK_KEY_Greek_OMICRON, 0x038C, 
GDK_KEY_Greek_UPSILON, 0x038E, 
GDK_KEY_Greek_OMEGA, 0x038F, 
GDK_KEY_Greek_alpha, 0x03AC, 
GDK_KEY_Greek_epsilon, 0x03AD, 
GDK_KEY_Greek_eta, 0x03AE, 
GDK_KEY_Greek_iota, 0x03AF, 
GDK_KEY_Greek_omicron, 0x03CC, 
GDK_KEY_Greek_upsilon, 0x03CD, 
GDK_KEY_Greek_omega, 0x03CE, 
GDK_KEY_dead_acute, 0x00B4, 
GDK_KEY_dead_diaeresis, GDK_KEY_space, 0x0385, 
GDK_KEY_dead_diaeresis, GDK_KEY_Greek_iota, 0x0390, 
GDK_KEY_dead_diaeresis, GDK_KEY_Greek_upsilon, 0x03B0, 
GDK_KEY_dead_stroke, GDK_KEY_O, 0x01FE, 
GDK_KEY_dead_stroke, GDK_KEY_o, 0x01FF, 
GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F0C, 
GDK_KEY_dead_psili, GDK_KEY_Greek_EPSILON, 0x1F1C, 
GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F2C, 
GDK_KEY_dead_psili, GDK_KEY_Greek_IOTA, 0x1F3C, 
GDK_KEY_dead_psili, GDK_KEY_Greek_OMICRON, 0x1F4C, 
GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1F6C, 
GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F04, 
GDK_KEY_dead_psili, GDK_KEY_Greek_epsilon, 0x1F14, 
GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F24, 
GDK_KEY_dead_psili, GDK_KEY_Greek_iota, 0x1F34, 
GDK_KEY_dead_psili, GDK_KEY_Greek_omicron, 0x1F44, 
GDK_KEY_dead_psili, GDK_KEY_Greek_upsilon, 0x1F54, 
GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1F64, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F0D, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_EPSILON, 0x1F1D, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F2D, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_IOTA, 0x1F3D, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_OMICRON, 0x1F4D, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_UPSILON, 0x1F5D, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1F6D, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F05, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_epsilon, 0x1F15, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F25, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_iota, 0x1F35, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_omicron, 0x1F45, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_upsilon, 0x1F55, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1F65, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_I, 0x1E2E, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_U, 0x01D7, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_i, 0x1E2F, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_u, 0x01D8, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_Greek_iota, 0x0390, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_Greek_upsilon, 0x03B0, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F0D, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_EPSILON, 0x1F1D, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F2D, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_IOTA, 0x1F3D, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_OMICRON, 0x1F4D, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_UPSILON, 0x1F5D, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1F6D, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F05, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_epsilon, 0x1F15, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F25, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_iota, 0x1F35, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_omicron, 0x1F45, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_upsilon, 0x1F55, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1F65, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F0C, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_EPSILON, 0x1F1C, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F2C, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_IOTA, 0x1F3C, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_OMICRON, 0x1F4C, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1F6C, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F04, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_epsilon, 0x1F14, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F24, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_iota, 0x1F34, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_omicron, 0x1F44, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_upsilon, 0x1F54, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1F64, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_O, 0x1EDA, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_U, 0x1EE8, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_o, 0x1EDB, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_u, 0x1EE9, 
GDK_KEY_Multi_key, GDK_KEY_comma, GDK_KEY_C, 0x1E08, 
GDK_KEY_Multi_key, GDK_KEY_comma, GDK_KEY_c, 0x1E09, 
GDK_KEY_Multi_key, GDK_KEY_slash, GDK_KEY_O, 0x01FE, 
GDK_KEY_Multi_key, GDK_KEY_slash, GDK_KEY_o, 0x01FF, 
GDK_KEY_Multi_key, GDK_KEY_U, GDK_KEY_A, 0x1EAE, 
GDK_KEY_Multi_key, GDK_KEY_U, GDK_KEY_a, 0x1EAF, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_A, 0x1EA4, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_E, 0x1EBE, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_O, 0x1ED0, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_a, 0x1EA5, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_e, 0x1EBF, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_o, 0x1ED1, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_E, 0x1E16, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_O, 0x1E52, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_e, 0x1E17, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_o, 0x1E53, 
GDK_KEY_Multi_key, GDK_KEY_b, GDK_KEY_A, 0x1EAE, 
GDK_KEY_Multi_key, GDK_KEY_b, GDK_KEY_a, 0x1EAF, 
GDK_KEY_Multi_key, GDK_KEY_o, GDK_KEY_A, 0x01FA, 
GDK_KEY_Multi_key, GDK_KEY_o, GDK_KEY_a, 0x01FB, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_O, 0x1E4C, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_U, 0x1E78, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_o, 0x1E4D, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_u, 0x1E79, 
GDK_KEY_Multi_key, GDK_KEY_macron, GDK_KEY_E, 0x1E16, 
GDK_KEY_Multi_key, GDK_KEY_macron, GDK_KEY_O, 0x1E52, 
GDK_KEY_Multi_key, GDK_KEY_macron, GDK_KEY_e, 0x1E17, 
GDK_KEY_Multi_key, GDK_KEY_macron, GDK_KEY_o, 0x1E53, 
GDK_KEY_Multi_key, GDK_KEY_cedilla, GDK_KEY_C, 0x1E08, 
GDK_KEY_Multi_key, GDK_KEY_cedilla, GDK_KEY_c, 0x1E09, 
GDK_KEY_Multi_key, GDK_KEY_KP_Divide, GDK_KEY_O, 0x01FE, 
GDK_KEY_Multi_key, GDK_KEY_KP_Divide, GDK_KEY_o, 0x01FF, 
GDK_KEY_space, 0x005E, 
GDK_KEY_parenleft, 0x207D, 
GDK_KEY_parenright, 0x207E, 
GDK_KEY_plus, 0x207A, 
GDK_KEY_minus, 0x207B, 
GDK_KEY_0, 0x2070, 
GDK_KEY_1, 0x00B9, 
GDK_KEY_2, 0x00B2, 
GDK_KEY_3, 0x00B3, 
GDK_KEY_4, 0x2074, 
GDK_KEY_5, 0x2075, 
GDK_KEY_6, 0x2076, 
GDK_KEY_7, 0x2077, 
GDK_KEY_8, 0x2078, 
GDK_KEY_9, 0x2079, 
GDK_KEY_equal, 0x207C, 
GDK_KEY_nobreakspace, 0x0302, 
GDK_KEY_Agrave, 0x1EA6, 
GDK_KEY_Aacute, 0x1EA4, 
GDK_KEY_Atilde, 0x1EAA, 
GDK_KEY_Egrave, 0x1EC0, 
GDK_KEY_Eacute, 0x1EBE, 
GDK_KEY_Ograve, 0x1ED2, 
GDK_KEY_Oacute, 0x1ED0, 
GDK_KEY_Otilde, 0x1ED6, 
GDK_KEY_agrave, 0x1EA7, 
GDK_KEY_aacute, 0x1EA5, 
GDK_KEY_atilde, 0x1EAB, 
GDK_KEY_egrave, 0x1EC1, 
GDK_KEY_eacute, 0x1EBF, 
GDK_KEY_ograve, 0x1ED3, 
GDK_KEY_oacute, 0x1ED1, 
GDK_KEY_otilde, 0x1ED7, 
0x2212, 0x207B, 
0x4E00, 0x3192, 
0x4E01, 0x319C, 
0x4E09, 0x3194, 
0x4E0A, 0x3196, 
0x4E0B, 0x3198, 
0x4E19, 0x319B, 
0x4E2D, 0x3197, 
0x4E59, 0x319A, 
0x4E8C, 0x3193, 
0x4EBA, 0x319F, 
0x56DB, 0x3195, 
0x5730, 0x319E, 
0x5929, 0x319D, 
0x7532, 0x3199, 
GDK_KEY_dead_circumflex, 0x005E, 
GDK_KEY_KP_Space, 0x00B2, 
GDK_KEY_KP_Add, 0x207A, 
GDK_KEY_KP_0, 0x2070, 
GDK_KEY_KP_1, 0x00B9, 
GDK_KEY_KP_2, 0x00B2, 
GDK_KEY_KP_3, 0x00B3, 
GDK_KEY_KP_4, 0x2074, 
GDK_KEY_KP_5, 0x2075, 
GDK_KEY_KP_6, 0x2076, 
GDK_KEY_KP_7, 0x2077, 
GDK_KEY_KP_8, 0x2078, 
GDK_KEY_KP_9, 0x2079, 
GDK_KEY_KP_Equal, 0x207C, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_A, 0x1EAC, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_E, 0x1EC6, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_O, 0x1ED8, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_a, 0x1EAD, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_e, 0x1EC7, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_o, 0x1ED9, 
GDK_KEY_Multi_key, GDK_KEY_S, GDK_KEY_M, 0x2120, 
GDK_KEY_Multi_key, GDK_KEY_S, GDK_KEY_m, 0x2120, 
GDK_KEY_Multi_key, GDK_KEY_T, GDK_KEY_M, 0x2122, 
GDK_KEY_Multi_key, GDK_KEY_T, GDK_KEY_m, 0x2122, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_a, 0x00AA, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_h, 0x02B0, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_i, 0x2071, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_j, 0x02B2, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_l, 0x02E1, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_n, 0x207F, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_o, 0x00BA, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_r, 0x02B3, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_s, 0x02E2, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_w, 0x02B7, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_x, 0x02E3, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_y, 0x02B8, 
GDK_KEY_Multi_key, GDK_KEY_underscore, 0x0263, 0x02E0, 
GDK_KEY_Multi_key, GDK_KEY_underscore, 0x0266, 0x02B1, 
GDK_KEY_Multi_key, GDK_KEY_underscore, 0x0279, 0x02B4, 
GDK_KEY_Multi_key, GDK_KEY_underscore, 0x027B, 0x02B5, 
GDK_KEY_Multi_key, GDK_KEY_underscore, 0x0281, 0x02B6, 
GDK_KEY_Multi_key, GDK_KEY_underscore, 0x0295, 0x02E4, 
GDK_KEY_Multi_key, GDK_KEY_s, GDK_KEY_M, 0x2120, 
GDK_KEY_Multi_key, GDK_KEY_s, GDK_KEY_m, 0x2120, 
GDK_KEY_Multi_key, GDK_KEY_t, GDK_KEY_M, 0x2122, 
GDK_KEY_Multi_key, GDK_KEY_t, GDK_KEY_m, 0x2122, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_a, 0x00AA, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_h, 0x02B0, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_i, 0x2071, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_j, 0x02B2, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_l, 0x02E1, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_n, 0x207F, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_o, 0x00BA, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_r, 0x02B3, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_s, 0x02E2, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_w, 0x02B7, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_x, 0x02E3, 
GDK_KEY_Multi_key, GDK_KEY_underbar, GDK_KEY_y, 0x02B8, 
GDK_KEY_Multi_key, GDK_KEY_underbar, 0x0263, 0x02E0, 
GDK_KEY_Multi_key, GDK_KEY_underbar, 0x0266, 0x02B1, 
GDK_KEY_Multi_key, GDK_KEY_underbar, 0x0279, 0x02B4, 
GDK_KEY_Multi_key, GDK_KEY_underbar, 0x027B, 0x02B5, 
GDK_KEY_Multi_key, GDK_KEY_underbar, 0x0281, 0x02B6, 
GDK_KEY_Multi_key, GDK_KEY_underbar, 0x0295, 0x02E4, 
GDK_KEY_space, 0x007E, 
GDK_KEY_less, 0x2272, 
GDK_KEY_equal, 0x2243, 
GDK_KEY_greater, 0x2273, 
GDK_KEY_nobreakspace, 0x0303, 
GDK_KEY_Oacute, 0x1E4C, 
GDK_KEY_Odiaeresis, 0x1E4E, 
GDK_KEY_Uacute, 0x1E78, 
GDK_KEY_oacute, 0x1E4D, 
GDK_KEY_odiaeresis, 0x1E4F, 
GDK_KEY_uacute, 0x1E79, 
GDK_KEY_Abreve, 0x1EB4, 
GDK_KEY_abreve, 0x1EB5, 
GDK_KEY_Omacron, 0x022C, 
GDK_KEY_omacron, 0x022D, 
GDK_KEY_Greek_iotadieresis, 0x1FD7, 
GDK_KEY_Greek_upsilondieresis, 0x1FE7, 
GDK_KEY_Greek_alpha, 0x1FB6, 
GDK_KEY_Greek_eta, 0x1FC6, 
GDK_KEY_Greek_iota, 0x1FD6, 
GDK_KEY_Greek_upsilon, 0x1FE6, 
GDK_KEY_Greek_omega, 0x1FF6, 
0x1F00, 0x1F06, 
0x1F01, 0x1F07, 
0x1F08, 0x1F0E, 
0x1F09, 0x1F0F, 
0x1F20, 0x1F26, 
0x1F21, 0x1F27, 
0x1F28, 0x1F2E, 
0x1F29, 0x1F2F, 
0x1F30, 0x1F36, 
0x1F31, 0x1F37, 
0x1F38, 0x1F3E, 
0x1F39, 0x1F3F, 
0x1F50, 0x1F56, 
0x1F51, 0x1F57, 
0x1F59, 0x1F5F, 
0x1F60, 0x1F66, 
0x1F61, 0x1F67, 
0x1F68, 0x1F6E, 
0x1F69, 0x1F6F, 
GDK_KEY_dead_tilde, 0x007E, 
GDK_KEY_dead_diaeresis, GDK_KEY_Greek_iota, 0x1FD7, 
GDK_KEY_dead_diaeresis, GDK_KEY_Greek_upsilon, 0x1FE7, 
GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F0E, 
GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F2E, 
GDK_KEY_dead_psili, GDK_KEY_Greek_IOTA, 0x1F3E, 
GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1F6E, 
GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F06, 
GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F26, 
GDK_KEY_dead_psili, GDK_KEY_Greek_iota, 0x1F36, 
GDK_KEY_dead_psili, GDK_KEY_Greek_upsilon, 0x1F56, 
GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1F66, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F0F, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F2F, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_IOTA, 0x1F3F, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_UPSILON, 0x1F5F, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1F6F, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F07, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F27, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_iota, 0x1F37, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_upsilon, 0x1F57, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1F67, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_Greek_iota, 0x1FD7, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_Greek_upsilon, 0x1FE7, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F0F, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F2F, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_IOTA, 0x1F3F, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_UPSILON, 0x1F5F, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1F6F, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F07, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F27, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_iota, 0x1F37, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_upsilon, 0x1F57, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1F67, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F0E, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F2E, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_IOTA, 0x1F3E, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1F6E, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F06, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F26, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_iota, 0x1F36, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_upsilon, 0x1F56, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1F66, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_O, 0x1EE0, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_U, 0x1EEE, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_o, 0x1EE1, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_u, 0x1EEF, 
GDK_KEY_Multi_key, GDK_KEY_U, GDK_KEY_A, 0x1EB4, 
GDK_KEY_Multi_key, GDK_KEY_U, GDK_KEY_a, 0x1EB5, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_A, 0x1EAA, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_E, 0x1EC4, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_O, 0x1ED6, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_a, 0x1EAB, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_e, 0x1EC5, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_o, 0x1ED7, 
GDK_KEY_Multi_key, GDK_KEY_b, GDK_KEY_A, 0x1EB4, 
GDK_KEY_Multi_key, GDK_KEY_b, GDK_KEY_a, 0x1EB5, 
GDK_KEY_space, 0x00AF, 
GDK_KEY_V, 0x01D5, 
GDK_KEY_v, 0x01D6, 
GDK_KEY_nobreakspace, 0x0304, 
GDK_KEY_Egrave, 0x1E14, 
GDK_KEY_Eacute, 0x1E16, 
GDK_KEY_Ograve, 0x1E50, 
GDK_KEY_Oacute, 0x1E52, 
GDK_KEY_egrave, 0x1E15, 
GDK_KEY_eacute, 0x1E17, 
GDK_KEY_ograve, 0x1E51, 
GDK_KEY_oacute, 0x1E53, 
GDK_KEY_Cyrillic_i, 0x04E3, 
GDK_KEY_Cyrillic_u, 0x04EF, 
GDK_KEY_Cyrillic_I, 0x04E2, 
GDK_KEY_Cyrillic_U, 0x04EE, 
GDK_KEY_Greek_ALPHA, 0x1FB9, 
GDK_KEY_Greek_IOTA, 0x1FD9, 
GDK_KEY_Greek_UPSILON, 0x1FE9, 
GDK_KEY_Greek_alpha, 0x1FB1, 
GDK_KEY_Greek_iota, 0x1FD1, 
GDK_KEY_Greek_upsilon, 0x1FE1, 
GDK_KEY_dead_macron, 0x00AF, 
GDK_KEY_dead_greek, GDK_KEY_A, 0x1FB9, 
GDK_KEY_dead_greek, GDK_KEY_I, 0x1FD9, 
GDK_KEY_dead_greek, GDK_KEY_U, 0x1FE9, 
GDK_KEY_dead_greek, GDK_KEY_a, 0x1FB1, 
GDK_KEY_dead_greek, GDK_KEY_i, 0x1FD1, 
GDK_KEY_dead_greek, GDK_KEY_u, 0x1FE1, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_L, 0x1E38, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_R, 0x1E5C, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_l, 0x1E39, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_r, 0x1E5D, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_A, 0x01DE, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_O, 0x022A, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_U, 0x01D5, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_a, 0x01DF, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_o, 0x022B, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_u, 0x01D6, 
GDK_KEY_Multi_key, GDK_KEY_period, GDK_KEY_A, 0x01E0, 
GDK_KEY_Multi_key, GDK_KEY_period, GDK_KEY_O, 0x0230, 
GDK_KEY_Multi_key, GDK_KEY_period, GDK_KEY_a, 0x01E1, 
GDK_KEY_Multi_key, GDK_KEY_period, GDK_KEY_o, 0x0231, 
GDK_KEY_Multi_key, GDK_KEY_semicolon, GDK_KEY_O, 0x01EC, 
GDK_KEY_Multi_key, GDK_KEY_semicolon, GDK_KEY_o, 0x01ED, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_O, 0x022C, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_o, 0x022D, 
GDK_KEY_space, 0x02D8, 
GDK_KEY_nobreakspace, 0x0306, 
GDK_KEY_Agrave, 0x1EB0, 
GDK_KEY_Aacute, 0x1EAE, 
GDK_KEY_Atilde, 0x1EB4, 
GDK_KEY_agrave, 0x1EB1, 
GDK_KEY_aacute, 0x1EAF, 
GDK_KEY_atilde, 0x1EB5, 
GDK_KEY_Cyrillic_a, 0x04D1, 
GDK_KEY_Cyrillic_ie, 0x04D7, 
GDK_KEY_Cyrillic_i, 0x0439, 
GDK_KEY_Cyrillic_u, 0x045E, 
GDK_KEY_Cyrillic_zhe, 0x04C2, 
GDK_KEY_Cyrillic_A, 0x04D0, 
GDK_KEY_Cyrillic_IE, 0x04D6, 
GDK_KEY_Cyrillic_I, 0x0419, 
GDK_KEY_Cyrillic_U, 0x040E, 
GDK_KEY_Cyrillic_ZHE, 0x04C1, 
GDK_KEY_Greek_ALPHA, 0x1FB8, 
GDK_KEY_Greek_IOTA, 0x1FD8, 
GDK_KEY_Greek_UPSILON, 0x1FE8, 
GDK_KEY_Greek_alpha, 0x1FB0, 
GDK_KEY_Greek_iota, 0x1FD0, 
GDK_KEY_Greek_upsilon, 0x1FE0, 
GDK_KEY_dead_breve, 0x02D8, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_A, 0x1EB6, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_a, 0x1EB7, 
GDK_KEY_Multi_key, GDK_KEY_comma, GDK_KEY_E, 0x1E1C, 
GDK_KEY_Multi_key, GDK_KEY_comma, GDK_KEY_e, 0x1E1D, 
GDK_KEY_Multi_key, GDK_KEY_cedilla, GDK_KEY_E, 0x1E1C, 
GDK_KEY_Multi_key, GDK_KEY_cedilla, GDK_KEY_e, 0x1E1D, 
GDK_KEY_space, 0x02D9, 
GDK_KEY_L, 0x013F, 
GDK_KEY_i, 0x0131, 
GDK_KEY_j, 0x0237, 
GDK_KEY_l, 0x0140, 
GDK_KEY_nobreakspace, 0x0307, 
GDK_KEY_Sacute, 0x1E64, 
GDK_KEY_Scaron, 0x1E66, 
GDK_KEY_sacute, 0x1E65, 
GDK_KEY_scaron, 0x1E67, 
GDK_KEY_Amacron, 0x01E0, 
GDK_KEY_Omacron, 0x0230, 
GDK_KEY_amacron, 0x01E1, 
GDK_KEY_omacron, 0x0231, 
GDK_KEY_dead_abovedot, 0x02D9, 
GDK_KEY_dead_stroke, GDK_KEY_j, 0x025F, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_S, 0x1E68, 
GDK_KEY_Multi_key, GDK_KEY_exclam, GDK_KEY_s, 0x1E69, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_S, 0x1E64, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_s, 0x1E65, 
GDK_KEY_Multi_key, GDK_KEY_c, GDK_KEY_S, 0x1E66, 
GDK_KEY_Multi_key, GDK_KEY_c, GDK_KEY_s, 0x1E67, 
GDK_KEY_Multi_key, GDK_KEY_f, GDK_KEY_s, 0x1E9B, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_S, 0x1E64, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_s, 0x1E65, 
GDK_KEY_space, 0x0022, 
GDK_KEY_apostrophe, 0x0344, 
GDK_KEY_nobreakspace, 0x0308, 
GDK_KEY_acute, 0x0344, 
GDK_KEY_Iacute, 0x1E2E, 
GDK_KEY_Ugrave, 0x01DB, 
GDK_KEY_Uacute, 0x01D7, 
GDK_KEY_iacute, 0x1E2F, 
GDK_KEY_ugrave, 0x01DC, 
GDK_KEY_uacute, 0x01D8, 
0x01D3, 0x01D9, 
0x01D4, 0x01DA, 
GDK_KEY_Amacron, 0x01DE, 
GDK_KEY_Umacron, 0x1E7A, 
GDK_KEY_amacron, 0x01DF, 
GDK_KEY_omacron, 0x022B, 
GDK_KEY_umacron, 0x1E7B, 
GDK_KEY_Ukrainian_i, 0x0457, 
GDK_KEY_Ukrainian_I, 0x0407, 
GDK_KEY_Cyrillic_a, 0x04D3, 
GDK_KEY_Cyrillic_ie, 0x0451, 
GDK_KEY_Cyrillic_i, 0x04E5, 
GDK_KEY_Cyrillic_o, 0x04E7, 
GDK_KEY_Cyrillic_u, 0x04F1, 
GDK_KEY_Cyrillic_zhe, 0x04DD, 
GDK_KEY_Cyrillic_yeru, 0x04F9, 
GDK_KEY_Cyrillic_ze, 0x04DF, 
GDK_KEY_Cyrillic_e, 0x04ED, 
GDK_KEY_Cyrillic_che, 0x04F5, 
GDK_KEY_Cyrillic_A, 0x04D2, 
GDK_KEY_Cyrillic_IE, 0x0401, 
GDK_KEY_Cyrillic_I, 0x04E4, 
GDK_KEY_Cyrillic_O, 0x04E6, 
GDK_KEY_Cyrillic_U, 0x04F0, 
GDK_KEY_Cyrillic_ZHE, 0x04DC, 
GDK_KEY_Cyrillic_YERU, 0x04F8, 
GDK_KEY_Cyrillic_ZE, 0x04DE, 
GDK_KEY_Cyrillic_E, 0x04EC, 
GDK_KEY_Cyrillic_CHE, 0x04F4, 
GDK_KEY_Greek_IOTA, 0x03AA, 
GDK_KEY_Greek_UPSILON, 0x03AB, 
GDK_KEY_Greek_iota, 0x03CA, 
GDK_KEY_Greek_upsilon, 0x03CB, 
GDK_KEY_dead_diaeresis, 0x00A8, 
GDK_KEY_dead_acute, GDK_KEY_space, 0x0385, 
GDK_KEY_dead_acute, GDK_KEY_Greek_iota, 0x0390, 
GDK_KEY_dead_acute, GDK_KEY_Greek_upsilon, 0x03B0, 
GDK_KEY_dead_belowdiaeresis, GDK_KEY_equal, 0x2A77, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_U, 0x1E7A, 
GDK_KEY_Multi_key, GDK_KEY_underscore, GDK_KEY_u, 0x1E7B, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_O, 0x1E4E, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_o, 0x1E4F, 
GDK_KEY_Multi_key, GDK_KEY_macron, GDK_KEY_U, 0x1E7A, 
GDK_KEY_Multi_key, GDK_KEY_macron, GDK_KEY_u, 0x1E7B, 
GDK_KEY_space, 0x00B0, 
GDK_KEY_nobreakspace, 0x030A, 
GDK_KEY_Aacute, 0x01FA, 
GDK_KEY_aacute, 0x01FB, 
GDK_KEY_dead_abovering, 0x00B0, 
GDK_KEY_space, 0x02DD, 
GDK_KEY_nobreakspace, 0x030B, 
GDK_KEY_Cyrillic_u, 0x04F3, 
GDK_KEY_Cyrillic_U, 0x04F2, 
GDK_KEY_dead_doubleacute, 0x02DD, 
GDK_KEY_space, 0x02C7, 
GDK_KEY_parenleft, 0x208D, 
GDK_KEY_parenright, 0x208E, 
GDK_KEY_plus, 0x208A, 
GDK_KEY_minus, 0x208B, 
GDK_KEY_0, 0x2080, 
GDK_KEY_1, 0x2081, 
GDK_KEY_2, 0x2082, 
GDK_KEY_3, 0x2083, 
GDK_KEY_4, 0x2084, 
GDK_KEY_5, 0x2085, 
GDK_KEY_6, 0x2086, 
GDK_KEY_7, 0x2087, 
GDK_KEY_8, 0x2088, 
GDK_KEY_9, 0x2089, 
GDK_KEY_equal, 0x208C, 
GDK_KEY_V, 0x01D9, 
GDK_KEY_v, 0x01DA, 
GDK_KEY_nobreakspace, 0x030C, 
0x01F2, 0x01C5, 
GDK_KEY_dead_caron, 0x02C7, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_U, 0x01D9, 
GDK_KEY_Multi_key, GDK_KEY_quotedbl, GDK_KEY_u, 0x01DA, 
GDK_KEY_space, 0x00B8, 
GDK_KEY_nobreakspace, 0x0327, 
GDK_KEY_cent, 0x20B5, 
GDK_KEY_Cacute, 0x1E08, 
GDK_KEY_cacute, 0x1E09, 
GDK_KEY_dead_cedilla, 0x00B8, 
GDK_KEY_dead_currency, GDK_KEY_C, 0x20B5, 
GDK_KEY_dead_currency, GDK_KEY_c, 0x20B5, 
GDK_KEY_space, 0x02DB, 
GDK_KEY_nobreakspace, 0x0328, 
GDK_KEY_Omacron, 0x01EC, 
GDK_KEY_omacron, 0x01ED, 
GDK_KEY_dead_ogonek, 0x02DB, 
GDK_KEY_space, 0x037A, 
GDK_KEY_Greek_alphaaccent, 0x1FB4, 
GDK_KEY_Greek_etaaccent, 0x1FC4, 
GDK_KEY_Greek_omegaaccent, 0x1FF4, 
GDK_KEY_Greek_ALPHA, 0x1FBC, 
GDK_KEY_Greek_ETA, 0x1FCC, 
GDK_KEY_Greek_OMEGA, 0x1FFC, 
GDK_KEY_Greek_alpha, 0x1FB3, 
GDK_KEY_Greek_eta, 0x1FC3, 
GDK_KEY_Greek_omega, 0x1FF3, 
GDK_KEY_dead_iota, 0x037A, 
GDK_KEY_dead_grave, GDK_KEY_Greek_alpha, 0x1FB2, 
GDK_KEY_dead_grave, GDK_KEY_Greek_eta, 0x1FC2, 
GDK_KEY_dead_grave, GDK_KEY_Greek_omega, 0x1FF2, 
GDK_KEY_dead_acute, GDK_KEY_Greek_alpha, 0x1FB4, 
GDK_KEY_dead_acute, GDK_KEY_Greek_eta, 0x1FC4, 
GDK_KEY_dead_acute, GDK_KEY_Greek_omega, 0x1FF4, 
GDK_KEY_dead_tilde, GDK_KEY_Greek_alpha, 0x1FB7, 
GDK_KEY_dead_tilde, GDK_KEY_Greek_eta, 0x1FC7, 
GDK_KEY_dead_tilde, GDK_KEY_Greek_omega, 0x1FF7, 
GDK_KEY_dead_tilde, 0x1F00, 0x1F86, 
GDK_KEY_dead_tilde, 0x1F01, 0x1F87, 
GDK_KEY_dead_tilde, 0x1F08, 0x1F8E, 
GDK_KEY_dead_tilde, 0x1F09, 0x1F8F, 
GDK_KEY_dead_tilde, 0x1F20, 0x1F96, 
GDK_KEY_dead_tilde, 0x1F21, 0x1F97, 
GDK_KEY_dead_tilde, 0x1F28, 0x1F9E, 
GDK_KEY_dead_tilde, 0x1F29, 0x1F9F, 
GDK_KEY_dead_tilde, 0x1F60, 0x1FA6, 
GDK_KEY_dead_tilde, 0x1F61, 0x1FA7, 
GDK_KEY_dead_tilde, 0x1F68, 0x1FAE, 
GDK_KEY_dead_tilde, 0x1F69, 0x1FAF, 
GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F88, 
GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F98, 
GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FA8, 
GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F80, 
GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F90, 
GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA0, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F89, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F99, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FA9, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F81, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F91, 
GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA1, 
GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8A, 
GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9A, 
GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAA, 
GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F82, 
GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F92, 
GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA2, 
GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8B, 
GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9B, 
GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAB, 
GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F83, 
GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F93, 
GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA3, 
GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8E, 
GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9E, 
GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAE, 
GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F86, 
GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F96, 
GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA6, 
GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8F, 
GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9F, 
GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAF, 
GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F87, 
GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F97, 
GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA7, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_Greek_alpha, 0x1FB4, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_Greek_eta, 0x1FC4, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_Greek_omega, 0x1FF4, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F00, 0x1F84, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F01, 0x1F85, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F08, 0x1F8C, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F09, 0x1F8D, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F20, 0x1F94, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F21, 0x1F95, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F28, 0x1F9C, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F29, 0x1F9D, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F60, 0x1FA4, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F61, 0x1FA5, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F68, 0x1FAC, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, 0x1F69, 0x1FAD, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F89, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F99, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FA9, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F81, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F91, 
GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA1, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F88, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F98, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FA8, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F80, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F90, 
GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA0, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_Greek_alpha, 0x1FB2, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_Greek_eta, 0x1FC2, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_Greek_omega, 0x1FF2, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F00, 0x1F82, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F01, 0x1F83, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F08, 0x1F8A, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F09, 0x1F8B, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F20, 0x1F92, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F21, 0x1F93, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F28, 0x1F9A, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F29, 0x1F9B, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F60, 0x1FA2, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F61, 0x1FA3, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F68, 0x1FAA, 
GDK_KEY_Multi_key, GDK_KEY_grave, 0x1F69, 0x1FAB, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_Greek_alpha, 0x1FB7, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_Greek_eta, 0x1FC7, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_Greek_omega, 0x1FF7, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F00, 0x1F86, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F01, 0x1F87, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F08, 0x1F8E, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F09, 0x1F8F, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F20, 0x1F96, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F21, 0x1F97, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F28, 0x1F9E, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F29, 0x1F9F, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F60, 0x1FA6, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F61, 0x1FA7, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F68, 0x1FAE, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, 0x1F69, 0x1FAF, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_Greek_alpha, 0x1FB4, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_Greek_eta, 0x1FC4, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_Greek_omega, 0x1FF4, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F00, 0x1F84, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F01, 0x1F85, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F08, 0x1F8C, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F09, 0x1F8D, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F20, 0x1F94, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F21, 0x1F95, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F28, 0x1F9C, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F29, 0x1F9D, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F60, 0x1FA4, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F61, 0x1FA5, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F68, 0x1FAC, 
GDK_KEY_Multi_key, GDK_KEY_acute, 0x1F69, 0x1FAD, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8B, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9B, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAB, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F83, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F93, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA3, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8A, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9A, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAA, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F82, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F92, 
GDK_KEY_dead_grave, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA2, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_dead_acute, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8F, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9F, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAF, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F87, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F97, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA7, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8E, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9E, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAE, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F86, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F96, 
GDK_KEY_dead_tilde, GDK_KEY_Multi_key, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA6, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_Multi_key, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8B, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9B, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAB, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F83, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F93, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA3, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8A, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9A, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAA, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F82, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F92, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA2, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8A, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9A, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAA, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F82, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F92, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA2, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8B, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9B, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAB, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F83, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F93, 
GDK_KEY_Multi_key, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA3, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8F, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9F, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAF, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F87, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F97, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA7, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8E, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9E, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAE, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F86, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F96, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA6, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8E, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9E, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAE, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F86, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F96, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA6, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8F, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9F, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAF, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F87, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F97, 
GDK_KEY_Multi_key, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA7, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_Multi_key, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_kana_WO, 0x30FA, 
GDK_KEY_kana_U, 0x30F4, 
GDK_KEY_kana_KA, 0x30AC, 
GDK_KEY_kana_KI, 0x30AE, 
GDK_KEY_kana_KU, 0x30B0, 
GDK_KEY_kana_KE, 0x30B2, 
GDK_KEY_kana_KO, 0x30B4, 
GDK_KEY_kana_SA, 0x30B6, 
GDK_KEY_kana_SHI, 0x30B8, 
GDK_KEY_kana_SU, 0x30BA, 
GDK_KEY_kana_SE, 0x30BC, 
GDK_KEY_kana_SO, 0x30BE, 
GDK_KEY_kana_TA, 0x30C0, 
GDK_KEY_kana_CHI, 0x30C2, 
GDK_KEY_kana_TSU, 0x30C5, 
GDK_KEY_kana_TE, 0x30C7, 
GDK_KEY_kana_TO, 0x30C9, 
GDK_KEY_kana_HA, 0x30D0, 
GDK_KEY_kana_HI, 0x30D3, 
GDK_KEY_kana_FU, 0x30D6, 
GDK_KEY_kana_HE, 0x30D9, 
GDK_KEY_kana_HO, 0x30DC, 
GDK_KEY_kana_WA, 0x30F7, 
GDK_KEY_kana_HA, 0x30D1, 
GDK_KEY_kana_HI, 0x30D4, 
GDK_KEY_kana_FU, 0x30D7, 
GDK_KEY_kana_HE, 0x30DA, 
GDK_KEY_kana_HO, 0x30DD, 
GDK_KEY_space, 0x0323, 
GDK_KEY_plus, 0x2A25, 
GDK_KEY_minus, 0x2A2A, 
GDK_KEY_equal, 0x2A66, 
GDK_KEY_nobreakspace, 0x0323, 
GDK_KEY_Abreve, 0x1EB6, 
GDK_KEY_abreve, 0x1EB7, 
GDK_KEY_dead_belowdot, 0x0323, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_O, 0x1EE2, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_U, 0x1EF0, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_o, 0x1EE3, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_u, 0x1EF1, 
GDK_KEY_space, 0x0309, 
GDK_KEY_B, 0x0181, 
GDK_KEY_C, 0x0187, 
GDK_KEY_D, 0x018A, 
GDK_KEY_F, 0x0191, 
GDK_KEY_G, 0x0193, 
GDK_KEY_K, 0x0198, 
GDK_KEY_M, 0x2C6E, 
GDK_KEY_N, 0x019D, 
GDK_KEY_P, 0x01A4, 
GDK_KEY_T, 0x01AC, 
GDK_KEY_V, 0x01B2, 
GDK_KEY_W, 0x2C72, 
GDK_KEY_Z, 0x0224, 
GDK_KEY_b, 0x0253, 
GDK_KEY_c, 0x0188, 
GDK_KEY_d, 0x0257, 
GDK_KEY_f, 0x0192, 
GDK_KEY_g, 0x0260, 
GDK_KEY_h, 0x0266, 
GDK_KEY_k, 0x0199, 
GDK_KEY_m, 0x0271, 
GDK_KEY_n, 0x0272, 
GDK_KEY_p, 0x01A5, 
GDK_KEY_q, 0x02A0, 
GDK_KEY_r, 0x027C, 
GDK_KEY_s, 0x0282, 
GDK_KEY_t, 0x01AD, 
GDK_KEY_v, 0x028B, 
GDK_KEY_w, 0x2C73, 
GDK_KEY_z, 0x0225, 
GDK_KEY_nobreakspace, 0x0309, 
GDK_KEY_Abreve, 0x1EB2, 
GDK_KEY_abreve, 0x1EB3, 
0x0256, 0x1D91, 
0x025C, 0x025D, 
0x025F, 0x0284, 
0x0279, 0x027B, 
GDK_KEY_dead_hook, 0x0309, 
GDK_KEY_dead_greek, GDK_KEY_U, 0x03D2, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_O, 0x1EDE, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_U, 0x1EEC, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_o, 0x1EDF, 
GDK_KEY_Multi_key, GDK_KEY_plus, GDK_KEY_u, 0x1EED, 
GDK_KEY_Multi_key, GDK_KEY_U, GDK_KEY_A, 0x1EB2, 
GDK_KEY_Multi_key, GDK_KEY_U, GDK_KEY_a, 0x1EB3, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_A, 0x1EA8, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_E, 0x1EC2, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_O, 0x1ED4, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_a, 0x1EA9, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_e, 0x1EC3, 
GDK_KEY_Multi_key, GDK_KEY_asciicircum, GDK_KEY_o, 0x1ED5, 
GDK_KEY_Multi_key, GDK_KEY_b, GDK_KEY_A, 0x1EB2, 
GDK_KEY_Multi_key, GDK_KEY_b, GDK_KEY_a, 0x1EB3, 
GDK_KEY_space, 0x031B, 
GDK_KEY_nobreakspace, 0x031B, 
GDK_KEY_Utilde, 0x1EEE, 
GDK_KEY_utilde, 0x1EEF, 
GDK_KEY_dead_horn, 0x031B, 
GDK_KEY_space, 0x002F, 
GDK_KEY_2, 0x01BB, 
GDK_KEY_A, 0x023A, 
GDK_KEY_B, 0x0243, 
GDK_KEY_C, 0x023B, 
GDK_KEY_D, 0x0110, 
GDK_KEY_E, 0x0246, 
GDK_KEY_G, 0x01E4, 
GDK_KEY_H, 0x0126, 
GDK_KEY_I, 0x0197, 
GDK_KEY_J, 0x0248, 
GDK_KEY_L, 0x0141, 
GDK_KEY_O, 0x00D8, 
GDK_KEY_P, 0x2C63, 
GDK_KEY_R, 0x024C, 
GDK_KEY_T, 0x0166, 
GDK_KEY_U, 0x0244, 
GDK_KEY_Y, 0x024E, 
GDK_KEY_Z, 0x01B5, 
GDK_KEY_a, 0x2C65, 
GDK_KEY_b, 0x0180, 
GDK_KEY_c, 0x023C, 
GDK_KEY_d, 0x0111, 
GDK_KEY_e, 0x0247, 
GDK_KEY_g, 0x01E5, 
GDK_KEY_h, 0x0127, 
GDK_KEY_i, 0x0268, 
GDK_KEY_j, 0x0249, 
GDK_KEY_l, 0x0142, 
GDK_KEY_o, 0x00F8, 
GDK_KEY_p, 0x1D7D, 
GDK_KEY_r, 0x024D, 
GDK_KEY_t, 0x0167, 
GDK_KEY_u, 0x0289, 
GDK_KEY_y, 0x024F, 
GDK_KEY_z, 0x01B6, 
GDK_KEY_nobreakspace, 0x0338, 
GDK_KEY_Oacute, 0x01FE, 
GDK_KEY_oacute, 0x01FF, 
0x0237, 0x025F, 
0x0269, 0x1D7C, 
GDK_KEY_lessthanequal, 0x2270, 
GDK_KEY_greaterthanequal, 0x2271, 
GDK_KEY_dead_stroke, 0x002F, 
GDK_KEY_dead_acute, GDK_KEY_O, 0x01FE, 
GDK_KEY_dead_acute, GDK_KEY_o, 0x01FF, 
GDK_KEY_dead_abovedot, GDK_KEY_j, 0x025F, 
GDK_KEY_dead_greek, GDK_KEY_r, 0x03FC, 
GDK_KEY_Greek_ALPHA, 0x1F08, 
GDK_KEY_Greek_EPSILON, 0x1F18, 
GDK_KEY_Greek_ETA, 0x1F28, 
GDK_KEY_Greek_IOTA, 0x1F38, 
GDK_KEY_Greek_OMICRON, 0x1F48, 
GDK_KEY_Greek_OMEGA, 0x1F68, 
GDK_KEY_Greek_alpha, 0x1F00, 
GDK_KEY_Greek_epsilon, 0x1F10, 
GDK_KEY_Greek_eta, 0x1F20, 
GDK_KEY_Greek_iota, 0x1F30, 
GDK_KEY_Greek_omicron, 0x1F40, 
GDK_KEY_Greek_rho, 0x1FE4, 
GDK_KEY_Greek_upsilon, 0x1F50, 
GDK_KEY_Greek_omega, 0x1F60, 
GDK_KEY_Greek_ALPHA, 0x1F09, 
GDK_KEY_Greek_EPSILON, 0x1F19, 
GDK_KEY_Greek_ETA, 0x1F29, 
GDK_KEY_Greek_IOTA, 0x1F39, 
GDK_KEY_Greek_OMICRON, 0x1F49, 
GDK_KEY_Greek_RHO, 0x1FEC, 
GDK_KEY_Greek_UPSILON, 0x1F59, 
GDK_KEY_Greek_OMEGA, 0x1F69, 
GDK_KEY_Greek_alpha, 0x1F01, 
GDK_KEY_Greek_epsilon, 0x1F11, 
GDK_KEY_Greek_eta, 0x1F21, 
GDK_KEY_Greek_iota, 0x1F31, 
GDK_KEY_Greek_omicron, 0x1F41, 
GDK_KEY_Greek_rho, 0x1FE5, 
GDK_KEY_Greek_upsilon, 0x1F51, 
GDK_KEY_Greek_omega, 0x1F61, 
GDK_KEY_bar, 0x2AF0, 
GDK_KEY_plus, 0x2A26, 
GDK_KEY_dead_diaeresis, GDK_KEY_equal, 0x2A77, 
GDK_KEY_space, 0x002C, 
GDK_KEY_S, 0x0218, 
GDK_KEY_T, 0x021A, 
GDK_KEY_s, 0x0219, 
GDK_KEY_t, 0x021B, 
GDK_KEY_nobreakspace, 0x0326, 
GDK_KEY_dead_belowcomma, 0x002C, 
GDK_KEY_space, 0x00A4, 
GDK_KEY_A, 0x20B3, 
GDK_KEY_B, 0x20B1, 
GDK_KEY_C, 0x20A1, 
GDK_KEY_D, 0x20AF, 
GDK_KEY_E, 0x20A0, 
GDK_KEY_F, 0x20A3, 
GDK_KEY_G, 0x20B2, 
GDK_KEY_H, 0x20B4, 
GDK_KEY_I, 0x17DB, 
GDK_KEY_K, 0x20AD, 
GDK_KEY_L, 0x20A4, 
GDK_KEY_M, 0x2133, 
GDK_KEY_N, 0x20A6, 
GDK_KEY_O, 0x0AF1, 
GDK_KEY_P, 0x20A7, 
GDK_KEY_R, 0x20A8, 
GDK_KEY_S, 0x0024, 
GDK_KEY_T, 0x20AE, 
GDK_KEY_U, 0x5713, 
GDK_KEY_W, 0x20A9, 
GDK_KEY_Y, 0x5186, 
GDK_KEY_a, 0x060B, 
GDK_KEY_b, 0x0E3F, 
GDK_KEY_c, 0x00A2, 
GDK_KEY_d, 0x20AB, 
GDK_KEY_e, 0x20AC, 
GDK_KEY_f, 0x0192, 
GDK_KEY_g, 0x20B2, 
GDK_KEY_h, 0x20B4, 
GDK_KEY_i, 0xFDFC, 
GDK_KEY_k, 0x20AD, 
GDK_KEY_l, 0x00A3, 
GDK_KEY_m, 0x20A5, 
GDK_KEY_n, 0x20A6, 
GDK_KEY_o, 0x0BF9, 
GDK_KEY_p, 0x20B0, 
GDK_KEY_r, 0x20A2, 
GDK_KEY_s, 0x20AA, 
GDK_KEY_t, 0x09F3, 
GDK_KEY_u, 0x5143, 
GDK_KEY_w, 0x20A9, 
GDK_KEY_y, 0x00A5, 
GDK_KEY_nobreakspace, 0x00A4, 
GDK_KEY_Ccedilla, 0x20B5, 
GDK_KEY_THORN, 0x09F2, 
GDK_KEY_ccedilla, 0x20B5, 
GDK_KEY_thorn, 0x09F2, 
GDK_KEY_dead_currency, 0x00A4, 
GDK_KEY_dead_cedilla, GDK_KEY_C, 0x20B5, 
GDK_KEY_dead_cedilla, GDK_KEY_c, 0x20B5, 
GDK_KEY_space, 0x00B5, 
GDK_KEY_A, 0x0391, 
GDK_KEY_B, 0x0392, 
GDK_KEY_D, 0x0394, 
GDK_KEY_E, 0x0395, 
GDK_KEY_F, 0x03A6, 
GDK_KEY_G, 0x0393, 
GDK_KEY_H, 0x0397, 
GDK_KEY_I, 0x0399, 
GDK_KEY_J, 0x0398, 
GDK_KEY_K, 0x039A, 
GDK_KEY_L, 0x039B, 
GDK_KEY_M, 0x039C, 
GDK_KEY_N, 0x039D, 
GDK_KEY_O, 0x039F, 
GDK_KEY_P, 0x03A0, 
GDK_KEY_Q, 0x03A7, 
GDK_KEY_R, 0x03A1, 
GDK_KEY_S, 0x03A3, 
GDK_KEY_T, 0x03A4, 
GDK_KEY_U, 0x03A5, 
GDK_KEY_W, 0x03A9, 
GDK_KEY_X, 0x039E, 
GDK_KEY_Y, 0x03A8, 
GDK_KEY_Z, 0x0396, 
GDK_KEY_a, 0x03B1, 
GDK_KEY_b, 0x03B2, 
GDK_KEY_d, 0x03B4, 
GDK_KEY_e, 0x03B5, 
GDK_KEY_f, 0x03C6, 
GDK_KEY_g, 0x03B3, 
GDK_KEY_h, 0x03B7, 
GDK_KEY_i, 0x03B9, 
GDK_KEY_j, 0x03B8, 
GDK_KEY_k, 0x03BA, 
GDK_KEY_l, 0x03BB, 
GDK_KEY_m, 0x03BC, 
GDK_KEY_n, 0x03BD, 
GDK_KEY_o, 0x03BF, 
GDK_KEY_p, 0x03C0, 
GDK_KEY_q, 0x03C7, 
GDK_KEY_r, 0x03C1, 
GDK_KEY_s, 0x03C3, 
GDK_KEY_t, 0x03C4, 
GDK_KEY_u, 0x03C5, 
GDK_KEY_w, 0x03C9, 
GDK_KEY_x, 0x03BE, 
GDK_KEY_y, 0x03C8, 
GDK_KEY_z, 0x03B6, 
GDK_KEY_nobreakspace, 0x00B5, 
GDK_KEY_dead_greek, 0x00B5, 
GDK_KEY_dead_macron, GDK_KEY_A, 0x1FB9, 
GDK_KEY_dead_macron, GDK_KEY_I, 0x1FD9, 
GDK_KEY_dead_macron, GDK_KEY_U, 0x1FE9, 
GDK_KEY_dead_macron, GDK_KEY_a, 0x1FB1, 
GDK_KEY_dead_macron, GDK_KEY_i, 0x1FD1, 
GDK_KEY_dead_macron, GDK_KEY_u, 0x1FE1, 
GDK_KEY_dead_hook, GDK_KEY_U, 0x03D2, 
GDK_KEY_dead_stroke, GDK_KEY_r, 0x03FC, 
GDK_KEY_space, GDK_KEY_space, 0x00A0, 
GDK_KEY_space, GDK_KEY_apostrophe, 0x0027, 
GDK_KEY_space, GDK_KEY_parenleft, 0x02D8, 
GDK_KEY_space, GDK_KEY_comma, 0x00B8, 
GDK_KEY_space, GDK_KEY_minus, 0x007E, 
GDK_KEY_space, GDK_KEY_period, 0x2008, 
GDK_KEY_space, GDK_KEY_less, 0x02C7, 
GDK_KEY_space, GDK_KEY_greater, 0x005E, 
GDK_KEY_space, GDK_KEY_asciicircum, 0x005E, 
GDK_KEY_space, GDK_KEY_grave, 0x0060, 
GDK_KEY_space, GDK_KEY_asciitilde, 0x007E, 
GDK_KEY_exclam, GDK_KEY_exclam, 0x00A1, 
GDK_KEY_exclam, GDK_KEY_question, 0x203D, 
GDK_KEY_exclam, GDK_KEY_A, 0x1EA0, 
GDK_KEY_exclam, GDK_KEY_B, 0x1E04, 
GDK_KEY_exclam, GDK_KEY_D, 0x1E0C, 
GDK_KEY_exclam, GDK_KEY_E, 0x1EB8, 
GDK_KEY_exclam, GDK_KEY_H, 0x1E24, 
GDK_KEY_exclam, GDK_KEY_I, 0x1ECA, 
GDK_KEY_exclam, GDK_KEY_K, 0x1E32, 
GDK_KEY_exclam, GDK_KEY_L, 0x1E36, 
GDK_KEY_exclam, GDK_KEY_M, 0x1E42, 
GDK_KEY_exclam, GDK_KEY_N, 0x1E46, 
GDK_KEY_exclam, GDK_KEY_O, 0x1ECC, 
GDK_KEY_exclam, GDK_KEY_R, 0x1E5A, 
GDK_KEY_exclam, GDK_KEY_S, 0x1E62, 
GDK_KEY_exclam, GDK_KEY_T, 0x1E6C, 
GDK_KEY_exclam, GDK_KEY_U, 0x1EE4, 
GDK_KEY_exclam, GDK_KEY_V, 0x1E7E, 
GDK_KEY_exclam, GDK_KEY_W, 0x1E88, 
GDK_KEY_exclam, GDK_KEY_Y, 0x1EF4, 
GDK_KEY_exclam, GDK_KEY_Z, 0x1E92, 
GDK_KEY_exclam, GDK_KEY_asciicircum, 0x00A6, 
GDK_KEY_exclam, GDK_KEY_a, 0x1EA1, 
GDK_KEY_exclam, GDK_KEY_b, 0x1E05, 
GDK_KEY_exclam, GDK_KEY_d, 0x1E0D, 
GDK_KEY_exclam, GDK_KEY_e, 0x1EB9, 
GDK_KEY_exclam, GDK_KEY_h, 0x1E25, 
GDK_KEY_exclam, GDK_KEY_i, 0x1ECB, 
GDK_KEY_exclam, GDK_KEY_k, 0x1E33, 
GDK_KEY_exclam, GDK_KEY_l, 0x1E37, 
GDK_KEY_exclam, GDK_KEY_m, 0x1E43, 
GDK_KEY_exclam, GDK_KEY_n, 0x1E47, 
GDK_KEY_exclam, GDK_KEY_o, 0x1ECD, 
GDK_KEY_exclam, GDK_KEY_r, 0x1E5B, 
GDK_KEY_exclam, GDK_KEY_s, 0x1E63, 
GDK_KEY_exclam, GDK_KEY_t, 0x1E6D, 
GDK_KEY_exclam, GDK_KEY_u, 0x1EE5, 
GDK_KEY_exclam, GDK_KEY_v, 0x1E7F, 
GDK_KEY_exclam, GDK_KEY_w, 0x1E89, 
GDK_KEY_exclam, GDK_KEY_y, 0x1EF5, 
GDK_KEY_exclam, GDK_KEY_z, 0x1E93, 
GDK_KEY_quotedbl, GDK_KEY_quotedbl, 0x00A8, 
GDK_KEY_quotedbl, GDK_KEY_apostrophe, 0x0344, 
GDK_KEY_quotedbl, GDK_KEY_comma, 0x201E, 
GDK_KEY_quotedbl, GDK_KEY_less, 0x201C, 
GDK_KEY_quotedbl, GDK_KEY_greater, 0x201D, 
GDK_KEY_quotedbl, GDK_KEY_A, 0x00C4, 
GDK_KEY_quotedbl, GDK_KEY_E, 0x00CB, 
GDK_KEY_quotedbl, GDK_KEY_H, 0x1E26, 
GDK_KEY_quotedbl, GDK_KEY_I, 0x00CF, 
GDK_KEY_quotedbl, GDK_KEY_O, 0x00D6, 
GDK_KEY_quotedbl, GDK_KEY_U, 0x00DC, 
GDK_KEY_quotedbl, GDK_KEY_W, 0x1E84, 
GDK_KEY_quotedbl, GDK_KEY_X, 0x1E8C, 
GDK_KEY_quotedbl, GDK_KEY_Y, 0x0178, 
GDK_KEY_quotedbl, GDK_KEY_a, 0x00E4, 
GDK_KEY_quotedbl, GDK_KEY_e, 0x00EB, 
GDK_KEY_quotedbl, GDK_KEY_h, 0x1E27, 
GDK_KEY_quotedbl, GDK_KEY_i, 0x00EF, 
GDK_KEY_quotedbl, GDK_KEY_o, 0x00F6, 
GDK_KEY_quotedbl, GDK_KEY_t, 0x1E97, 
GDK_KEY_quotedbl, GDK_KEY_u, 0x00FC, 
GDK_KEY_quotedbl, GDK_KEY_w, 0x1E85, 
GDK_KEY_quotedbl, GDK_KEY_x, 0x1E8D, 
GDK_KEY_quotedbl, GDK_KEY_y, 0x00FF, 
GDK_KEY_quotedbl, GDK_KEY_acute, 0x0344, 
GDK_KEY_quotedbl, GDK_KEY_Otilde, 0x1E4E, 
GDK_KEY_quotedbl, GDK_KEY_otilde, 0x1E4F, 
GDK_KEY_quotedbl, 0x03D2, 0x03D4, 
GDK_KEY_quotedbl, GDK_KEY_Umacron, 0x1E7A, 
GDK_KEY_quotedbl, GDK_KEY_umacron, 0x1E7B, 
GDK_KEY_quotedbl, 0x04D8, 0x04DA, 
GDK_KEY_quotedbl, 0x04D9, 0x04DB, 
GDK_KEY_quotedbl, 0x04E8, 0x04EA, 
GDK_KEY_quotedbl, 0x04E9, 0x04EB, 
GDK_KEY_quotedbl, GDK_KEY_Ukrainian_i, 0x0457, 
GDK_KEY_quotedbl, GDK_KEY_Ukrainian_I, 0x0407, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_a, 0x04D3, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_ie, 0x0451, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_i, 0x04E5, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_o, 0x04E7, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_u, 0x04F1, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_zhe, 0x04DD, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_yeru, 0x04F9, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_ze, 0x04DF, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_e, 0x04ED, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_che, 0x04F5, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_A, 0x04D2, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_IE, 0x0401, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_I, 0x04E4, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_O, 0x04E6, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_U, 0x04F0, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_ZHE, 0x04DC, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_YERU, 0x04F8, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_ZE, 0x04DE, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_E, 0x04EC, 
GDK_KEY_quotedbl, GDK_KEY_Cyrillic_CHE, 0x04F4, 
GDK_KEY_quotedbl, GDK_KEY_Greek_IOTA, 0x03AA, 
GDK_KEY_quotedbl, GDK_KEY_Greek_UPSILON, 0x03AB, 
GDK_KEY_quotedbl, GDK_KEY_Greek_iota, 0x03CA, 
GDK_KEY_quotedbl, GDK_KEY_Greek_upsilon, 0x03CB, 
GDK_KEY_quotedbl, GDK_KEY_dead_acute, 0x0344, 
GDK_KEY_numbersign, GDK_KEY_numbersign, 0x266F, 
GDK_KEY_numbersign, GDK_KEY_E, 0x266B, 
GDK_KEY_numbersign, GDK_KEY_S, 0x266C, 
GDK_KEY_numbersign, GDK_KEY_b, 0x266D, 
GDK_KEY_numbersign, GDK_KEY_e, 0x266A, 
GDK_KEY_numbersign, GDK_KEY_f, 0x266E, 
GDK_KEY_numbersign, GDK_KEY_q, 0x2669, 
GDK_KEY_percent, GDK_KEY_o, 0x2030, 
GDK_KEY_apostrophe, GDK_KEY_space, 0x0027, 
GDK_KEY_apostrophe, GDK_KEY_apostrophe, 0x00B4, 
GDK_KEY_apostrophe, GDK_KEY_comma, 0x201A, 
GDK_KEY_apostrophe, GDK_KEY_less, 0x2018, 
GDK_KEY_apostrophe, GDK_KEY_greater, 0x2019, 
GDK_KEY_apostrophe, GDK_KEY_A, 0x00C1, 
GDK_KEY_apostrophe, GDK_KEY_C, 0x0106, 
GDK_KEY_apostrophe, GDK_KEY_E, 0x00C9, 
GDK_KEY_apostrophe, GDK_KEY_G, 0x01F4, 
GDK_KEY_apostrophe, GDK_KEY_I, 0x00CD, 
GDK_KEY_apostrophe, GDK_KEY_K, 0x1E30, 
GDK_KEY_apostrophe, GDK_KEY_L, 0x0139, 
GDK_KEY_apostrophe, GDK_KEY_M, 0x1E3E, 
GDK_KEY_apostrophe, GDK_KEY_N, 0x0143, 
GDK_KEY_apostrophe, GDK_KEY_O, 0x00D3, 
GDK_KEY_apostrophe, GDK_KEY_P, 0x1E54, 
GDK_KEY_apostrophe, GDK_KEY_R, 0x0154, 
GDK_KEY_apostrophe, GDK_KEY_S, 0x015A, 
GDK_KEY_apostrophe, GDK_KEY_U, 0x00DA, 
GDK_KEY_apostrophe, GDK_KEY_W, 0x1E82, 
GDK_KEY_apostrophe, GDK_KEY_Y, 0x00DD, 
GDK_KEY_apostrophe, GDK_KEY_Z, 0x0179, 
GDK_KEY_apostrophe, GDK_KEY_a, 0x00E1, 
GDK_KEY_apostrophe, GDK_KEY_c, 0x0107, 
GDK_KEY_apostrophe, GDK_KEY_e, 0x00E9, 
GDK_KEY_apostrophe, GDK_KEY_g, 0x01F5, 
GDK_KEY_apostrophe, GDK_KEY_i, 0x00ED, 
GDK_KEY_apostrophe, GDK_KEY_k, 0x1E31, 
GDK_KEY_apostrophe, GDK_KEY_l, 0x013A, 
GDK_KEY_apostrophe, GDK_KEY_m, 0x1E3F, 
GDK_KEY_apostrophe, GDK_KEY_n, 0x0144, 
GDK_KEY_apostrophe, GDK_KEY_o, 0x00F3, 
GDK_KEY_apostrophe, GDK_KEY_p, 0x1E55, 
GDK_KEY_apostrophe, GDK_KEY_r, 0x0155, 
GDK_KEY_apostrophe, GDK_KEY_s, 0x015B, 
GDK_KEY_apostrophe, GDK_KEY_u, 0x00FA, 
GDK_KEY_apostrophe, GDK_KEY_w, 0x1E83, 
GDK_KEY_apostrophe, GDK_KEY_y, 0x00FD, 
GDK_KEY_apostrophe, GDK_KEY_z, 0x017A, 
GDK_KEY_apostrophe, GDK_KEY_Acircumflex, 0x1EA4, 
GDK_KEY_apostrophe, GDK_KEY_Aring, 0x01FA, 
GDK_KEY_apostrophe, GDK_KEY_AE, 0x01FC, 
GDK_KEY_apostrophe, GDK_KEY_Ccedilla, 0x1E08, 
GDK_KEY_apostrophe, GDK_KEY_Ecircumflex, 0x1EBE, 
GDK_KEY_apostrophe, GDK_KEY_Idiaeresis, 0x1E2E, 
GDK_KEY_apostrophe, GDK_KEY_Ocircumflex, 0x1ED0, 
GDK_KEY_apostrophe, GDK_KEY_Otilde, 0x1E4C, 
GDK_KEY_apostrophe, GDK_KEY_Ooblique, 0x01FE, 
GDK_KEY_apostrophe, GDK_KEY_Udiaeresis, 0x01D7, 
GDK_KEY_apostrophe, GDK_KEY_acircumflex, 0x1EA5, 
GDK_KEY_apostrophe, GDK_KEY_aring, 0x01FB, 
GDK_KEY_apostrophe, GDK_KEY_ae, 0x01FD, 
GDK_KEY_apostrophe, GDK_KEY_ccedilla, 0x1E09, 
GDK_KEY_apostrophe, GDK_KEY_ecircumflex, 0x1EBF, 
GDK_KEY_apostrophe, GDK_KEY_idiaeresis, 0x1E2F, 
GDK_KEY_apostrophe, GDK_KEY_ocircumflex, 0x1ED1, 
GDK_KEY_apostrophe, GDK_KEY_otilde, 0x1E4D, 
GDK_KEY_apostrophe, GDK_KEY_oslash, 0x01FF, 
GDK_KEY_apostrophe, GDK_KEY_udiaeresis, 0x01D8, 
GDK_KEY_apostrophe, GDK_KEY_Abreve, 0x1EAE, 
GDK_KEY_apostrophe, GDK_KEY_abreve, 0x1EAF, 
GDK_KEY_apostrophe, GDK_KEY_Emacron, 0x1E16, 
GDK_KEY_apostrophe, GDK_KEY_emacron, 0x1E17, 
GDK_KEY_apostrophe, GDK_KEY_Omacron, 0x1E52, 
GDK_KEY_apostrophe, GDK_KEY_Utilde, 0x1E78, 
GDK_KEY_apostrophe, GDK_KEY_omacron, 0x1E53, 
GDK_KEY_apostrophe, GDK_KEY_utilde, 0x1E79, 
GDK_KEY_apostrophe, GDK_KEY_Cyrillic_ghe, 0x0453, 
GDK_KEY_apostrophe, GDK_KEY_Cyrillic_ka, 0x045C, 
GDK_KEY_apostrophe, GDK_KEY_Cyrillic_GHE, 0x0403, 
GDK_KEY_apostrophe, GDK_KEY_Cyrillic_KA, 0x040C, 
GDK_KEY_apostrophe, GDK_KEY_Greek_iotadieresis, 0x0390, 
GDK_KEY_apostrophe, GDK_KEY_Greek_upsilondieresis, 0x03B0, 
GDK_KEY_apostrophe, GDK_KEY_Greek_ALPHA, 0x0386, 
GDK_KEY_apostrophe, GDK_KEY_Greek_EPSILON, 0x0388, 
GDK_KEY_apostrophe, GDK_KEY_Greek_ETA, 0x0389, 
GDK_KEY_apostrophe, GDK_KEY_Greek_IOTA, 0x038A, 
GDK_KEY_apostrophe, GDK_KEY_Greek_OMICRON, 0x038C, 
GDK_KEY_apostrophe, GDK_KEY_Greek_UPSILON, 0x038E, 
GDK_KEY_apostrophe, GDK_KEY_Greek_OMEGA, 0x038F, 
GDK_KEY_apostrophe, GDK_KEY_Greek_alpha, 0x03AC, 
GDK_KEY_apostrophe, GDK_KEY_Greek_epsilon, 0x03AD, 
GDK_KEY_apostrophe, GDK_KEY_Greek_eta, 0x03AE, 
GDK_KEY_apostrophe, GDK_KEY_Greek_iota, 0x03AF, 
GDK_KEY_apostrophe, GDK_KEY_Greek_omicron, 0x03CC, 
GDK_KEY_apostrophe, GDK_KEY_Greek_upsilon, 0x03CD, 
GDK_KEY_apostrophe, GDK_KEY_Greek_omega, 0x03CE, 
GDK_KEY_apostrophe, 0x1F00, 0x1F04, 
GDK_KEY_apostrophe, 0x1F01, 0x1F05, 
GDK_KEY_apostrophe, 0x1F08, 0x1F0C, 
GDK_KEY_apostrophe, 0x1F09, 0x1F0D, 
GDK_KEY_apostrophe, 0x1F10, 0x1F14, 
GDK_KEY_apostrophe, 0x1F11, 0x1F15, 
GDK_KEY_apostrophe, 0x1F18, 0x1F1C, 
GDK_KEY_apostrophe, 0x1F19, 0x1F1D, 
GDK_KEY_apostrophe, 0x1F20, 0x1F24, 
GDK_KEY_apostrophe, 0x1F21, 0x1F25, 
GDK_KEY_apostrophe, 0x1F28, 0x1F2C, 
GDK_KEY_apostrophe, 0x1F29, 0x1F2D, 
GDK_KEY_apostrophe, 0x1F30, 0x1F34, 
GDK_KEY_apostrophe, 0x1F31, 0x1F35, 
GDK_KEY_apostrophe, 0x1F38, 0x1F3C, 
GDK_KEY_apostrophe, 0x1F39, 0x1F3D, 
GDK_KEY_apostrophe, 0x1F40, 0x1F44, 
GDK_KEY_apostrophe, 0x1F41, 0x1F45, 
GDK_KEY_apostrophe, 0x1F48, 0x1F4C, 
GDK_KEY_apostrophe, 0x1F49, 0x1F4D, 
GDK_KEY_apostrophe, 0x1F50, 0x1F54, 
GDK_KEY_apostrophe, 0x1F51, 0x1F55, 
GDK_KEY_apostrophe, 0x1F59, 0x1F5D, 
GDK_KEY_apostrophe, 0x1F60, 0x1F64, 
GDK_KEY_apostrophe, 0x1F61, 0x1F65, 
GDK_KEY_apostrophe, 0x1F68, 0x1F6C, 
GDK_KEY_apostrophe, 0x1F69, 0x1F6D, 
GDK_KEY_apostrophe, 0x2395, 0x235E, 
GDK_KEY_parenleft, GDK_KEY_space, 0x02D8, 
GDK_KEY_parenleft, GDK_KEY_parenleft, 0x005B, 
GDK_KEY_parenleft, GDK_KEY_minus, 0x007B, 
GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F09, 
GDK_KEY_parenleft, GDK_KEY_Greek_EPSILON, 0x1F19, 
GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F29, 
GDK_KEY_parenleft, GDK_KEY_Greek_IOTA, 0x1F39, 
GDK_KEY_parenleft, GDK_KEY_Greek_OMICRON, 0x1F49, 
GDK_KEY_parenleft, GDK_KEY_Greek_RHO, 0x1FEC, 
GDK_KEY_parenleft, GDK_KEY_Greek_UPSILON, 0x1F59, 
GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1F69, 
GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F01, 
GDK_KEY_parenleft, GDK_KEY_Greek_epsilon, 0x1F11, 
GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F21, 
GDK_KEY_parenleft, GDK_KEY_Greek_iota, 0x1F31, 
GDK_KEY_parenleft, GDK_KEY_Greek_omicron, 0x1F41, 
GDK_KEY_parenleft, GDK_KEY_Greek_rho, 0x1FE5, 
GDK_KEY_parenleft, GDK_KEY_Greek_upsilon, 0x1F51, 
GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1F61, 
GDK_KEY_parenright, GDK_KEY_parenright, 0x005D, 
GDK_KEY_parenright, GDK_KEY_minus, 0x007D, 
GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F08, 
GDK_KEY_parenright, GDK_KEY_Greek_EPSILON, 0x1F18, 
GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F28, 
GDK_KEY_parenright, GDK_KEY_Greek_IOTA, 0x1F38, 
GDK_KEY_parenright, GDK_KEY_Greek_OMICRON, 0x1F48, 
GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1F68, 
GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F00, 
GDK_KEY_parenright, GDK_KEY_Greek_epsilon, 0x1F10, 
GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F20, 
GDK_KEY_parenright, GDK_KEY_Greek_iota, 0x1F30, 
GDK_KEY_parenright, GDK_KEY_Greek_omicron, 0x1F40, 
GDK_KEY_parenright, GDK_KEY_Greek_rho, 0x1FE4, 
GDK_KEY_parenright, GDK_KEY_Greek_upsilon, 0x1F50, 
GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1F60, 
GDK_KEY_asterisk, GDK_KEY_0, 0x00B0, 
GDK_KEY_asterisk, GDK_KEY_A, 0x00C5, 
GDK_KEY_asterisk, GDK_KEY_U, 0x016E, 
GDK_KEY_asterisk, GDK_KEY_a, 0x00E5, 
GDK_KEY_asterisk, GDK_KEY_u, 0x016F, 
GDK_KEY_asterisk, GDK_KEY_diaeresis, 0x2363, 
GDK_KEY_asterisk, GDK_KEY_emopencircle, 0x235F, 
GDK_KEY_plus, GDK_KEY_plus, 0x0023, 
GDK_KEY_plus, GDK_KEY_minus, 0x00B1, 
GDK_KEY_plus, GDK_KEY_O, 0x01A0, 
GDK_KEY_plus, GDK_KEY_U, 0x01AF, 
GDK_KEY_plus, GDK_KEY_o, 0x01A1, 
GDK_KEY_plus, GDK_KEY_u, 0x01B0, 
GDK_KEY_comma, GDK_KEY_space, 0x00B8, 
GDK_KEY_comma, GDK_KEY_quotedbl, 0x201E, 
GDK_KEY_comma, GDK_KEY_apostrophe, 0x201A, 
GDK_KEY_comma, GDK_KEY_comma, 0x00B8, 
GDK_KEY_comma, GDK_KEY_minus, 0x00AC, 
GDK_KEY_comma, GDK_KEY_A, 0x0104, 
GDK_KEY_comma, GDK_KEY_C, 0x00C7, 
GDK_KEY_comma, GDK_KEY_D, 0x1E10, 
GDK_KEY_comma, GDK_KEY_E, 0x0118, 
GDK_KEY_comma, GDK_KEY_G, 0x0122, 
GDK_KEY_comma, GDK_KEY_H, 0x1E28, 
GDK_KEY_comma, GDK_KEY_I, 0x012E, 
GDK_KEY_comma, GDK_KEY_K, 0x0136, 
GDK_KEY_comma, GDK_KEY_L, 0x013B, 
GDK_KEY_comma, GDK_KEY_N, 0x0145, 
GDK_KEY_comma, GDK_KEY_O, 0x01EA, 
GDK_KEY_comma, GDK_KEY_R, 0x0156, 
GDK_KEY_comma, GDK_KEY_S, 0x015E, 
GDK_KEY_comma, GDK_KEY_T, 0x0162, 
GDK_KEY_comma, GDK_KEY_U, 0x0172, 
GDK_KEY_comma, GDK_KEY_a, 0x0105, 
GDK_KEY_comma, GDK_KEY_c, 0x00E7, 
GDK_KEY_comma, GDK_KEY_d, 0x1E11, 
GDK_KEY_comma, GDK_KEY_e, 0x0119, 
GDK_KEY_comma, GDK_KEY_g, 0x0123, 
GDK_KEY_comma, GDK_KEY_h, 0x1E29, 
GDK_KEY_comma, GDK_KEY_i, 0x012F, 
GDK_KEY_comma, GDK_KEY_k, 0x0137, 
GDK_KEY_comma, GDK_KEY_l, 0x013C, 
GDK_KEY_comma, GDK_KEY_n, 0x0146, 
GDK_KEY_comma, GDK_KEY_o, 0x01EB, 
GDK_KEY_comma, GDK_KEY_r, 0x0157, 
GDK_KEY_comma, GDK_KEY_s, 0x015F, 
GDK_KEY_comma, GDK_KEY_t, 0x0163, 
GDK_KEY_comma, GDK_KEY_u, 0x0173, 
GDK_KEY_minus, GDK_KEY_space, 0x007E, 
GDK_KEY_minus, GDK_KEY_parenleft, 0x007B, 
GDK_KEY_minus, GDK_KEY_parenright, 0x007D, 
GDK_KEY_minus, GDK_KEY_plus, 0x00B1, 
GDK_KEY_minus, GDK_KEY_comma, 0x00AC, 
GDK_KEY_minus, GDK_KEY_slash, 0x233F, 
GDK_KEY_minus, GDK_KEY_colon, 0x00F7, 
GDK_KEY_minus, GDK_KEY_greater, 0x2192, 
GDK_KEY_minus, GDK_KEY_A, 0x0100, 
GDK_KEY_minus, GDK_KEY_D, 0x0110, 
GDK_KEY_minus, GDK_KEY_E, 0x0112, 
GDK_KEY_minus, GDK_KEY_I, 0x012A, 
GDK_KEY_minus, GDK_KEY_L, 0x00A3, 
GDK_KEY_minus, GDK_KEY_O, 0x014C, 
GDK_KEY_minus, GDK_KEY_U, 0x016A, 
GDK_KEY_minus, GDK_KEY_Y, 0x00A5, 
GDK_KEY_minus, GDK_KEY_backslash, 0x2340, 
GDK_KEY_minus, GDK_KEY_asciicircum, 0x00AF, 
GDK_KEY_minus, GDK_KEY_a, 0x0101, 
GDK_KEY_minus, GDK_KEY_d, 0x0111, 
GDK_KEY_minus, GDK_KEY_e, 0x0113, 
GDK_KEY_minus, GDK_KEY_i, 0x012B, 
GDK_KEY_minus, GDK_KEY_l, 0x00A3, 
GDK_KEY_minus, GDK_KEY_o, 0x014D, 
GDK_KEY_minus, GDK_KEY_u, 0x016B, 
GDK_KEY_minus, GDK_KEY_y, 0x00A5, 
GDK_KEY_minus, 0x2191, 0x234F, 
GDK_KEY_minus, 0x2193, 0x2356, 
GDK_KEY_minus, GDK_KEY_emopencircle, 0x2296, 
GDK_KEY_period, GDK_KEY_minus, 0x00B7, 
GDK_KEY_period, GDK_KEY_period, 0x2026, 
GDK_KEY_period, GDK_KEY_colon, 0x2235, 
GDK_KEY_period, GDK_KEY_less, 0x2039, 
GDK_KEY_period, GDK_KEY_equal, 0x2022, 
GDK_KEY_period, GDK_KEY_greater, 0x203A, 
GDK_KEY_period, GDK_KEY_A, 0x0226, 
GDK_KEY_period, GDK_KEY_B, 0x1E02, 
GDK_KEY_period, GDK_KEY_C, 0x010A, 
GDK_KEY_period, GDK_KEY_D, 0x1E0A, 
GDK_KEY_period, GDK_KEY_E, 0x0116, 
GDK_KEY_period, GDK_KEY_F, 0x1E1E, 
GDK_KEY_period, GDK_KEY_G, 0x0120, 
GDK_KEY_period, GDK_KEY_H, 0x1E22, 
GDK_KEY_period, GDK_KEY_I, 0x0130, 
GDK_KEY_period, GDK_KEY_M, 0x1E40, 
GDK_KEY_period, GDK_KEY_N, 0x1E44, 
GDK_KEY_period, GDK_KEY_O, 0x022E, 
GDK_KEY_period, GDK_KEY_P, 0x1E56, 
GDK_KEY_period, GDK_KEY_R, 0x1E58, 
GDK_KEY_period, GDK_KEY_S, 0x1E60, 
GDK_KEY_period, GDK_KEY_T, 0x1E6A, 
GDK_KEY_period, GDK_KEY_W, 0x1E86, 
GDK_KEY_period, GDK_KEY_X, 0x1E8A, 
GDK_KEY_period, GDK_KEY_Y, 0x1E8E, 
GDK_KEY_period, GDK_KEY_Z, 0x017B, 
GDK_KEY_period, GDK_KEY_asciicircum, 0x00B7, 
GDK_KEY_period, GDK_KEY_a, 0x0227, 
GDK_KEY_period, GDK_KEY_b, 0x1E03, 
GDK_KEY_period, GDK_KEY_c, 0x010B, 
GDK_KEY_period, GDK_KEY_d, 0x1E0B, 
GDK_KEY_period, GDK_KEY_e, 0x0117, 
GDK_KEY_period, GDK_KEY_f, 0x1E1F, 
GDK_KEY_period, GDK_KEY_g, 0x0121, 
GDK_KEY_period, GDK_KEY_h, 0x1E23, 
GDK_KEY_period, GDK_KEY_i, 0x0131, 
GDK_KEY_period, GDK_KEY_m, 0x1E41, 
GDK_KEY_period, GDK_KEY_n, 0x1E45, 
GDK_KEY_period, GDK_KEY_o, 0x022F, 
GDK_KEY_period, GDK_KEY_p, 0x1E57, 
GDK_KEY_period, GDK_KEY_r, 0x1E59, 
GDK_KEY_period, GDK_KEY_s, 0x1E61, 
GDK_KEY_period, GDK_KEY_t, 0x1E6B, 
GDK_KEY_period, GDK_KEY_w, 0x1E87, 
GDK_KEY_period, GDK_KEY_x, 0x1E8B, 
GDK_KEY_period, GDK_KEY_y, 0x1E8F, 
GDK_KEY_period, GDK_KEY_z, 0x017C, 
GDK_KEY_period, 0x017F, 0x1E9B, 
GDK_KEY_period, GDK_KEY_Sacute, 0x1E64, 
GDK_KEY_period, GDK_KEY_Scaron, 0x1E66, 
GDK_KEY_period, GDK_KEY_sacute, 0x1E65, 
GDK_KEY_period, GDK_KEY_scaron, 0x1E67, 
GDK_KEY_period, 0x1E62, 0x1E68, 
GDK_KEY_period, 0x1E63, 0x1E69, 
GDK_KEY_period, GDK_KEY_emopencircle, 0x2299, 
GDK_KEY_slash, GDK_KEY_minus, 0x233F, 
GDK_KEY_slash, GDK_KEY_slash, 0x005C, 
GDK_KEY_slash, GDK_KEY_less, 0x005C, 
GDK_KEY_slash, GDK_KEY_equal, 0x2260, 
GDK_KEY_slash, GDK_KEY_C, 0x20A1, 
GDK_KEY_slash, GDK_KEY_D, 0x0110, 
GDK_KEY_slash, GDK_KEY_G, 0x01E4, 
GDK_KEY_slash, GDK_KEY_H, 0x0126, 
GDK_KEY_slash, GDK_KEY_I, 0x0197, 
GDK_KEY_slash, GDK_KEY_L, 0x0141, 
GDK_KEY_slash, GDK_KEY_O, 0x00D8, 
GDK_KEY_slash, GDK_KEY_T, 0x0166, 
GDK_KEY_slash, GDK_KEY_Z, 0x01B5, 
GDK_KEY_slash, GDK_KEY_asciicircum, 0x007C, 
GDK_KEY_slash, GDK_KEY_b, 0x0180, 
GDK_KEY_slash, GDK_KEY_c, 0x00A2, 
GDK_KEY_slash, GDK_KEY_d, 0x0111, 
GDK_KEY_slash, GDK_KEY_g, 0x01E5, 
GDK_KEY_slash, GDK_KEY_h, 0x0127, 
GDK_KEY_slash, GDK_KEY_i, 0x0268, 
GDK_KEY_slash, GDK_KEY_l, 0x0142, 
GDK_KEY_slash, GDK_KEY_m, 0x20A5, 
GDK_KEY_slash, GDK_KEY_o, 0x00F8, 
GDK_KEY_slash, GDK_KEY_t, 0x0167, 
GDK_KEY_slash, GDK_KEY_u, 0x00B5, 
GDK_KEY_slash, GDK_KEY_v, 0x221A, 
GDK_KEY_slash, GDK_KEY_z, 0x01B6, 
GDK_KEY_slash, 0x0294, 0x02A1, 
GDK_KEY_slash, 0x04AE, 0x04B0, 
GDK_KEY_slash, 0x04AF, 0x04B1, 
GDK_KEY_slash, GDK_KEY_Cyrillic_ghe, 0x0493, 
GDK_KEY_slash, GDK_KEY_Cyrillic_ka, 0x049F, 
GDK_KEY_slash, GDK_KEY_Cyrillic_GHE, 0x0492, 
GDK_KEY_slash, GDK_KEY_Cyrillic_KA, 0x049E, 
GDK_KEY_slash, GDK_KEY_leftarrow, 0x219A, 
GDK_KEY_slash, GDK_KEY_rightarrow, 0x219B, 
GDK_KEY_slash, 0x2194, 0x21AE, 
GDK_KEY_slash, 0x2395, 0x2341, 
GDK_KEY_0, GDK_KEY_asterisk, 0x00B0, 
GDK_KEY_0, GDK_KEY_3, 0x2189, 
GDK_KEY_0, GDK_KEY_asciitilde, 0x236C, 
GDK_KEY_1, GDK_KEY_2, 0x00BD, 
GDK_KEY_1, GDK_KEY_3, 0x2153, 
GDK_KEY_1, GDK_KEY_4, 0x00BC, 
GDK_KEY_1, GDK_KEY_5, 0x2155, 
GDK_KEY_1, GDK_KEY_6, 0x2159, 
GDK_KEY_1, GDK_KEY_7, 0x2150, 
GDK_KEY_1, GDK_KEY_8, 0x215B, 
GDK_KEY_1, GDK_KEY_9, 0x2151, 
GDK_KEY_1, GDK_KEY_asciicircum, 0x00B9, 
GDK_KEY_2, GDK_KEY_3, 0x2154, 
GDK_KEY_2, GDK_KEY_5, 0x2156, 
GDK_KEY_2, GDK_KEY_asciicircum, 0x00B2, 
GDK_KEY_3, GDK_KEY_4, 0x00BE, 
GDK_KEY_3, GDK_KEY_5, 0x2157, 
GDK_KEY_3, GDK_KEY_8, 0x215C, 
GDK_KEY_3, GDK_KEY_asciicircum, 0x00B3, 
GDK_KEY_4, GDK_KEY_5, 0x2158, 
GDK_KEY_5, GDK_KEY_6, 0x215A, 
GDK_KEY_5, GDK_KEY_8, 0x215D, 
GDK_KEY_7, GDK_KEY_8, 0x215E, 
GDK_KEY_8, GDK_KEY_8, 0x221E, 
GDK_KEY_colon, GDK_KEY_parenleft, 0x2639, 
GDK_KEY_colon, GDK_KEY_parenright, 0x263A, 
GDK_KEY_colon, GDK_KEY_minus, 0x00F7, 
GDK_KEY_colon, GDK_KEY_period, 0x2234, 
GDK_KEY_colon, 0x2395, 0x2360, 
GDK_KEY_semicolon, GDK_KEY_A, 0x0104, 
GDK_KEY_semicolon, GDK_KEY_E, 0x0118, 
GDK_KEY_semicolon, GDK_KEY_I, 0x012E, 
GDK_KEY_semicolon, GDK_KEY_O, 0x01EA, 
GDK_KEY_semicolon, GDK_KEY_S, 0x0218, 
GDK_KEY_semicolon, GDK_KEY_T, 0x021A, 
GDK_KEY_semicolon, GDK_KEY_U, 0x0172, 
GDK_KEY_semicolon, GDK_KEY_underscore, 0x236E, 
GDK_KEY_semicolon, GDK_KEY_a, 0x0105, 
GDK_KEY_semicolon, GDK_KEY_e, 0x0119, 
GDK_KEY_semicolon, GDK_KEY_i, 0x012F, 
GDK_KEY_semicolon, GDK_KEY_o, 0x01EB, 
GDK_KEY_semicolon, GDK_KEY_s, 0x0219, 
GDK_KEY_semicolon, GDK_KEY_t, 0x021B, 
GDK_KEY_semicolon, GDK_KEY_u, 0x0173, 
GDK_KEY_less, GDK_KEY_space, 0x02C7, 
GDK_KEY_less, GDK_KEY_quotedbl, 0x201C, 
GDK_KEY_less, GDK_KEY_apostrophe, 0x2018, 
GDK_KEY_less, GDK_KEY_minus, 0x2190, 
GDK_KEY_less, GDK_KEY_slash, 0x005C, 
GDK_KEY_less, GDK_KEY_3, 0x2665, 
GDK_KEY_less, GDK_KEY_less, 0x00AB, 
GDK_KEY_less, GDK_KEY_equal, 0x2264, 
GDK_KEY_less, GDK_KEY_greater, 0x22C4, 
GDK_KEY_less, GDK_KEY_C, 0x010C, 
GDK_KEY_less, GDK_KEY_D, 0x010E, 
GDK_KEY_less, GDK_KEY_E, 0x011A, 
GDK_KEY_less, GDK_KEY_L, 0x013D, 
GDK_KEY_less, GDK_KEY_N, 0x0147, 
GDK_KEY_less, GDK_KEY_R, 0x0158, 
GDK_KEY_less, GDK_KEY_S, 0x0160, 
GDK_KEY_less, GDK_KEY_T, 0x0164, 
GDK_KEY_less, GDK_KEY_Z, 0x017D, 
GDK_KEY_less, GDK_KEY_underscore, 0x2264, 
GDK_KEY_less, GDK_KEY_c, 0x010D, 
GDK_KEY_less, GDK_KEY_d, 0x010F, 
GDK_KEY_less, GDK_KEY_e, 0x011B, 
GDK_KEY_less, GDK_KEY_l, 0x013E, 
GDK_KEY_less, GDK_KEY_n, 0x0148, 
GDK_KEY_less, GDK_KEY_r, 0x0159, 
GDK_KEY_less, GDK_KEY_s, 0x0161, 
GDK_KEY_less, GDK_KEY_t, 0x0165, 
GDK_KEY_less, GDK_KEY_z, 0x017E, 
GDK_KEY_less, 0x0338, 0x226E, 
GDK_KEY_less, 0x2395, 0x2343, 
GDK_KEY_equal, GDK_KEY_slash, 0x2260, 
GDK_KEY_equal, GDK_KEY_greater, 0x21D2, 
GDK_KEY_equal, GDK_KEY_C, 0x20AC, 
GDK_KEY_equal, GDK_KEY_E, 0x20AC, 
GDK_KEY_equal, GDK_KEY_L, 0x20A4, 
GDK_KEY_equal, GDK_KEY_N, 0x20A6, 
GDK_KEY_equal, GDK_KEY_O, 0x0150, 
GDK_KEY_equal, GDK_KEY_U, 0x0170, 
GDK_KEY_equal, GDK_KEY_W, 0x20A9, 
GDK_KEY_equal, GDK_KEY_Y, 0x00A5, 
GDK_KEY_equal, GDK_KEY_underscore, 0x2261, 
GDK_KEY_equal, GDK_KEY_c, 0x20AC, 
GDK_KEY_equal, GDK_KEY_d, 0x20AB, 
GDK_KEY_equal, GDK_KEY_e, 0x20AC, 
GDK_KEY_equal, GDK_KEY_o, 0x0151, 
GDK_KEY_equal, GDK_KEY_u, 0x0171, 
GDK_KEY_equal, GDK_KEY_y, 0x00A5, 
GDK_KEY_equal, 0x0338, 0x2260, 
GDK_KEY_equal, GDK_KEY_Cyrillic_u, 0x04F3, 
GDK_KEY_equal, GDK_KEY_Cyrillic_IE, 0x20AC, 
GDK_KEY_equal, GDK_KEY_Cyrillic_ES, 0x20AC, 
GDK_KEY_equal, GDK_KEY_Cyrillic_U, 0x04F2, 
GDK_KEY_equal, 0x2395, 0x2338, 
GDK_KEY_greater, GDK_KEY_space, 0x005E, 
GDK_KEY_greater, GDK_KEY_quotedbl, 0x201D, 
GDK_KEY_greater, GDK_KEY_apostrophe, 0x2019, 
GDK_KEY_greater, GDK_KEY_less, 0x22C4, 
GDK_KEY_greater, GDK_KEY_equal, 0x2265, 
GDK_KEY_greater, GDK_KEY_greater, 0x00BB, 
GDK_KEY_greater, GDK_KEY_A, 0x00C2, 
GDK_KEY_greater, GDK_KEY_E, 0x00CA, 
GDK_KEY_greater, GDK_KEY_I, 0x00CE, 
GDK_KEY_greater, GDK_KEY_O, 0x00D4, 
GDK_KEY_greater, GDK_KEY_U, 0x00DB, 
GDK_KEY_greater, GDK_KEY_underscore, 0x2265, 
GDK_KEY_greater, GDK_KEY_a, 0x00E2, 
GDK_KEY_greater, GDK_KEY_e, 0x00EA, 
GDK_KEY_greater, GDK_KEY_i, 0x00EE, 
GDK_KEY_greater, GDK_KEY_o, 0x00F4, 
GDK_KEY_greater, GDK_KEY_u, 0x00FB, 
GDK_KEY_greater, GDK_KEY_diaeresis, 0x2369, 
GDK_KEY_greater, 0x0338, 0x226F, 
GDK_KEY_greater, 0x2395, 0x2344, 
GDK_KEY_question, GDK_KEY_exclam, 0x2E18, 
GDK_KEY_question, GDK_KEY_question, 0x00BF, 
GDK_KEY_question, GDK_KEY_A, 0x1EA2, 
GDK_KEY_question, GDK_KEY_E, 0x1EBA, 
GDK_KEY_question, GDK_KEY_I, 0x1EC8, 
GDK_KEY_question, GDK_KEY_O, 0x1ECE, 
GDK_KEY_question, GDK_KEY_U, 0x1EE6, 
GDK_KEY_question, GDK_KEY_Y, 0x1EF6, 
GDK_KEY_question, GDK_KEY_a, 0x1EA3, 
GDK_KEY_question, GDK_KEY_e, 0x1EBB, 
GDK_KEY_question, GDK_KEY_i, 0x1EC9, 
GDK_KEY_question, GDK_KEY_o, 0x1ECF, 
GDK_KEY_question, GDK_KEY_u, 0x1EE7, 
GDK_KEY_question, GDK_KEY_y, 0x1EF7, 
GDK_KEY_question, GDK_KEY_Acircumflex, 0x1EA8, 
GDK_KEY_question, GDK_KEY_Ecircumflex, 0x1EC2, 
GDK_KEY_question, GDK_KEY_Ocircumflex, 0x1ED4, 
GDK_KEY_question, GDK_KEY_acircumflex, 0x1EA9, 
GDK_KEY_question, GDK_KEY_ecircumflex, 0x1EC3, 
GDK_KEY_question, GDK_KEY_ocircumflex, 0x1ED5, 
GDK_KEY_question, GDK_KEY_Abreve, 0x1EB2, 
GDK_KEY_question, GDK_KEY_abreve, 0x1EB3, 
GDK_KEY_question, 0x2395, 0x2370, 
GDK_KEY_A, GDK_KEY_quotedbl, 0x00C4, 
GDK_KEY_A, GDK_KEY_apostrophe, 0x00C1, 
GDK_KEY_A, GDK_KEY_parenleft, 0x0102, 
GDK_KEY_A, GDK_KEY_asterisk, 0x00C5, 
GDK_KEY_A, GDK_KEY_comma, 0x0104, 
GDK_KEY_A, GDK_KEY_minus, 0x0100, 
GDK_KEY_A, GDK_KEY_semicolon, 0x0104, 
GDK_KEY_A, GDK_KEY_greater, 0x00C2, 
GDK_KEY_A, GDK_KEY_A, 0x00C5, 
GDK_KEY_A, GDK_KEY_E, 0x00C6, 
GDK_KEY_A, GDK_KEY_T, 0x0040, 
GDK_KEY_A, GDK_KEY_asciicircum, 0x00C2, 
GDK_KEY_A, GDK_KEY_underscore, 0x0100, 
GDK_KEY_A, GDK_KEY_grave, 0x00C0, 
GDK_KEY_A, GDK_KEY_asciitilde, 0x00C3, 
GDK_KEY_A, GDK_KEY_diaeresis, 0x00C4, 
GDK_KEY_A, GDK_KEY_acute, 0x00C1, 
GDK_KEY_B, GDK_KEY_period, 0x1E02, 
GDK_KEY_C, GDK_KEY_apostrophe, 0x0106, 
GDK_KEY_C, GDK_KEY_comma, 0x00C7, 
GDK_KEY_C, GDK_KEY_period, 0x010A, 
GDK_KEY_C, GDK_KEY_slash, 0x20A1, 
GDK_KEY_C, GDK_KEY_less, 0x010C, 
GDK_KEY_C, GDK_KEY_equal, 0x20AC, 
GDK_KEY_C, GDK_KEY_E, 0x20A0, 
GDK_KEY_C, GDK_KEY_O, 0x00A9, 
GDK_KEY_C, GDK_KEY_o, 0x00A9, 
GDK_KEY_C, GDK_KEY_r, 0x20A2, 
GDK_KEY_C, GDK_KEY_bar, 0x00A2, 
GDK_KEY_D, GDK_KEY_comma, 0x1E10, 
GDK_KEY_D, GDK_KEY_minus, 0x0110, 
GDK_KEY_D, GDK_KEY_period, 0x1E0A, 
GDK_KEY_D, GDK_KEY_less, 0x010E, 
GDK_KEY_D, GDK_KEY_H, 0x00D0, 
GDK_KEY_E, GDK_KEY_quotedbl, 0x00CB, 
GDK_KEY_E, GDK_KEY_apostrophe, 0x00C9, 
GDK_KEY_E, GDK_KEY_comma, 0x0118, 
GDK_KEY_E, GDK_KEY_minus, 0x0112, 
GDK_KEY_E, GDK_KEY_period, 0x0116, 
GDK_KEY_E, GDK_KEY_semicolon, 0x0118, 
GDK_KEY_E, GDK_KEY_less, 0x011A, 
GDK_KEY_E, GDK_KEY_equal, 0x20AC, 
GDK_KEY_E, GDK_KEY_greater, 0x00CA, 
GDK_KEY_E, GDK_KEY_asciicircum, 0x00CA, 
GDK_KEY_E, GDK_KEY_underscore, 0x0112, 
GDK_KEY_E, GDK_KEY_grave, 0x00C8, 
GDK_KEY_E, GDK_KEY_diaeresis, 0x00CB, 
GDK_KEY_E, GDK_KEY_acute, 0x00C9, 
GDK_KEY_F, GDK_KEY_period, 0x1E1E, 
GDK_KEY_F, GDK_KEY_i, 0xFB03, 
GDK_KEY_F, GDK_KEY_l, 0xFB04, 
GDK_KEY_F, GDK_KEY_r, 0x20A3, 
GDK_KEY_G, GDK_KEY_parenleft, 0x011E, 
GDK_KEY_G, GDK_KEY_comma, 0x0122, 
GDK_KEY_G, GDK_KEY_period, 0x0120, 
GDK_KEY_G, GDK_KEY_U, 0x011E, 
GDK_KEY_G, GDK_KEY_breve, 0x011E, 
GDK_KEY_H, GDK_KEY_comma, 0x1E28, 
GDK_KEY_I, GDK_KEY_quotedbl, 0x00CF, 
GDK_KEY_I, GDK_KEY_apostrophe, 0x00CD, 
GDK_KEY_I, GDK_KEY_comma, 0x012E, 
GDK_KEY_I, GDK_KEY_minus, 0x012A, 
GDK_KEY_I, GDK_KEY_period, 0x0130, 
GDK_KEY_I, GDK_KEY_semicolon, 0x012E, 
GDK_KEY_I, GDK_KEY_greater, 0x00CE, 
GDK_KEY_I, GDK_KEY_J, 0x0132, 
GDK_KEY_I, GDK_KEY_asciicircum, 0x00CE, 
GDK_KEY_I, GDK_KEY_underscore, 0x012A, 
GDK_KEY_I, GDK_KEY_grave, 0x00CC, 
GDK_KEY_I, GDK_KEY_j, 0x0132, 
GDK_KEY_I, GDK_KEY_asciitilde, 0x0128, 
GDK_KEY_I, GDK_KEY_diaeresis, 0x00CF, 
GDK_KEY_I, GDK_KEY_acute, 0x00CD, 
GDK_KEY_K, GDK_KEY_comma, 0x0136, 
GDK_KEY_L, GDK_KEY_apostrophe, 0x0139, 
GDK_KEY_L, GDK_KEY_comma, 0x013B, 
GDK_KEY_L, GDK_KEY_minus, 0x00A3, 
GDK_KEY_L, GDK_KEY_slash, 0x0141, 
GDK_KEY_L, GDK_KEY_less, 0x013D, 
GDK_KEY_L, GDK_KEY_equal, 0x20A4, 
GDK_KEY_L, GDK_KEY_V, 0x007C, 
GDK_KEY_M, GDK_KEY_period, 0x1E40, 
GDK_KEY_N, GDK_KEY_apostrophe, 0x0143, 
GDK_KEY_N, GDK_KEY_comma, 0x0145, 
GDK_KEY_N, GDK_KEY_less, 0x0147, 
GDK_KEY_N, GDK_KEY_equal, 0x20A6, 
GDK_KEY_N, GDK_KEY_G, 0x014A, 
GDK_KEY_N, GDK_KEY_O, 0x2116, 
GDK_KEY_N, GDK_KEY_o, 0x2116, 
GDK_KEY_N, GDK_KEY_asciitilde, 0x00D1, 
GDK_KEY_O, GDK_KEY_quotedbl, 0x00D6, 
GDK_KEY_O, GDK_KEY_apostrophe, 0x00D3, 
GDK_KEY_O, GDK_KEY_comma, 0x01EA, 
GDK_KEY_O, GDK_KEY_minus, 0x014C, 
GDK_KEY_O, GDK_KEY_slash, 0x00D8, 
GDK_KEY_O, GDK_KEY_semicolon, 0x01EA, 
GDK_KEY_O, GDK_KEY_greater, 0x00D4, 
GDK_KEY_O, GDK_KEY_A, 0x24B6, 
GDK_KEY_O, GDK_KEY_C, 0x00A9, 
GDK_KEY_O, GDK_KEY_E, 0x0152, 
GDK_KEY_O, GDK_KEY_R, 0x00AE, 
GDK_KEY_O, GDK_KEY_S, 0x00A7, 
GDK_KEY_O, GDK_KEY_X, 0x00A4, 
GDK_KEY_O, GDK_KEY_asciicircum, 0x00D4, 
GDK_KEY_O, GDK_KEY_underscore, 0x014C, 
GDK_KEY_O, GDK_KEY_grave, 0x00D2, 
GDK_KEY_O, GDK_KEY_c, 0x00A9, 
GDK_KEY_O, GDK_KEY_r, 0x00AE, 
GDK_KEY_O, GDK_KEY_x, 0x00A4, 
GDK_KEY_O, GDK_KEY_asciitilde, 0x00D5, 
GDK_KEY_O, GDK_KEY_diaeresis, 0x00D6, 
GDK_KEY_O, GDK_KEY_acute, 0x00D3, 
GDK_KEY_P, GDK_KEY_exclam, 0x00B6, 
GDK_KEY_P, GDK_KEY_period, 0x1E56, 
GDK_KEY_P, GDK_KEY_P, 0x00B6, 
GDK_KEY_P, GDK_KEY_t, 0x20A7, 
GDK_KEY_R, GDK_KEY_apostrophe, 0x0154, 
GDK_KEY_R, GDK_KEY_comma, 0x0156, 
GDK_KEY_R, GDK_KEY_less, 0x0158, 
GDK_KEY_R, GDK_KEY_O, 0x00AE, 
GDK_KEY_R, GDK_KEY_s, 0x20A8, 
GDK_KEY_S, GDK_KEY_exclam, 0x00A7, 
GDK_KEY_S, GDK_KEY_apostrophe, 0x015A, 
GDK_KEY_S, GDK_KEY_comma, 0x015E, 
GDK_KEY_S, GDK_KEY_period, 0x1E60, 
GDK_KEY_S, GDK_KEY_semicolon, 0x0218, 
GDK_KEY_S, GDK_KEY_less, 0x0160, 
GDK_KEY_S, GDK_KEY_M, 0x2120, 
GDK_KEY_S, GDK_KEY_O, 0x00A7, 
GDK_KEY_S, GDK_KEY_S, 0x1E9E, 
GDK_KEY_S, GDK_KEY_m, 0x2120, 
GDK_KEY_T, GDK_KEY_comma, 0x0162, 
GDK_KEY_T, GDK_KEY_minus, 0x0166, 
GDK_KEY_T, GDK_KEY_period, 0x1E6A, 
GDK_KEY_T, GDK_KEY_slash, 0x0166, 
GDK_KEY_T, GDK_KEY_semicolon, 0x021A, 
GDK_KEY_T, GDK_KEY_less, 0x0164, 
GDK_KEY_T, GDK_KEY_H, 0x00DE, 
GDK_KEY_T, GDK_KEY_M, 0x2122, 
GDK_KEY_T, GDK_KEY_m, 0x2122, 
GDK_KEY_U, GDK_KEY_quotedbl, 0x00DC, 
GDK_KEY_U, GDK_KEY_apostrophe, 0x00DA, 
GDK_KEY_U, GDK_KEY_asterisk, 0x016E, 
GDK_KEY_U, GDK_KEY_comma, 0x0172, 
GDK_KEY_U, GDK_KEY_minus, 0x016A, 
GDK_KEY_U, GDK_KEY_semicolon, 0x0172, 
GDK_KEY_U, GDK_KEY_greater, 0x00DB, 
GDK_KEY_U, GDK_KEY_A, 0x0102, 
GDK_KEY_U, GDK_KEY_E, 0x0114, 
GDK_KEY_U, GDK_KEY_G, 0x011E, 
GDK_KEY_U, GDK_KEY_I, 0x012C, 
GDK_KEY_U, GDK_KEY_O, 0x014E, 
GDK_KEY_U, GDK_KEY_U, 0x016C, 
GDK_KEY_U, GDK_KEY_asciicircum, 0x00DB, 
GDK_KEY_U, GDK_KEY_underscore, 0x016A, 
GDK_KEY_U, GDK_KEY_grave, 0x00D9, 
GDK_KEY_U, GDK_KEY_a, 0x0103, 
GDK_KEY_U, GDK_KEY_e, 0x0115, 
GDK_KEY_U, GDK_KEY_g, 0x011F, 
GDK_KEY_U, GDK_KEY_i, 0x012D, 
GDK_KEY_U, GDK_KEY_o, 0x014F, 
GDK_KEY_U, GDK_KEY_u, 0x016D, 
GDK_KEY_U, GDK_KEY_asciitilde, 0x0168, 
GDK_KEY_U, GDK_KEY_diaeresis, 0x00DC, 
GDK_KEY_U, GDK_KEY_acute, 0x00DA, 
GDK_KEY_U, 0x0228, 0x1E1C, 
GDK_KEY_U, 0x0229, 0x1E1D, 
GDK_KEY_U, GDK_KEY_Cyrillic_a, 0x04D1, 
GDK_KEY_U, GDK_KEY_Cyrillic_ie, 0x04D7, 
GDK_KEY_U, GDK_KEY_Cyrillic_i, 0x0439, 
GDK_KEY_U, GDK_KEY_Cyrillic_u, 0x045E, 
GDK_KEY_U, GDK_KEY_Cyrillic_zhe, 0x04C2, 
GDK_KEY_U, GDK_KEY_Cyrillic_A, 0x04D0, 
GDK_KEY_U, GDK_KEY_Cyrillic_IE, 0x04D6, 
GDK_KEY_U, GDK_KEY_Cyrillic_I, 0x0419, 
GDK_KEY_U, GDK_KEY_Cyrillic_U, 0x040E, 
GDK_KEY_U, GDK_KEY_Cyrillic_ZHE, 0x04C1, 
GDK_KEY_U, GDK_KEY_Greek_ALPHA, 0x1FB8, 
GDK_KEY_U, GDK_KEY_Greek_IOTA, 0x1FD8, 
GDK_KEY_U, GDK_KEY_Greek_UPSILON, 0x1FE8, 
GDK_KEY_U, GDK_KEY_Greek_alpha, 0x1FB0, 
GDK_KEY_U, GDK_KEY_Greek_iota, 0x1FD0, 
GDK_KEY_U, GDK_KEY_Greek_upsilon, 0x1FE0, 
GDK_KEY_U, 0x1EA0, 0x1EB6, 
GDK_KEY_U, 0x1EA1, 0x1EB7, 
GDK_KEY_V, GDK_KEY_L, 0x007C, 
GDK_KEY_W, GDK_KEY_equal, 0x20A9, 
GDK_KEY_W, GDK_KEY_asciicircum, 0x0174, 
GDK_KEY_X, GDK_KEY_O, 0x00A4, 
GDK_KEY_X, GDK_KEY_o, 0x00A4, 
GDK_KEY_Y, GDK_KEY_quotedbl, 0x0178, 
GDK_KEY_Y, GDK_KEY_apostrophe, 0x00DD, 
GDK_KEY_Y, GDK_KEY_minus, 0x00A5, 
GDK_KEY_Y, GDK_KEY_equal, 0x00A5, 
GDK_KEY_Y, GDK_KEY_asciicircum, 0x0176, 
GDK_KEY_Y, GDK_KEY_diaeresis, 0x0178, 
GDK_KEY_Y, GDK_KEY_acute, 0x00DD, 
GDK_KEY_Z, GDK_KEY_apostrophe, 0x0179, 
GDK_KEY_Z, GDK_KEY_period, 0x017B, 
GDK_KEY_Z, GDK_KEY_less, 0x017D, 
GDK_KEY_bracketleft, GDK_KEY_bracketright, 0x2337, 
GDK_KEY_backslash, GDK_KEY_minus, 0x2340, 
GDK_KEY_backslash, 0x2395, 0x2342, 
GDK_KEY_backslash, GDK_KEY_emopencircle, 0x2349, 
GDK_KEY_bracketright, GDK_KEY_bracketleft, 0x2337, 
GDK_KEY_asciicircum, GDK_KEY_space, 0x005E, 
GDK_KEY_asciicircum, GDK_KEY_parenleft, 0x207D, 
GDK_KEY_asciicircum, GDK_KEY_parenright, 0x207E, 
GDK_KEY_asciicircum, GDK_KEY_plus, 0x207A, 
GDK_KEY_asciicircum, GDK_KEY_minus, 0x00AF, 
GDK_KEY_asciicircum, GDK_KEY_period, 0x00B7, 
GDK_KEY_asciicircum, GDK_KEY_slash, 0x007C, 
GDK_KEY_asciicircum, GDK_KEY_0, 0x2070, 
GDK_KEY_asciicircum, GDK_KEY_1, 0x00B9, 
GDK_KEY_asciicircum, GDK_KEY_2, 0x00B2, 
GDK_KEY_asciicircum, GDK_KEY_3, 0x00B3, 
GDK_KEY_asciicircum, GDK_KEY_4, 0x2074, 
GDK_KEY_asciicircum, GDK_KEY_5, 0x2075, 
GDK_KEY_asciicircum, GDK_KEY_6, 0x2076, 
GDK_KEY_asciicircum, GDK_KEY_7, 0x2077, 
GDK_KEY_asciicircum, GDK_KEY_8, 0x2078, 
GDK_KEY_asciicircum, GDK_KEY_9, 0x2079, 
GDK_KEY_asciicircum, GDK_KEY_equal, 0x207C, 
GDK_KEY_asciicircum, GDK_KEY_A, 0x00C2, 
GDK_KEY_asciicircum, GDK_KEY_C, 0x0108, 
GDK_KEY_asciicircum, GDK_KEY_E, 0x00CA, 
GDK_KEY_asciicircum, GDK_KEY_G, 0x011C, 
GDK_KEY_asciicircum, GDK_KEY_H, 0x0124, 
GDK_KEY_asciicircum, GDK_KEY_I, 0x00CE, 
GDK_KEY_asciicircum, GDK_KEY_J, 0x0134, 
GDK_KEY_asciicircum, GDK_KEY_O, 0x00D4, 
GDK_KEY_asciicircum, GDK_KEY_S, 0x015C, 
GDK_KEY_asciicircum, GDK_KEY_U, 0x00DB, 
GDK_KEY_asciicircum, GDK_KEY_W, 0x0174, 
GDK_KEY_asciicircum, GDK_KEY_Y, 0x0176, 
GDK_KEY_asciicircum, GDK_KEY_Z, 0x1E90, 
GDK_KEY_asciicircum, GDK_KEY_a, 0x00E2, 
GDK_KEY_asciicircum, GDK_KEY_c, 0x0109, 
GDK_KEY_asciicircum, GDK_KEY_e, 0x00EA, 
GDK_KEY_asciicircum, GDK_KEY_g, 0x011D, 
GDK_KEY_asciicircum, GDK_KEY_h, 0x0125, 
GDK_KEY_asciicircum, GDK_KEY_i, 0x00EE, 
GDK_KEY_asciicircum, GDK_KEY_j, 0x0135, 
GDK_KEY_asciicircum, GDK_KEY_o, 0x00F4, 
GDK_KEY_asciicircum, GDK_KEY_s, 0x015D, 
GDK_KEY_asciicircum, GDK_KEY_u, 0x00FB, 
GDK_KEY_asciicircum, GDK_KEY_w, 0x0175, 
GDK_KEY_asciicircum, GDK_KEY_y, 0x0177, 
GDK_KEY_asciicircum, GDK_KEY_z, 0x1E91, 
GDK_KEY_asciicircum, 0x1EA0, 0x1EAC, 
GDK_KEY_asciicircum, 0x1EA1, 0x1EAD, 
GDK_KEY_asciicircum, 0x1EB8, 0x1EC6, 
GDK_KEY_asciicircum, 0x1EB9, 0x1EC7, 
GDK_KEY_asciicircum, 0x1ECC, 0x1ED8, 
GDK_KEY_asciicircum, 0x1ECD, 0x1ED9, 
GDK_KEY_asciicircum, 0x2212, 0x207B, 
GDK_KEY_asciicircum, 0x4E00, 0x3192, 
GDK_KEY_asciicircum, 0x4E01, 0x319C, 
GDK_KEY_asciicircum, 0x4E09, 0x3194, 
GDK_KEY_asciicircum, 0x4E0A, 0x3196, 
GDK_KEY_asciicircum, 0x4E0B, 0x3198, 
GDK_KEY_asciicircum, 0x4E19, 0x319B, 
GDK_KEY_asciicircum, 0x4E2D, 0x3197, 
GDK_KEY_asciicircum, 0x4E59, 0x319A, 
GDK_KEY_asciicircum, 0x4E8C, 0x3193, 
GDK_KEY_asciicircum, 0x4EBA, 0x319F, 
GDK_KEY_asciicircum, 0x56DB, 0x3195, 
GDK_KEY_asciicircum, 0x5730, 0x319E, 
GDK_KEY_asciicircum, 0x5929, 0x319D, 
GDK_KEY_asciicircum, 0x7532, 0x3199, 
GDK_KEY_asciicircum, GDK_KEY_KP_Space, 0x00B2, 
GDK_KEY_asciicircum, GDK_KEY_KP_Add, 0x207A, 
GDK_KEY_asciicircum, GDK_KEY_KP_0, 0x2070, 
GDK_KEY_asciicircum, GDK_KEY_KP_1, 0x00B9, 
GDK_KEY_asciicircum, GDK_KEY_KP_2, 0x00B2, 
GDK_KEY_asciicircum, GDK_KEY_KP_3, 0x00B3, 
GDK_KEY_asciicircum, GDK_KEY_KP_4, 0x2074, 
GDK_KEY_asciicircum, GDK_KEY_KP_5, 0x2075, 
GDK_KEY_asciicircum, GDK_KEY_KP_6, 0x2076, 
GDK_KEY_asciicircum, GDK_KEY_KP_7, 0x2077, 
GDK_KEY_asciicircum, GDK_KEY_KP_8, 0x2078, 
GDK_KEY_asciicircum, GDK_KEY_KP_9, 0x2079, 
GDK_KEY_asciicircum, GDK_KEY_KP_Equal, 0x207C, 
GDK_KEY_underscore, GDK_KEY_apostrophe, 0x2358, 
GDK_KEY_underscore, GDK_KEY_parenleft, 0x208D, 
GDK_KEY_underscore, GDK_KEY_parenright, 0x208E, 
GDK_KEY_underscore, GDK_KEY_plus, 0x208A, 
GDK_KEY_underscore, GDK_KEY_0, 0x2080, 
GDK_KEY_underscore, GDK_KEY_1, 0x2081, 
GDK_KEY_underscore, GDK_KEY_2, 0x2082, 
GDK_KEY_underscore, GDK_KEY_3, 0x2083, 
GDK_KEY_underscore, GDK_KEY_4, 0x2084, 
GDK_KEY_underscore, GDK_KEY_5, 0x2085, 
GDK_KEY_underscore, GDK_KEY_6, 0x2086, 
GDK_KEY_underscore, GDK_KEY_7, 0x2087, 
GDK_KEY_underscore, GDK_KEY_8, 0x2088, 
GDK_KEY_underscore, GDK_KEY_9, 0x2089, 
GDK_KEY_underscore, GDK_KEY_less, 0x2264, 
GDK_KEY_underscore, GDK_KEY_equal, 0x208C, 
GDK_KEY_underscore, GDK_KEY_greater, 0x2265, 
GDK_KEY_underscore, GDK_KEY_A, 0x0100, 
GDK_KEY_underscore, GDK_KEY_E, 0x0112, 
GDK_KEY_underscore, GDK_KEY_G, 0x1E20, 
GDK_KEY_underscore, GDK_KEY_I, 0x012A, 
GDK_KEY_underscore, GDK_KEY_O, 0x014C, 
GDK_KEY_underscore, GDK_KEY_U, 0x016A, 
GDK_KEY_underscore, GDK_KEY_Y, 0x0232, 
GDK_KEY_underscore, GDK_KEY_asciicircum, 0x00AF, 
GDK_KEY_underscore, GDK_KEY_underscore, 0x00AF, 
GDK_KEY_underscore, GDK_KEY_a, 0x0101, 
GDK_KEY_underscore, GDK_KEY_e, 0x0113, 
GDK_KEY_underscore, GDK_KEY_g, 0x1E21, 
GDK_KEY_underscore, GDK_KEY_i, 0x012B, 
GDK_KEY_underscore, GDK_KEY_o, 0x014D, 
GDK_KEY_underscore, GDK_KEY_u, 0x016B, 
GDK_KEY_underscore, GDK_KEY_y, 0x0233, 
GDK_KEY_underscore, GDK_KEY_Adiaeresis, 0x01DE, 
GDK_KEY_underscore, GDK_KEY_AE, 0x01E2, 
GDK_KEY_underscore, GDK_KEY_Otilde, 0x022C, 
GDK_KEY_underscore, GDK_KEY_Odiaeresis, 0x022A, 
GDK_KEY_underscore, GDK_KEY_Udiaeresis, 0x01D5, 
GDK_KEY_underscore, GDK_KEY_adiaeresis, 0x01DF, 
GDK_KEY_underscore, GDK_KEY_ae, 0x01E3, 
GDK_KEY_underscore, GDK_KEY_otilde, 0x022D, 
GDK_KEY_underscore, GDK_KEY_odiaeresis, 0x022B, 
GDK_KEY_underscore, GDK_KEY_udiaeresis, 0x01D6, 
GDK_KEY_underscore, 0x01EA, 0x01EC, 
GDK_KEY_underscore, 0x01EB, 0x01ED, 
GDK_KEY_underscore, 0x0226, 0x01E0, 
GDK_KEY_underscore, 0x0227, 0x01E1, 
GDK_KEY_underscore, 0x022E, 0x0230, 
GDK_KEY_underscore, 0x022F, 0x0231, 
GDK_KEY_underscore, GDK_KEY_Cyrillic_i, 0x04E3, 
GDK_KEY_underscore, GDK_KEY_Cyrillic_u, 0x04EF, 
GDK_KEY_underscore, GDK_KEY_Cyrillic_I, 0x04E2, 
GDK_KEY_underscore, GDK_KEY_Cyrillic_U, 0x04EE, 
GDK_KEY_underscore, GDK_KEY_Greek_ALPHA, 0x1FB9, 
GDK_KEY_underscore, GDK_KEY_Greek_IOTA, 0x1FD9, 
GDK_KEY_underscore, GDK_KEY_Greek_UPSILON, 0x1FE9, 
GDK_KEY_underscore, GDK_KEY_Greek_alpha, 0x1FB1, 
GDK_KEY_underscore, GDK_KEY_Greek_iota, 0x1FD1, 
GDK_KEY_underscore, GDK_KEY_Greek_upsilon, 0x1FE1, 
GDK_KEY_underscore, 0x1E36, 0x1E38, 
GDK_KEY_underscore, 0x1E37, 0x1E39, 
GDK_KEY_underscore, 0x1E5A, 0x1E5C, 
GDK_KEY_underscore, 0x1E5B, 0x1E5D, 
GDK_KEY_underscore, 0x2206, 0x2359, 
GDK_KEY_underscore, 0x220A, 0x2377, 
GDK_KEY_underscore, 0x2212, 0x208B, 
GDK_KEY_underscore, 0x2218, 0x235B, 
GDK_KEY_underscore, 0x2260, 0x2262, 
GDK_KEY_underscore, 0x2282, 0x2286, 
GDK_KEY_underscore, 0x2283, 0x2287, 
GDK_KEY_underscore, GDK_KEY_downtack, 0x234A, 
GDK_KEY_underscore, 0x22C4, 0x235A, 
GDK_KEY_underscore, 0x2373, 0x2378, 
GDK_KEY_underscore, 0x2375, 0x2379, 
GDK_KEY_underscore, 0x237A, 0x2376, 
GDK_KEY_underscore, GDK_KEY_emopencircle, 0x235C, 
GDK_KEY_underscore, GDK_KEY_KP_Space, 0x2082, 
GDK_KEY_underscore, GDK_KEY_KP_Add, 0x208A, 
GDK_KEY_underscore, GDK_KEY_KP_0, 0x2080, 
GDK_KEY_underscore, GDK_KEY_KP_1, 0x2081, 
GDK_KEY_underscore, GDK_KEY_KP_2, 0x2082, 
GDK_KEY_underscore, GDK_KEY_KP_3, 0x2083, 
GDK_KEY_underscore, GDK_KEY_KP_4, 0x2084, 
GDK_KEY_underscore, GDK_KEY_KP_5, 0x2085, 
GDK_KEY_underscore, GDK_KEY_KP_6, 0x2086, 
GDK_KEY_underscore, GDK_KEY_KP_7, 0x2087, 
GDK_KEY_underscore, GDK_KEY_KP_8, 0x2088, 
GDK_KEY_underscore, GDK_KEY_KP_9, 0x2089, 
GDK_KEY_underscore, GDK_KEY_KP_Equal, 0x208C, 
GDK_KEY_grave, GDK_KEY_space, 0x0060, 
GDK_KEY_grave, GDK_KEY_A, 0x00C0, 
GDK_KEY_grave, GDK_KEY_E, 0x00C8, 
GDK_KEY_grave, GDK_KEY_I, 0x00CC, 
GDK_KEY_grave, GDK_KEY_N, 0x01F8, 
GDK_KEY_grave, GDK_KEY_O, 0x00D2, 
GDK_KEY_grave, GDK_KEY_U, 0x00D9, 
GDK_KEY_grave, GDK_KEY_W, 0x1E80, 
GDK_KEY_grave, GDK_KEY_Y, 0x1EF2, 
GDK_KEY_grave, GDK_KEY_a, 0x00E0, 
GDK_KEY_grave, GDK_KEY_e, 0x00E8, 
GDK_KEY_grave, GDK_KEY_i, 0x00EC, 
GDK_KEY_grave, GDK_KEY_n, 0x01F9, 
GDK_KEY_grave, GDK_KEY_o, 0x00F2, 
GDK_KEY_grave, GDK_KEY_u, 0x00F9, 
GDK_KEY_grave, GDK_KEY_w, 0x1E81, 
GDK_KEY_grave, GDK_KEY_y, 0x1EF3, 
GDK_KEY_grave, GDK_KEY_Acircumflex, 0x1EA6, 
GDK_KEY_grave, GDK_KEY_Ecircumflex, 0x1EC0, 
GDK_KEY_grave, GDK_KEY_Ocircumflex, 0x1ED2, 
GDK_KEY_grave, GDK_KEY_Udiaeresis, 0x01DB, 
GDK_KEY_grave, GDK_KEY_acircumflex, 0x1EA7, 
GDK_KEY_grave, GDK_KEY_ecircumflex, 0x1EC1, 
GDK_KEY_grave, GDK_KEY_ocircumflex, 0x1ED3, 
GDK_KEY_grave, GDK_KEY_udiaeresis, 0x01DC, 
GDK_KEY_grave, GDK_KEY_Abreve, 0x1EB0, 
GDK_KEY_grave, GDK_KEY_abreve, 0x1EB1, 
GDK_KEY_grave, GDK_KEY_Emacron, 0x1E14, 
GDK_KEY_grave, GDK_KEY_emacron, 0x1E15, 
GDK_KEY_grave, GDK_KEY_Omacron, 0x1E50, 
GDK_KEY_grave, GDK_KEY_omacron, 0x1E51, 
GDK_KEY_grave, GDK_KEY_Cyrillic_ie, 0x0450, 
GDK_KEY_grave, GDK_KEY_Cyrillic_i, 0x045D, 
GDK_KEY_grave, GDK_KEY_Cyrillic_IE, 0x0400, 
GDK_KEY_grave, GDK_KEY_Cyrillic_I, 0x040D, 
GDK_KEY_grave, GDK_KEY_Greek_iotadieresis, 0x1FD2, 
GDK_KEY_grave, GDK_KEY_Greek_upsilondieresis, 0x1FE2, 
GDK_KEY_grave, GDK_KEY_Greek_ALPHA, 0x1FBA, 
GDK_KEY_grave, GDK_KEY_Greek_EPSILON, 0x1FC8, 
GDK_KEY_grave, GDK_KEY_Greek_ETA, 0x1FCA, 
GDK_KEY_grave, GDK_KEY_Greek_IOTA, 0x1FDA, 
GDK_KEY_grave, GDK_KEY_Greek_OMICRON, 0x1FF8, 
GDK_KEY_grave, GDK_KEY_Greek_UPSILON, 0x1FEA, 
GDK_KEY_grave, GDK_KEY_Greek_OMEGA, 0x1FFA, 
GDK_KEY_grave, GDK_KEY_Greek_alpha, 0x1F70, 
GDK_KEY_grave, GDK_KEY_Greek_epsilon, 0x1F72, 
GDK_KEY_grave, GDK_KEY_Greek_eta, 0x1F74, 
GDK_KEY_grave, GDK_KEY_Greek_iota, 0x1F76, 
GDK_KEY_grave, GDK_KEY_Greek_omicron, 0x1F78, 
GDK_KEY_grave, GDK_KEY_Greek_upsilon, 0x1F7A, 
GDK_KEY_grave, GDK_KEY_Greek_omega, 0x1F7C, 
GDK_KEY_grave, 0x1F00, 0x1F02, 
GDK_KEY_grave, 0x1F01, 0x1F03, 
GDK_KEY_grave, 0x1F08, 0x1F0A, 
GDK_KEY_grave, 0x1F09, 0x1F0B, 
GDK_KEY_grave, 0x1F10, 0x1F12, 
GDK_KEY_grave, 0x1F11, 0x1F13, 
GDK_KEY_grave, 0x1F18, 0x1F1A, 
GDK_KEY_grave, 0x1F19, 0x1F1B, 
GDK_KEY_grave, 0x1F20, 0x1F22, 
GDK_KEY_grave, 0x1F21, 0x1F23, 
GDK_KEY_grave, 0x1F28, 0x1F2A, 
GDK_KEY_grave, 0x1F29, 0x1F2B, 
GDK_KEY_grave, 0x1F30, 0x1F32, 
GDK_KEY_grave, 0x1F31, 0x1F33, 
GDK_KEY_grave, 0x1F38, 0x1F3A, 
GDK_KEY_grave, 0x1F39, 0x1F3B, 
GDK_KEY_grave, 0x1F40, 0x1F42, 
GDK_KEY_grave, 0x1F41, 0x1F43, 
GDK_KEY_grave, 0x1F48, 0x1F4A, 
GDK_KEY_grave, 0x1F49, 0x1F4B, 
GDK_KEY_grave, 0x1F50, 0x1F52, 
GDK_KEY_grave, 0x1F51, 0x1F53, 
GDK_KEY_grave, 0x1F59, 0x1F5B, 
GDK_KEY_grave, 0x1F60, 0x1F62, 
GDK_KEY_grave, 0x1F61, 0x1F63, 
GDK_KEY_grave, 0x1F68, 0x1F6A, 
GDK_KEY_grave, 0x1F69, 0x1F6B, 
GDK_KEY_a, GDK_KEY_quotedbl, 0x00E4, 
GDK_KEY_a, GDK_KEY_apostrophe, 0x00E1, 
GDK_KEY_a, GDK_KEY_parenleft, 0x0103, 
GDK_KEY_a, GDK_KEY_asterisk, 0x00E5, 
GDK_KEY_a, GDK_KEY_comma, 0x0105, 
GDK_KEY_a, GDK_KEY_minus, 0x0101, 
GDK_KEY_a, GDK_KEY_semicolon, 0x0105, 
GDK_KEY_a, GDK_KEY_greater, 0x00E2, 
GDK_KEY_a, GDK_KEY_asciicircum, 0x00E2, 
GDK_KEY_a, GDK_KEY_underscore, 0x0101, 
GDK_KEY_a, GDK_KEY_grave, 0x00E0, 
GDK_KEY_a, GDK_KEY_a, 0x00E5, 
GDK_KEY_a, GDK_KEY_e, 0x00E6, 
GDK_KEY_a, GDK_KEY_asciitilde, 0x00E3, 
GDK_KEY_a, GDK_KEY_diaeresis, 0x00E4, 
GDK_KEY_a, GDK_KEY_acute, 0x00E1, 
GDK_KEY_b, GDK_KEY_period, 0x1E03, 
GDK_KEY_b, GDK_KEY_A, 0x0102, 
GDK_KEY_b, GDK_KEY_E, 0x0114, 
GDK_KEY_b, GDK_KEY_G, 0x011E, 
GDK_KEY_b, GDK_KEY_I, 0x012C, 
GDK_KEY_b, GDK_KEY_O, 0x014E, 
GDK_KEY_b, GDK_KEY_U, 0x016C, 
GDK_KEY_b, GDK_KEY_a, 0x0103, 
GDK_KEY_b, GDK_KEY_e, 0x0115, 
GDK_KEY_b, GDK_KEY_g, 0x011F, 
GDK_KEY_b, GDK_KEY_i, 0x012D, 
GDK_KEY_b, GDK_KEY_o, 0x014F, 
GDK_KEY_b, GDK_KEY_u, 0x016D, 
GDK_KEY_b, 0x0228, 0x1E1C, 
GDK_KEY_b, 0x0229, 0x1E1D, 
GDK_KEY_b, GDK_KEY_Cyrillic_a, 0x04D1, 
GDK_KEY_b, GDK_KEY_Cyrillic_ie, 0x04D7, 
GDK_KEY_b, GDK_KEY_Cyrillic_i, 0x0439, 
GDK_KEY_b, GDK_KEY_Cyrillic_u, 0x045E, 
GDK_KEY_b, GDK_KEY_Cyrillic_zhe, 0x04C2, 
GDK_KEY_b, GDK_KEY_Cyrillic_A, 0x04D0, 
GDK_KEY_b, GDK_KEY_Cyrillic_IE, 0x04D6, 
GDK_KEY_b, GDK_KEY_Cyrillic_I, 0x0419, 
GDK_KEY_b, GDK_KEY_Cyrillic_U, 0x040E, 
GDK_KEY_b, GDK_KEY_Cyrillic_ZHE, 0x04C1, 
GDK_KEY_b, GDK_KEY_Greek_ALPHA, 0x1FB8, 
GDK_KEY_b, GDK_KEY_Greek_IOTA, 0x1FD8, 
GDK_KEY_b, GDK_KEY_Greek_UPSILON, 0x1FE8, 
GDK_KEY_b, GDK_KEY_Greek_alpha, 0x1FB0, 
GDK_KEY_b, GDK_KEY_Greek_iota, 0x1FD0, 
GDK_KEY_b, GDK_KEY_Greek_upsilon, 0x1FE0, 
GDK_KEY_b, 0x1EA0, 0x1EB6, 
GDK_KEY_b, 0x1EA1, 0x1EB7, 
GDK_KEY_c, GDK_KEY_apostrophe, 0x0107, 
GDK_KEY_c, GDK_KEY_comma, 0x00E7, 
GDK_KEY_c, GDK_KEY_period, 0x010B, 
GDK_KEY_c, GDK_KEY_slash, 0x00A2, 
GDK_KEY_c, GDK_KEY_less, 0x010D, 
GDK_KEY_c, GDK_KEY_equal, 0x20AC, 
GDK_KEY_c, GDK_KEY_A, 0x01CD, 
GDK_KEY_c, GDK_KEY_C, 0x010C, 
GDK_KEY_c, GDK_KEY_D, 0x010E, 
GDK_KEY_c, GDK_KEY_E, 0x011A, 
GDK_KEY_c, GDK_KEY_G, 0x01E6, 
GDK_KEY_c, GDK_KEY_H, 0x021E, 
GDK_KEY_c, GDK_KEY_I, 0x01CF, 
GDK_KEY_c, GDK_KEY_K, 0x01E8, 
GDK_KEY_c, GDK_KEY_L, 0x013D, 
GDK_KEY_c, GDK_KEY_N, 0x0147, 
GDK_KEY_c, GDK_KEY_O, 0x01D1, 
GDK_KEY_c, GDK_KEY_R, 0x0158, 
GDK_KEY_c, GDK_KEY_S, 0x0160, 
GDK_KEY_c, GDK_KEY_T, 0x0164, 
GDK_KEY_c, GDK_KEY_U, 0x01D3, 
GDK_KEY_c, GDK_KEY_Z, 0x017D, 
GDK_KEY_c, GDK_KEY_a, 0x01CE, 
GDK_KEY_c, GDK_KEY_c, 0x010D, 
GDK_KEY_c, GDK_KEY_d, 0x010F, 
GDK_KEY_c, GDK_KEY_e, 0x011B, 
GDK_KEY_c, GDK_KEY_g, 0x01E7, 
GDK_KEY_c, GDK_KEY_h, 0x021F, 
GDK_KEY_c, GDK_KEY_i, 0x01D0, 
GDK_KEY_c, GDK_KEY_j, 0x01F0, 
GDK_KEY_c, GDK_KEY_k, 0x01E9, 
GDK_KEY_c, GDK_KEY_l, 0x013E, 
GDK_KEY_c, GDK_KEY_n, 0x0148, 
GDK_KEY_c, GDK_KEY_o, 0x01D2, 
GDK_KEY_c, GDK_KEY_r, 0x0159, 
GDK_KEY_c, GDK_KEY_s, 0x0161, 
GDK_KEY_c, GDK_KEY_t, 0x0165, 
GDK_KEY_c, GDK_KEY_u, 0x01D4, 
GDK_KEY_c, GDK_KEY_z, 0x017E, 
GDK_KEY_c, GDK_KEY_bar, 0x00A2, 
GDK_KEY_c, GDK_KEY_Udiaeresis, 0x01D9, 
GDK_KEY_c, GDK_KEY_udiaeresis, 0x01DA, 
GDK_KEY_d, GDK_KEY_comma, 0x1E11, 
GDK_KEY_d, GDK_KEY_minus, 0x0111, 
GDK_KEY_d, GDK_KEY_period, 0x1E0B, 
GDK_KEY_d, GDK_KEY_less, 0x010F, 
GDK_KEY_d, GDK_KEY_equal, 0x20AB, 
GDK_KEY_d, GDK_KEY_h, 0x00F0, 
GDK_KEY_d, GDK_KEY_i, 0x2300, 
GDK_KEY_e, GDK_KEY_quotedbl, 0x00EB, 
GDK_KEY_e, GDK_KEY_apostrophe, 0x00E9, 
GDK_KEY_e, GDK_KEY_comma, 0x0119, 
GDK_KEY_e, GDK_KEY_minus, 0x0113, 
GDK_KEY_e, GDK_KEY_period, 0x0117, 
GDK_KEY_e, GDK_KEY_semicolon, 0x0119, 
GDK_KEY_e, GDK_KEY_less, 0x011B, 
GDK_KEY_e, GDK_KEY_equal, 0x20AC, 
GDK_KEY_e, GDK_KEY_greater, 0x00EA, 
GDK_KEY_e, GDK_KEY_asciicircum, 0x00EA, 
GDK_KEY_e, GDK_KEY_underscore, 0x0113, 
GDK_KEY_e, GDK_KEY_grave, 0x00E8, 
GDK_KEY_e, GDK_KEY_e, 0x0259, 
GDK_KEY_e, GDK_KEY_diaeresis, 0x00EB, 
GDK_KEY_e, GDK_KEY_acute, 0x00E9, 
GDK_KEY_f, GDK_KEY_period, 0x1E1F, 
GDK_KEY_f, GDK_KEY_S, 0x017F, 
GDK_KEY_f, GDK_KEY_f, 0xFB00, 
GDK_KEY_f, GDK_KEY_i, 0xFB01, 
GDK_KEY_f, GDK_KEY_l, 0xFB02, 
GDK_KEY_f, GDK_KEY_s, 0x017F, 
GDK_KEY_g, GDK_KEY_parenleft, 0x011F, 
GDK_KEY_g, GDK_KEY_comma, 0x0123, 
GDK_KEY_g, GDK_KEY_period, 0x0121, 
GDK_KEY_g, GDK_KEY_U, 0x011F, 
GDK_KEY_g, GDK_KEY_breve, 0x011F, 
GDK_KEY_h, GDK_KEY_comma, 0x1E29, 
GDK_KEY_i, GDK_KEY_quotedbl, 0x00EF, 
GDK_KEY_i, GDK_KEY_apostrophe, 0x00ED, 
GDK_KEY_i, GDK_KEY_comma, 0x012F, 
GDK_KEY_i, GDK_KEY_minus, 0x012B, 
GDK_KEY_i, GDK_KEY_period, 0x0131, 
GDK_KEY_i, GDK_KEY_semicolon, 0x012F, 
GDK_KEY_i, GDK_KEY_greater, 0x00EE, 
GDK_KEY_i, GDK_KEY_asciicircum, 0x00EE, 
GDK_KEY_i, GDK_KEY_underscore, 0x012B, 
GDK_KEY_i, GDK_KEY_grave, 0x00EC, 
GDK_KEY_i, GDK_KEY_j, 0x0133, 
GDK_KEY_i, GDK_KEY_asciitilde, 0x0129, 
GDK_KEY_i, GDK_KEY_diaeresis, 0x00EF, 
GDK_KEY_i, GDK_KEY_acute, 0x00ED, 
GDK_KEY_k, GDK_KEY_comma, 0x0137, 
GDK_KEY_k, GDK_KEY_k, 0x0138, 
GDK_KEY_l, GDK_KEY_apostrophe, 0x013A, 
GDK_KEY_l, GDK_KEY_comma, 0x013C, 
GDK_KEY_l, GDK_KEY_minus, 0x00A3, 
GDK_KEY_l, GDK_KEY_slash, 0x0142, 
GDK_KEY_l, GDK_KEY_less, 0x013E, 
GDK_KEY_l, GDK_KEY_v, 0x007C, 
GDK_KEY_m, GDK_KEY_period, 0x1E41, 
GDK_KEY_m, GDK_KEY_slash, 0x20A5, 
GDK_KEY_m, GDK_KEY_u, 0x00B5, 
GDK_KEY_n, GDK_KEY_apostrophe, 0x0144, 
GDK_KEY_n, GDK_KEY_comma, 0x0146, 
GDK_KEY_n, GDK_KEY_less, 0x0148, 
GDK_KEY_n, GDK_KEY_g, 0x014B, 
GDK_KEY_n, GDK_KEY_asciitilde, 0x00F1, 
GDK_KEY_o, GDK_KEY_quotedbl, 0x00F6, 
GDK_KEY_o, GDK_KEY_apostrophe, 0x00F3, 
GDK_KEY_o, GDK_KEY_comma, 0x01EB, 
GDK_KEY_o, GDK_KEY_minus, 0x014D, 
GDK_KEY_o, GDK_KEY_slash, 0x00F8, 
GDK_KEY_o, GDK_KEY_semicolon, 0x01EB, 
GDK_KEY_o, GDK_KEY_greater, 0x00F4, 
GDK_KEY_o, GDK_KEY_A, 0x00C5, 
GDK_KEY_o, GDK_KEY_C, 0x00A9, 
GDK_KEY_o, GDK_KEY_R, 0x00AE, 
GDK_KEY_o, GDK_KEY_U, 0x016E, 
GDK_KEY_o, GDK_KEY_X, 0x00A4, 
GDK_KEY_o, GDK_KEY_asciicircum, 0x00F4, 
GDK_KEY_o, GDK_KEY_underscore, 0x014D, 
GDK_KEY_o, GDK_KEY_grave, 0x00F2, 
GDK_KEY_o, GDK_KEY_a, 0x00E5, 
GDK_KEY_o, GDK_KEY_c, 0x00A9, 
GDK_KEY_o, GDK_KEY_e, 0x0153, 
GDK_KEY_o, GDK_KEY_o, 0x00B0, 
GDK_KEY_o, GDK_KEY_r, 0x00AE, 
GDK_KEY_o, GDK_KEY_s, 0x00A7, 
GDK_KEY_o, GDK_KEY_u, 0x016F, 
GDK_KEY_o, GDK_KEY_w, 0x1E98, 
GDK_KEY_o, GDK_KEY_x, 0x00A4, 
GDK_KEY_o, GDK_KEY_y, 0x1E99, 
GDK_KEY_o, GDK_KEY_asciitilde, 0x00F5, 
GDK_KEY_o, GDK_KEY_diaeresis, 0x00F6, 
GDK_KEY_o, GDK_KEY_acute, 0x00F3, 
GDK_KEY_p, GDK_KEY_exclam, 0x00B6, 
GDK_KEY_p, GDK_KEY_period, 0x1E57, 
GDK_KEY_r, GDK_KEY_apostrophe, 0x0155, 
GDK_KEY_r, GDK_KEY_comma, 0x0157, 
GDK_KEY_r, GDK_KEY_less, 0x0159, 
GDK_KEY_s, GDK_KEY_exclam, 0x00A7, 
GDK_KEY_s, GDK_KEY_apostrophe, 0x015B, 
GDK_KEY_s, GDK_KEY_comma, 0x015F, 
GDK_KEY_s, GDK_KEY_period, 0x1E61, 
GDK_KEY_s, GDK_KEY_semicolon, 0x0219, 
GDK_KEY_s, GDK_KEY_less, 0x0161, 
GDK_KEY_s, GDK_KEY_M, 0x2120, 
GDK_KEY_s, GDK_KEY_m, 0x2120, 
GDK_KEY_s, GDK_KEY_o, 0x00A7, 
GDK_KEY_s, GDK_KEY_s, 0x00DF, 
GDK_KEY_s, GDK_KEY_cedilla, 0x015F, 
GDK_KEY_t, GDK_KEY_comma, 0x0163, 
GDK_KEY_t, GDK_KEY_minus, 0x0167, 
GDK_KEY_t, GDK_KEY_period, 0x1E6B, 
GDK_KEY_t, GDK_KEY_slash, 0x0167, 
GDK_KEY_t, GDK_KEY_semicolon, 0x021B, 
GDK_KEY_t, GDK_KEY_less, 0x0165, 
GDK_KEY_t, GDK_KEY_M, 0x2122, 
GDK_KEY_t, GDK_KEY_h, 0x00FE, 
GDK_KEY_t, GDK_KEY_m, 0x2122, 
GDK_KEY_u, GDK_KEY_quotedbl, 0x00FC, 
GDK_KEY_u, GDK_KEY_apostrophe, 0x00FA, 
GDK_KEY_u, GDK_KEY_asterisk, 0x016F, 
GDK_KEY_u, GDK_KEY_comma, 0x0173, 
GDK_KEY_u, GDK_KEY_minus, 0x016B, 
GDK_KEY_u, GDK_KEY_slash, 0x00B5, 
GDK_KEY_u, GDK_KEY_semicolon, 0x0173, 
GDK_KEY_u, GDK_KEY_greater, 0x00FB, 
GDK_KEY_u, GDK_KEY_A, 0x0102, 
GDK_KEY_u, GDK_KEY_U, 0x016C, 
GDK_KEY_u, GDK_KEY_asciicircum, 0x00FB, 
GDK_KEY_u, GDK_KEY_underscore, 0x016B, 
GDK_KEY_u, GDK_KEY_grave, 0x00F9, 
GDK_KEY_u, GDK_KEY_a, 0x0103, 
GDK_KEY_u, GDK_KEY_u, 0x016D, 
GDK_KEY_u, GDK_KEY_asciitilde, 0x0169, 
GDK_KEY_u, GDK_KEY_diaeresis, 0x00FC, 
GDK_KEY_u, GDK_KEY_acute, 0x00FA, 
GDK_KEY_v, GDK_KEY_slash, 0x221A, 
GDK_KEY_v, GDK_KEY_Z, 0x017D, 
GDK_KEY_v, GDK_KEY_l, 0x007C, 
GDK_KEY_v, GDK_KEY_z, 0x017E, 
GDK_KEY_w, GDK_KEY_asciicircum, 0x0175, 
GDK_KEY_x, GDK_KEY_O, 0x00A4, 
GDK_KEY_x, GDK_KEY_o, 0x00A4, 
GDK_KEY_x, GDK_KEY_x, 0x00D7, 
GDK_KEY_y, GDK_KEY_quotedbl, 0x00FF, 
GDK_KEY_y, GDK_KEY_apostrophe, 0x00FD, 
GDK_KEY_y, GDK_KEY_minus, 0x00A5, 
GDK_KEY_y, GDK_KEY_equal, 0x00A5, 
GDK_KEY_y, GDK_KEY_asciicircum, 0x0177, 
GDK_KEY_y, GDK_KEY_diaeresis, 0x00FF, 
GDK_KEY_y, GDK_KEY_acute, 0x00FD, 
GDK_KEY_z, GDK_KEY_apostrophe, 0x017A, 
GDK_KEY_z, GDK_KEY_period, 0x017C, 
GDK_KEY_z, GDK_KEY_less, 0x017E, 
GDK_KEY_braceleft, GDK_KEY_braceright, 0x2205, 
GDK_KEY_bar, GDK_KEY_C, 0x00A2, 
GDK_KEY_bar, GDK_KEY_c, 0x00A2, 
GDK_KEY_bar, GDK_KEY_asciitilde, 0x236D, 
GDK_KEY_bar, 0x2190, 0x2345, 
GDK_KEY_bar, 0x2192, 0x2346, 
GDK_KEY_bar, 0x2206, 0x234B, 
GDK_KEY_bar, 0x2207, 0x2352, 
GDK_KEY_bar, GDK_KEY_union, 0x2366, 
GDK_KEY_bar, 0x2282, 0x2367, 
GDK_KEY_bar, GDK_KEY_emopencircle, 0x233D, 
GDK_KEY_asciitilde, GDK_KEY_space, 0x007E, 
GDK_KEY_asciitilde, GDK_KEY_0, 0x236C, 
GDK_KEY_asciitilde, GDK_KEY_A, 0x00C3, 
GDK_KEY_asciitilde, GDK_KEY_E, 0x1EBC, 
GDK_KEY_asciitilde, GDK_KEY_I, 0x0128, 
GDK_KEY_asciitilde, GDK_KEY_N, 0x00D1, 
GDK_KEY_asciitilde, GDK_KEY_O, 0x00D5, 
GDK_KEY_asciitilde, GDK_KEY_U, 0x0168, 
GDK_KEY_asciitilde, GDK_KEY_V, 0x1E7C, 
GDK_KEY_asciitilde, GDK_KEY_Y, 0x1EF8, 
GDK_KEY_asciitilde, GDK_KEY_a, 0x00E3, 
GDK_KEY_asciitilde, GDK_KEY_e, 0x1EBD, 
GDK_KEY_asciitilde, GDK_KEY_i, 0x0129, 
GDK_KEY_asciitilde, GDK_KEY_n, 0x00F1, 
GDK_KEY_asciitilde, GDK_KEY_o, 0x00F5, 
GDK_KEY_asciitilde, GDK_KEY_u, 0x0169, 
GDK_KEY_asciitilde, GDK_KEY_v, 0x1E7D, 
GDK_KEY_asciitilde, GDK_KEY_y, 0x1EF9, 
GDK_KEY_asciitilde, GDK_KEY_bar, 0x236D, 
GDK_KEY_asciitilde, GDK_KEY_diaeresis, 0x2368, 
GDK_KEY_asciitilde, GDK_KEY_Acircumflex, 0x1EAA, 
GDK_KEY_asciitilde, GDK_KEY_Ecircumflex, 0x1EC4, 
GDK_KEY_asciitilde, GDK_KEY_Ocircumflex, 0x1ED6, 
GDK_KEY_asciitilde, GDK_KEY_acircumflex, 0x1EAB, 
GDK_KEY_asciitilde, GDK_KEY_ecircumflex, 0x1EC5, 
GDK_KEY_asciitilde, GDK_KEY_ocircumflex, 0x1ED7, 
GDK_KEY_asciitilde, GDK_KEY_Abreve, 0x1EB4, 
GDK_KEY_asciitilde, GDK_KEY_abreve, 0x1EB5, 
GDK_KEY_asciitilde, GDK_KEY_Greek_iotadieresis, 0x1FD7, 
GDK_KEY_asciitilde, GDK_KEY_Greek_upsilondieresis, 0x1FE7, 
GDK_KEY_asciitilde, GDK_KEY_Greek_alpha, 0x1FB6, 
GDK_KEY_asciitilde, GDK_KEY_Greek_eta, 0x1FC6, 
GDK_KEY_asciitilde, GDK_KEY_Greek_iota, 0x1FD6, 
GDK_KEY_asciitilde, GDK_KEY_Greek_upsilon, 0x1FE6, 
GDK_KEY_asciitilde, GDK_KEY_Greek_omega, 0x1FF6, 
GDK_KEY_asciitilde, 0x1F00, 0x1F06, 
GDK_KEY_asciitilde, 0x1F01, 0x1F07, 
GDK_KEY_asciitilde, 0x1F08, 0x1F0E, 
GDK_KEY_asciitilde, 0x1F09, 0x1F0F, 
GDK_KEY_asciitilde, 0x1F20, 0x1F26, 
GDK_KEY_asciitilde, 0x1F21, 0x1F27, 
GDK_KEY_asciitilde, 0x1F28, 0x1F2E, 
GDK_KEY_asciitilde, 0x1F29, 0x1F2F, 
GDK_KEY_asciitilde, 0x1F30, 0x1F36, 
GDK_KEY_asciitilde, 0x1F31, 0x1F37, 
GDK_KEY_asciitilde, 0x1F38, 0x1F3E, 
GDK_KEY_asciitilde, 0x1F39, 0x1F3F, 
GDK_KEY_asciitilde, 0x1F50, 0x1F56, 
GDK_KEY_asciitilde, 0x1F51, 0x1F57, 
GDK_KEY_asciitilde, 0x1F59, 0x1F5F, 
GDK_KEY_asciitilde, 0x1F60, 0x1F66, 
GDK_KEY_asciitilde, 0x1F61, 0x1F67, 
GDK_KEY_asciitilde, 0x1F68, 0x1F6E, 
GDK_KEY_asciitilde, 0x1F69, 0x1F6F, 
GDK_KEY_asciitilde, 0x2207, 0x236B, 
GDK_KEY_asciitilde, 0x2227, 0x2372, 
GDK_KEY_asciitilde, 0x2228, 0x2371, 
GDK_KEY_diaeresis, GDK_KEY_apostrophe, 0x0385, 
GDK_KEY_diaeresis, GDK_KEY_asterisk, 0x2363, 
GDK_KEY_diaeresis, GDK_KEY_greater, 0x2369, 
GDK_KEY_diaeresis, GDK_KEY_A, 0x00C4, 
GDK_KEY_diaeresis, GDK_KEY_E, 0x00CB, 
GDK_KEY_diaeresis, GDK_KEY_I, 0x00CF, 
GDK_KEY_diaeresis, GDK_KEY_O, 0x00D6, 
GDK_KEY_diaeresis, GDK_KEY_U, 0x00DC, 
GDK_KEY_diaeresis, GDK_KEY_Y, 0x0178, 
GDK_KEY_diaeresis, GDK_KEY_grave, 0x1FED, 
GDK_KEY_diaeresis, GDK_KEY_a, 0x00E4, 
GDK_KEY_diaeresis, GDK_KEY_e, 0x00EB, 
GDK_KEY_diaeresis, GDK_KEY_i, 0x00EF, 
GDK_KEY_diaeresis, GDK_KEY_o, 0x00F6, 
GDK_KEY_diaeresis, GDK_KEY_u, 0x00FC, 
GDK_KEY_diaeresis, GDK_KEY_y, 0x00FF, 
GDK_KEY_diaeresis, GDK_KEY_asciitilde, 0x1FC1, 
GDK_KEY_diaeresis, GDK_KEY_acute, 0x0385, 
GDK_KEY_diaeresis, 0x2207, 0x2362, 
GDK_KEY_diaeresis, 0x2218, 0x2364, 
GDK_KEY_diaeresis, GDK_KEY_uptack, 0x2361, 
GDK_KEY_diaeresis, GDK_KEY_emopencircle, 0x2365, 
GDK_KEY_diaeresis, GDK_KEY_dead_grave, 0x1FED, 
GDK_KEY_diaeresis, GDK_KEY_dead_acute, 0x0385, 
GDK_KEY_diaeresis, GDK_KEY_dead_tilde, 0x1FC1, 
GDK_KEY_macron, GDK_KEY_A, 0x0100, 
GDK_KEY_macron, GDK_KEY_E, 0x0112, 
GDK_KEY_macron, GDK_KEY_G, 0x1E20, 
GDK_KEY_macron, GDK_KEY_I, 0x012A, 
GDK_KEY_macron, GDK_KEY_O, 0x014C, 
GDK_KEY_macron, GDK_KEY_U, 0x016A, 
GDK_KEY_macron, GDK_KEY_Y, 0x0232, 
GDK_KEY_macron, GDK_KEY_a, 0x0101, 
GDK_KEY_macron, GDK_KEY_e, 0x0113, 
GDK_KEY_macron, GDK_KEY_g, 0x1E21, 
GDK_KEY_macron, GDK_KEY_i, 0x012B, 
GDK_KEY_macron, GDK_KEY_o, 0x014D, 
GDK_KEY_macron, GDK_KEY_u, 0x016B, 
GDK_KEY_macron, GDK_KEY_y, 0x0233, 
GDK_KEY_macron, GDK_KEY_Adiaeresis, 0x01DE, 
GDK_KEY_macron, GDK_KEY_AE, 0x01E2, 
GDK_KEY_macron, GDK_KEY_Otilde, 0x022C, 
GDK_KEY_macron, GDK_KEY_Odiaeresis, 0x022A, 
GDK_KEY_macron, GDK_KEY_Udiaeresis, 0x01D5, 
GDK_KEY_macron, GDK_KEY_adiaeresis, 0x01DF, 
GDK_KEY_macron, GDK_KEY_ae, 0x01E3, 
GDK_KEY_macron, GDK_KEY_otilde, 0x022D, 
GDK_KEY_macron, GDK_KEY_odiaeresis, 0x022B, 
GDK_KEY_macron, GDK_KEY_udiaeresis, 0x01D6, 
GDK_KEY_macron, 0x01EA, 0x01EC, 
GDK_KEY_macron, 0x01EB, 0x01ED, 
GDK_KEY_macron, 0x0226, 0x01E0, 
GDK_KEY_macron, 0x0227, 0x01E1, 
GDK_KEY_macron, 0x022E, 0x0230, 
GDK_KEY_macron, 0x022F, 0x0231, 
GDK_KEY_macron, GDK_KEY_Cyrillic_i, 0x04E3, 
GDK_KEY_macron, GDK_KEY_Cyrillic_u, 0x04EF, 
GDK_KEY_macron, GDK_KEY_Cyrillic_I, 0x04E2, 
GDK_KEY_macron, GDK_KEY_Cyrillic_U, 0x04EE, 
GDK_KEY_macron, GDK_KEY_Greek_ALPHA, 0x1FB9, 
GDK_KEY_macron, GDK_KEY_Greek_IOTA, 0x1FD9, 
GDK_KEY_macron, GDK_KEY_Greek_UPSILON, 0x1FE9, 
GDK_KEY_macron, GDK_KEY_Greek_alpha, 0x1FB1, 
GDK_KEY_macron, GDK_KEY_Greek_iota, 0x1FD1, 
GDK_KEY_macron, GDK_KEY_Greek_upsilon, 0x1FE1, 
GDK_KEY_macron, 0x1E36, 0x1E38, 
GDK_KEY_macron, 0x1E37, 0x1E39, 
GDK_KEY_macron, 0x1E5A, 0x1E5C, 
GDK_KEY_macron, 0x1E5B, 0x1E5D, 
GDK_KEY_macron, GDK_KEY_uptack, 0x2351, 
GDK_KEY_acute, GDK_KEY_A, 0x00C1, 
GDK_KEY_acute, GDK_KEY_C, 0x0106, 
GDK_KEY_acute, GDK_KEY_E, 0x00C9, 
GDK_KEY_acute, GDK_KEY_G, 0x01F4, 
GDK_KEY_acute, GDK_KEY_I, 0x00CD, 
GDK_KEY_acute, GDK_KEY_K, 0x1E30, 
GDK_KEY_acute, GDK_KEY_L, 0x0139, 
GDK_KEY_acute, GDK_KEY_M, 0x1E3E, 
GDK_KEY_acute, GDK_KEY_N, 0x0143, 
GDK_KEY_acute, GDK_KEY_O, 0x00D3, 
GDK_KEY_acute, GDK_KEY_P, 0x1E54, 
GDK_KEY_acute, GDK_KEY_R, 0x0154, 
GDK_KEY_acute, GDK_KEY_S, 0x015A, 
GDK_KEY_acute, GDK_KEY_U, 0x00DA, 
GDK_KEY_acute, GDK_KEY_W, 0x1E82, 
GDK_KEY_acute, GDK_KEY_Y, 0x00DD, 
GDK_KEY_acute, GDK_KEY_Z, 0x0179, 
GDK_KEY_acute, GDK_KEY_a, 0x00E1, 
GDK_KEY_acute, GDK_KEY_c, 0x0107, 
GDK_KEY_acute, GDK_KEY_e, 0x00E9, 
GDK_KEY_acute, GDK_KEY_g, 0x01F5, 
GDK_KEY_acute, GDK_KEY_i, 0x00ED, 
GDK_KEY_acute, GDK_KEY_k, 0x1E31, 
GDK_KEY_acute, GDK_KEY_l, 0x013A, 
GDK_KEY_acute, GDK_KEY_m, 0x1E3F, 
GDK_KEY_acute, GDK_KEY_n, 0x0144, 
GDK_KEY_acute, GDK_KEY_o, 0x00F3, 
GDK_KEY_acute, GDK_KEY_p, 0x1E55, 
GDK_KEY_acute, GDK_KEY_r, 0x0155, 
GDK_KEY_acute, GDK_KEY_s, 0x015B, 
GDK_KEY_acute, GDK_KEY_u, 0x00FA, 
GDK_KEY_acute, GDK_KEY_w, 0x1E83, 
GDK_KEY_acute, GDK_KEY_y, 0x00FD, 
GDK_KEY_acute, GDK_KEY_z, 0x017A, 
GDK_KEY_acute, GDK_KEY_Acircumflex, 0x1EA4, 
GDK_KEY_acute, GDK_KEY_Aring, 0x01FA, 
GDK_KEY_acute, GDK_KEY_AE, 0x01FC, 
GDK_KEY_acute, GDK_KEY_Ccedilla, 0x1E08, 
GDK_KEY_acute, GDK_KEY_Ecircumflex, 0x1EBE, 
GDK_KEY_acute, GDK_KEY_Idiaeresis, 0x1E2E, 
GDK_KEY_acute, GDK_KEY_Ocircumflex, 0x1ED0, 
GDK_KEY_acute, GDK_KEY_Otilde, 0x1E4C, 
GDK_KEY_acute, GDK_KEY_Ooblique, 0x01FE, 
GDK_KEY_acute, GDK_KEY_Udiaeresis, 0x01D7, 
GDK_KEY_acute, GDK_KEY_acircumflex, 0x1EA5, 
GDK_KEY_acute, GDK_KEY_aring, 0x01FB, 
GDK_KEY_acute, GDK_KEY_ae, 0x01FD, 
GDK_KEY_acute, GDK_KEY_ccedilla, 0x1E09, 
GDK_KEY_acute, GDK_KEY_ecircumflex, 0x1EBF, 
GDK_KEY_acute, GDK_KEY_idiaeresis, 0x1E2F, 
GDK_KEY_acute, GDK_KEY_ocircumflex, 0x1ED1, 
GDK_KEY_acute, GDK_KEY_otilde, 0x1E4D, 
GDK_KEY_acute, GDK_KEY_oslash, 0x01FF, 
GDK_KEY_acute, GDK_KEY_udiaeresis, 0x01D8, 
GDK_KEY_acute, GDK_KEY_Abreve, 0x1EAE, 
GDK_KEY_acute, GDK_KEY_abreve, 0x1EAF, 
GDK_KEY_acute, GDK_KEY_Emacron, 0x1E16, 
GDK_KEY_acute, GDK_KEY_emacron, 0x1E17, 
GDK_KEY_acute, GDK_KEY_Omacron, 0x1E52, 
GDK_KEY_acute, GDK_KEY_Utilde, 0x1E78, 
GDK_KEY_acute, GDK_KEY_omacron, 0x1E53, 
GDK_KEY_acute, GDK_KEY_utilde, 0x1E79, 
GDK_KEY_acute, GDK_KEY_Cyrillic_ghe, 0x0453, 
GDK_KEY_acute, GDK_KEY_Cyrillic_ka, 0x045C, 
GDK_KEY_acute, GDK_KEY_Cyrillic_GHE, 0x0403, 
GDK_KEY_acute, GDK_KEY_Cyrillic_KA, 0x040C, 
GDK_KEY_acute, GDK_KEY_Greek_iotadieresis, 0x0390, 
GDK_KEY_acute, GDK_KEY_Greek_upsilondieresis, 0x03B0, 
GDK_KEY_acute, GDK_KEY_Greek_ALPHA, 0x0386, 
GDK_KEY_acute, GDK_KEY_Greek_EPSILON, 0x0388, 
GDK_KEY_acute, GDK_KEY_Greek_ETA, 0x0389, 
GDK_KEY_acute, GDK_KEY_Greek_IOTA, 0x038A, 
GDK_KEY_acute, GDK_KEY_Greek_OMICRON, 0x038C, 
GDK_KEY_acute, GDK_KEY_Greek_UPSILON, 0x038E, 
GDK_KEY_acute, GDK_KEY_Greek_OMEGA, 0x038F, 
GDK_KEY_acute, GDK_KEY_Greek_alpha, 0x03AC, 
GDK_KEY_acute, GDK_KEY_Greek_epsilon, 0x03AD, 
GDK_KEY_acute, GDK_KEY_Greek_eta, 0x03AE, 
GDK_KEY_acute, GDK_KEY_Greek_iota, 0x03AF, 
GDK_KEY_acute, GDK_KEY_Greek_omicron, 0x03CC, 
GDK_KEY_acute, GDK_KEY_Greek_upsilon, 0x03CD, 
GDK_KEY_acute, GDK_KEY_Greek_omega, 0x03CE, 
GDK_KEY_acute, 0x1F00, 0x1F04, 
GDK_KEY_acute, 0x1F01, 0x1F05, 
GDK_KEY_acute, 0x1F08, 0x1F0C, 
GDK_KEY_acute, 0x1F09, 0x1F0D, 
GDK_KEY_acute, 0x1F10, 0x1F14, 
GDK_KEY_acute, 0x1F11, 0x1F15, 
GDK_KEY_acute, 0x1F18, 0x1F1C, 
GDK_KEY_acute, 0x1F19, 0x1F1D, 
GDK_KEY_acute, 0x1F20, 0x1F24, 
GDK_KEY_acute, 0x1F21, 0x1F25, 
GDK_KEY_acute, 0x1F28, 0x1F2C, 
GDK_KEY_acute, 0x1F29, 0x1F2D, 
GDK_KEY_acute, 0x1F30, 0x1F34, 
GDK_KEY_acute, 0x1F31, 0x1F35, 
GDK_KEY_acute, 0x1F38, 0x1F3C, 
GDK_KEY_acute, 0x1F39, 0x1F3D, 
GDK_KEY_acute, 0x1F40, 0x1F44, 
GDK_KEY_acute, 0x1F41, 0x1F45, 
GDK_KEY_acute, 0x1F48, 0x1F4C, 
GDK_KEY_acute, 0x1F49, 0x1F4D, 
GDK_KEY_acute, 0x1F50, 0x1F54, 
GDK_KEY_acute, 0x1F51, 0x1F55, 
GDK_KEY_acute, 0x1F59, 0x1F5D, 
GDK_KEY_acute, 0x1F60, 0x1F64, 
GDK_KEY_acute, 0x1F61, 0x1F65, 
GDK_KEY_acute, 0x1F68, 0x1F6C, 
GDK_KEY_acute, 0x1F69, 0x1F6D, 
GDK_KEY_cedilla, GDK_KEY_C, 0x00C7, 
GDK_KEY_cedilla, GDK_KEY_D, 0x1E10, 
GDK_KEY_cedilla, GDK_KEY_E, 0x0228, 
GDK_KEY_cedilla, GDK_KEY_G, 0x0122, 
GDK_KEY_cedilla, GDK_KEY_H, 0x1E28, 
GDK_KEY_cedilla, GDK_KEY_K, 0x0136, 
GDK_KEY_cedilla, GDK_KEY_L, 0x013B, 
GDK_KEY_cedilla, GDK_KEY_N, 0x0145, 
GDK_KEY_cedilla, GDK_KEY_R, 0x0156, 
GDK_KEY_cedilla, GDK_KEY_S, 0x015E, 
GDK_KEY_cedilla, GDK_KEY_T, 0x0162, 
GDK_KEY_cedilla, GDK_KEY_c, 0x00E7, 
GDK_KEY_cedilla, GDK_KEY_d, 0x1E11, 
GDK_KEY_cedilla, GDK_KEY_e, 0x0229, 
GDK_KEY_cedilla, GDK_KEY_g, 0x0123, 
GDK_KEY_cedilla, GDK_KEY_h, 0x1E29, 
GDK_KEY_cedilla, GDK_KEY_k, 0x0137, 
GDK_KEY_cedilla, GDK_KEY_l, 0x013C, 
GDK_KEY_cedilla, GDK_KEY_n, 0x0146, 
GDK_KEY_cedilla, GDK_KEY_r, 0x0157, 
GDK_KEY_cedilla, GDK_KEY_s, 0x015F, 
GDK_KEY_cedilla, GDK_KEY_t, 0x0163, 
GDK_KEY_division, 0x2395, 0x2339, 
GDK_KEY_breve, GDK_KEY_G, 0x011E, 
GDK_KEY_breve, GDK_KEY_g, 0x011F, 
0x05B4, GDK_KEY_hebrew_yod, 0xFB1D, 
0x05B7, 0x05F2, 0xFB1F, 
0x05B7, GDK_KEY_hebrew_aleph, 0xFB2E, 
0x05B8, GDK_KEY_hebrew_aleph, 0xFB2F, 
0x05B9, GDK_KEY_hebrew_waw, 0xFB4B, 
0x05BC, GDK_KEY_hebrew_aleph, 0xFB30, 
0x05BC, GDK_KEY_hebrew_beth, 0xFB31, 
0x05BC, GDK_KEY_hebrew_gimmel, 0xFB32, 
0x05BC, GDK_KEY_hebrew_daleth, 0xFB33, 
0x05BC, GDK_KEY_hebrew_he, 0xFB34, 
0x05BC, GDK_KEY_hebrew_waw, 0xFB35, 
0x05BC, GDK_KEY_hebrew_zayin, 0xFB36, 
0x05BC, GDK_KEY_hebrew_teth, 0xFB38, 
0x05BC, GDK_KEY_hebrew_yod, 0xFB39, 
0x05BC, GDK_KEY_hebrew_finalkaph, 0xFB3A, 
0x05BC, GDK_KEY_hebrew_kaph, 0xFB3B, 
0x05BC, GDK_KEY_hebrew_lamed, 0xFB3C, 
0x05BC, GDK_KEY_hebrew_mem, 0xFB3E, 
0x05BC, GDK_KEY_hebrew_nun, 0xFB40, 
0x05BC, GDK_KEY_hebrew_samekh, 0xFB41, 
0x05BC, GDK_KEY_hebrew_finalpe, 0xFB43, 
0x05BC, GDK_KEY_hebrew_pe, 0xFB44, 
0x05BC, GDK_KEY_hebrew_zadi, 0xFB46, 
0x05BC, GDK_KEY_hebrew_qoph, 0xFB47, 
0x05BC, GDK_KEY_hebrew_resh, 0xFB48, 
0x05BC, GDK_KEY_hebrew_shin, 0xFB49, 
0x05BC, GDK_KEY_hebrew_taw, 0xFB4A, 
0x05BF, GDK_KEY_hebrew_beth, 0xFB4C, 
0x05BF, GDK_KEY_hebrew_kaph, 0xFB4D, 
0x05BF, GDK_KEY_hebrew_pe, 0xFB4E, 
0x05C1, GDK_KEY_hebrew_shin, 0xFB2A, 
0x05C1, 0xFB49, 0xFB2C, 
0x05C2, GDK_KEY_hebrew_shin, 0xFB2B, 
0x05C2, 0xFB49, 0xFB2D, 
0x0653, GDK_KEY_Arabic_alef, 0x0622, 
0x0654, GDK_KEY_Arabic_alef, 0x0623, 
0x0654, GDK_KEY_Arabic_waw, 0x0624, 
0x0654, GDK_KEY_Arabic_yeh, 0x0626, 
0x0654, 0x06C1, 0x06C2, 
0x0654, 0x06D2, 0x06D3, 
0x0654, 0x06D5, 0x06C0, 
0x0655, GDK_KEY_Arabic_alef, 0x0625, 
GDK_KEY_Cyrillic_pe, GDK_KEY_Cyrillic_a, 0x00A7, 
GDK_KEY_Cyrillic_IE, GDK_KEY_equal, 0x20AC, 
GDK_KEY_Cyrillic_EN, GDK_KEY_Cyrillic_o, 0x2116, 
GDK_KEY_Cyrillic_EN, GDK_KEY_Cyrillic_O, 0x2116, 
GDK_KEY_Cyrillic_ES, GDK_KEY_equal, 0x20AC, 
GDK_KEY_Greek_ALPHA, GDK_KEY_apostrophe, 0x0386, 
GDK_KEY_Greek_EPSILON, GDK_KEY_apostrophe, 0x0388, 
GDK_KEY_Greek_ETA, GDK_KEY_apostrophe, 0x0389, 
GDK_KEY_Greek_IOTA, GDK_KEY_quotedbl, 0x03AA, 
GDK_KEY_Greek_IOTA, GDK_KEY_apostrophe, 0x038A, 
GDK_KEY_Greek_OMICRON, GDK_KEY_apostrophe, 0x038C, 
GDK_KEY_Greek_UPSILON, GDK_KEY_quotedbl, 0x03AB, 
GDK_KEY_Greek_UPSILON, GDK_KEY_apostrophe, 0x038E, 
GDK_KEY_Greek_OMEGA, GDK_KEY_apostrophe, 0x038F, 
GDK_KEY_Greek_alpha, GDK_KEY_apostrophe, 0x03AC, 
GDK_KEY_Greek_epsilon, GDK_KEY_apostrophe, 0x03AD, 
GDK_KEY_Greek_eta, GDK_KEY_apostrophe, 0x03AE, 
GDK_KEY_Greek_iota, GDK_KEY_quotedbl, 0x03CA, 
GDK_KEY_Greek_iota, GDK_KEY_Greek_alphaaccent, 0x1FB4, 
GDK_KEY_Greek_iota, GDK_KEY_Greek_etaaccent, 0x1FC4, 
GDK_KEY_Greek_iota, GDK_KEY_Greek_omegaaccent, 0x1FF4, 
GDK_KEY_Greek_iota, GDK_KEY_Greek_ALPHA, 0x1FBC, 
GDK_KEY_Greek_iota, GDK_KEY_Greek_ETA, 0x1FCC, 
GDK_KEY_Greek_iota, GDK_KEY_Greek_OMEGA, 0x1FFC, 
GDK_KEY_Greek_iota, GDK_KEY_Greek_alpha, 0x1FB3, 
GDK_KEY_Greek_iota, GDK_KEY_Greek_eta, 0x1FC3, 
GDK_KEY_Greek_iota, GDK_KEY_Greek_omega, 0x1FF3, 
GDK_KEY_Greek_iota, 0x1F00, 0x1F80, 
GDK_KEY_Greek_iota, 0x1F01, 0x1F81, 
GDK_KEY_Greek_iota, 0x1F02, 0x1F82, 
GDK_KEY_Greek_iota, 0x1F03, 0x1F83, 
GDK_KEY_Greek_iota, 0x1F04, 0x1F84, 
GDK_KEY_Greek_iota, 0x1F05, 0x1F85, 
GDK_KEY_Greek_iota, 0x1F06, 0x1F86, 
GDK_KEY_Greek_iota, 0x1F07, 0x1F87, 
GDK_KEY_Greek_iota, 0x1F08, 0x1F88, 
GDK_KEY_Greek_iota, 0x1F09, 0x1F89, 
GDK_KEY_Greek_iota, 0x1F0A, 0x1F8A, 
GDK_KEY_Greek_iota, 0x1F0B, 0x1F8B, 
GDK_KEY_Greek_iota, 0x1F0C, 0x1F8C, 
GDK_KEY_Greek_iota, 0x1F0D, 0x1F8D, 
GDK_KEY_Greek_iota, 0x1F0E, 0x1F8E, 
GDK_KEY_Greek_iota, 0x1F0F, 0x1F8F, 
GDK_KEY_Greek_iota, 0x1F20, 0x1F90, 
GDK_KEY_Greek_iota, 0x1F21, 0x1F91, 
GDK_KEY_Greek_iota, 0x1F22, 0x1F92, 
GDK_KEY_Greek_iota, 0x1F23, 0x1F93, 
GDK_KEY_Greek_iota, 0x1F24, 0x1F94, 
GDK_KEY_Greek_iota, 0x1F25, 0x1F95, 
GDK_KEY_Greek_iota, 0x1F26, 0x1F96, 
GDK_KEY_Greek_iota, 0x1F27, 0x1F97, 
GDK_KEY_Greek_iota, 0x1F28, 0x1F98, 
GDK_KEY_Greek_iota, 0x1F29, 0x1F99, 
GDK_KEY_Greek_iota, 0x1F2A, 0x1F9A, 
GDK_KEY_Greek_iota, 0x1F2B, 0x1F9B, 
GDK_KEY_Greek_iota, 0x1F2C, 0x1F9C, 
GDK_KEY_Greek_iota, 0x1F2D, 0x1F9D, 
GDK_KEY_Greek_iota, 0x1F2E, 0x1F9E, 
GDK_KEY_Greek_iota, 0x1F2F, 0x1F9F, 
GDK_KEY_Greek_iota, 0x1F60, 0x1FA0, 
GDK_KEY_Greek_iota, 0x1F61, 0x1FA1, 
GDK_KEY_Greek_iota, 0x1F62, 0x1FA2, 
GDK_KEY_Greek_iota, 0x1F63, 0x1FA3, 
GDK_KEY_Greek_iota, 0x1F64, 0x1FA4, 
GDK_KEY_Greek_iota, 0x1F65, 0x1FA5, 
GDK_KEY_Greek_iota, 0x1F66, 0x1FA6, 
GDK_KEY_Greek_iota, 0x1F67, 0x1FA7, 
GDK_KEY_Greek_iota, 0x1F68, 0x1FA8, 
GDK_KEY_Greek_iota, 0x1F69, 0x1FA9, 
GDK_KEY_Greek_iota, 0x1F6A, 0x1FAA, 
GDK_KEY_Greek_iota, 0x1F6B, 0x1FAB, 
GDK_KEY_Greek_iota, 0x1F6C, 0x1FAC, 
GDK_KEY_Greek_iota, 0x1F6D, 0x1FAD, 
GDK_KEY_Greek_iota, 0x1F6E, 0x1FAE, 
GDK_KEY_Greek_iota, 0x1F6F, 0x1FAF, 
GDK_KEY_Greek_iota, 0x1F70, 0x1FB2, 
GDK_KEY_Greek_iota, 0x1F74, 0x1FC2, 
GDK_KEY_Greek_iota, 0x1F7C, 0x1FF2, 
GDK_KEY_Greek_iota, 0x1FB6, 0x1FB7, 
GDK_KEY_Greek_iota, 0x1FC6, 0x1FC7, 
GDK_KEY_Greek_iota, 0x1FF6, 0x1FF7, 
GDK_KEY_Greek_omicron, GDK_KEY_apostrophe, 0x03CC, 
GDK_KEY_Greek_upsilon, GDK_KEY_quotedbl, 0x03CB, 
GDK_KEY_Greek_upsilon, GDK_KEY_apostrophe, 0x03CD, 
GDK_KEY_Greek_omega, GDK_KEY_apostrophe, 0x03CE, 
GDK_KEY_lessthanequal, 0x0338, 0x2270, 
GDK_KEY_greaterthanequal, 0x0338, 0x2271, 
GDK_KEY_approximate, 0x0338, 0x2247, 
GDK_KEY_identical, 0x0338, 0x2262, 
GDK_KEY_includedin, 0x0338, 0x2284, 
GDK_KEY_includes, 0x0338, 0x2285, 
0x093C, 0x0915, 0x0958, 
0x093C, 0x0916, 0x0959, 
0x093C, 0x0917, 0x095A, 
0x093C, 0x091C, 0x095B, 
0x093C, 0x0921, 0x095C, 
0x093C, 0x0922, 0x095D, 
0x093C, 0x0928, 0x0929, 
0x093C, 0x092B, 0x095E, 
0x093C, 0x092F, 0x095F, 
0x093C, 0x0930, 0x0931, 
0x093C, 0x0933, 0x0934, 
0x09BC, 0x09A1, 0x09DC, 
0x09BC, 0x09A2, 0x09DD, 
0x09BC, 0x09AF, 0x09DF, 
0x09C7, 0x09BE, 0x09CB, 
0x09C7, 0x09D7, 0x09CC, 
0x0A3C, 0x0A16, 0x0A59, 
0x0A3C, 0x0A17, 0x0A5A, 
0x0A3C, 0x0A1C, 0x0A5B, 
0x0A3C, 0x0A2B, 0x0A5E, 
0x0A3C, 0x0A32, 0x0A33, 
0x0A3C, 0x0A38, 0x0A36, 
0x0B3C, 0x0B21, 0x0B5C, 
0x0B3C, 0x0B22, 0x0B5D, 
0x0B47, 0x0B3E, 0x0B4B, 
0x0B47, 0x0B56, 0x0B48, 
0x0B47, 0x0B57, 0x0B4C, 
GDK_KEY_leftcaret, 0x0338, 0x226E, 
GDK_KEY_rightcaret, 0x0338, 0x226F, 
GDK_KEY_underbar, GDK_KEY_parenleft, 0x208D, 
GDK_KEY_underbar, GDK_KEY_parenright, 0x208E, 
GDK_KEY_underbar, GDK_KEY_plus, 0x208A, 
GDK_KEY_underbar, GDK_KEY_0, 0x2080, 
GDK_KEY_underbar, GDK_KEY_1, 0x2081, 
GDK_KEY_underbar, GDK_KEY_2, 0x2082, 
GDK_KEY_underbar, GDK_KEY_3, 0x2083, 
GDK_KEY_underbar, GDK_KEY_4, 0x2084, 
GDK_KEY_underbar, GDK_KEY_5, 0x2085, 
GDK_KEY_underbar, GDK_KEY_6, 0x2086, 
GDK_KEY_underbar, GDK_KEY_7, 0x2087, 
GDK_KEY_underbar, GDK_KEY_8, 0x2088, 
GDK_KEY_underbar, GDK_KEY_9, 0x2089, 
GDK_KEY_underbar, GDK_KEY_equal, 0x208C, 
0x0BC6, 0x0BBE, 0x0BCA, 
0x0BC6, 0x0BD7, 0x0BCC, 
GDK_KEY_underbar, 0x2212, 0x208B, 
GDK_KEY_underbar, GDK_KEY_KP_Space, 0x2082, 
GDK_KEY_underbar, GDK_KEY_KP_Add, 0x208A, 
GDK_KEY_underbar, GDK_KEY_KP_0, 0x2080, 
GDK_KEY_underbar, GDK_KEY_KP_1, 0x2081, 
GDK_KEY_underbar, GDK_KEY_KP_2, 0x2082, 
GDK_KEY_underbar, GDK_KEY_KP_3, 0x2083, 
GDK_KEY_underbar, GDK_KEY_KP_4, 0x2084, 
GDK_KEY_underbar, GDK_KEY_KP_5, 0x2085, 
GDK_KEY_underbar, GDK_KEY_KP_6, 0x2086, 
GDK_KEY_underbar, GDK_KEY_KP_7, 0x2087, 
GDK_KEY_underbar, GDK_KEY_KP_8, 0x2088, 
GDK_KEY_underbar, GDK_KEY_KP_9, 0x2089, 
GDK_KEY_underbar, GDK_KEY_KP_Equal, 0x208C, 
0x0BC7, 0x0BBE, 0x0BCB, 
0x0BD7, 0x0B92, 0x0B94, 
GDK_KEY_rightshoe, 0x0338, 0x2285, 
GDK_KEY_leftshoe, 0x0338, 0x2284, 
GDK_KEY_righttack, 0x0338, 0x22AC, 
0x0C46, 0x0C56, 0x0C48, 
0x0CBF, 0x0CD5, 0x0CC0, 
0x0CC6, 0x0CC2, 0x0CCA, 
0x0CC6, 0x0CD5, 0x0CC7, 
0x0CC6, 0x0CD6, 0x0CC8, 
0x0CCA, 0x0CD5, 0x0CCB, 
0x0D46, 0x0D3E, 0x0D4A, 
0x0D46, 0x0D57, 0x0D4C, 
0x0D47, 0x0D3E, 0x0D4B, 
0x0DD9, 0x0DCA, 0x0DDA, 
0x0DD9, 0x0DCF, 0x0DDC, 
0x0DD9, 0x0DDF, 0x0DDE, 
0x0DDC, 0x0DCA, 0x0DDD, 
0x0F71, 0x0F72, 0x0F73, 
0x0F71, 0x0F74, 0x0F75, 
0x0F71, 0x0F80, 0x0F81, 
0x0F90, 0x0FB5, 0x0FB9, 
0x0F92, 0x0FB7, 0x0F93, 
0x0F9C, 0x0FB7, 0x0F9D, 
0x0FA1, 0x0FB7, 0x0FA2, 
0x0FA6, 0x0FB7, 0x0FA7, 
0x0FAB, 0x0FB7, 0x0FAC, 
0x0FB2, 0x0F80, 0x0F76, 
0x0FB3, 0x0F80, 0x0F78, 
0x0FB5, 0x0F40, 0x0F69, 
0x0FB7, 0x0F42, 0x0F43, 
0x0FB7, 0x0F4C, 0x0F4D, 
0x0FB7, 0x0F51, 0x0F52, 
0x0FB7, 0x0F56, 0x0F57, 
0x0FB7, 0x0F5B, 0x0F5C, 
0x102E, 0x1025, 0x1026, 
0x1100, 0x1100, 0x1101, 
0x1102, 0x1100, 0x1113, 
0x1102, 0x1102, 0x1114, 
0x1102, 0x1103, 0x1115, 
0x1102, 0x1107, 0x1116, 
0x1103, 0x1100, 0x1117, 
0x1103, 0x1103, 0x1104, 
0x1105, 0x1102, 0x1118, 
0x1105, 0x1105, 0x1119, 
0x1105, 0x110B, 0x111B, 
0x1105, 0x1112, 0x111A, 
0x1106, 0x1107, 0x111C, 
0x1106, 0x110B, 0x111D, 
0x1107, 0x1100, 0x111E, 
0x1107, 0x1102, 0x111F, 
0x1107, 0x1103, 0x1120, 
0x1107, 0x1107, 0x1108, 
0x1107, 0x1109, 0x1121, 
0x1107, 0x110A, 0x1125, 
0x1107, 0x110B, 0x112B, 
0x1107, 0x110C, 0x1127, 
0x1107, 0x110E, 0x1128, 
0x1107, 0x1110, 0x1129, 
0x1107, 0x1111, 0x112A, 
0x1107, 0x112B, 0x112C, 
0x1107, 0x112D, 0x1122, 
0x1107, 0x112F, 0x1123, 
0x1107, 0x1132, 0x1124, 
0x1107, 0x1136, 0x1126, 
0x1108, 0x110B, 0x112C, 
0x1109, 0x1100, 0x112D, 
0x1109, 0x1102, 0x112E, 
0x1109, 0x1103, 0x112F, 
0x1109, 0x1105, 0x1130, 
0x1109, 0x1106, 0x1131, 
0x1109, 0x1107, 0x1132, 
0x1109, 0x1109, 0x110A, 
0x1109, 0x110A, 0x1134, 
0x1109, 0x110B, 0x1135, 
0x1109, 0x110C, 0x1136, 
0x1109, 0x110E, 0x1137, 
0x1109, 0x110F, 0x1138, 
0x1109, 0x1110, 0x1139, 
0x1109, 0x1111, 0x113A, 
0x1109, 0x1112, 0x113B, 
0x1109, 0x111E, 0x1133, 
0x110A, 0x1109, 0x1134, 
0x110B, 0x1100, 0x1141, 
0x110B, 0x1103, 0x1142, 
0x110B, 0x1106, 0x1143, 
0x110B, 0x1107, 0x1144, 
0x110B, 0x1109, 0x1145, 
0x110B, 0x110B, 0x1147, 
0x110B, 0x110C, 0x1148, 
0x110B, 0x110E, 0x1149, 
0x110B, 0x1110, 0x114A, 
0x110B, 0x1111, 0x114B, 
0x110B, 0x1140, 0x1146, 
0x110C, 0x110B, 0x114D, 
0x110C, 0x110C, 0x110D, 
0x110E, 0x110F, 0x1152, 
0x110E, 0x1112, 0x1153, 
0x1111, 0x1107, 0x1156, 
0x1111, 0x110B, 0x1157, 
0x1112, 0x1112, 0x1158, 
0x1121, 0x1100, 0x1122, 
0x1121, 0x1103, 0x1123, 
0x1121, 0x1107, 0x1124, 
0x1121, 0x1109, 0x1125, 
0x1121, 0x110C, 0x1126, 
0x1132, 0x1100, 0x1133, 
0x113C, 0x113C, 0x113D, 
0x113E, 0x113E, 0x113F, 
0x114E, 0x114E, 0x114F, 
0x1150, 0x1150, 0x1151, 
0x1161, 0x1169, 0x1176, 
0x1161, 0x116E, 0x1177, 
0x1161, 0x1175, 0x1162, 
0x1163, 0x1169, 0x1178, 
0x1163, 0x116D, 0x1179, 
0x1163, 0x1175, 0x1164, 
0x1165, 0x1169, 0x117A, 
0x1165, 0x116E, 0x117B, 
0x1165, 0x1173, 0x117C, 
0x1165, 0x1175, 0x1166, 
0x1167, 0x1169, 0x117D, 
0x1167, 0x116E, 0x117E, 
0x1167, 0x1175, 0x1168, 
0x1169, 0x1161, 0x116A, 
0x1169, 0x1162, 0x116B, 
0x1169, 0x1165, 0x117F, 
0x1169, 0x1166, 0x1180, 
0x1169, 0x1168, 0x1181, 
0x1169, 0x1169, 0x1182, 
0x1169, 0x116E, 0x1183, 
0x1169, 0x1175, 0x116C, 
0x116A, 0x1175, 0x116B, 
0x116D, 0x1163, 0x1184, 
0x116D, 0x1164, 0x1185, 
0x116D, 0x1167, 0x1186, 
0x116D, 0x1169, 0x1187, 
0x116D, 0x1175, 0x1188, 
0x116E, 0x1161, 0x1189, 
0x116E, 0x1162, 0x118A, 
0x116E, 0x1165, 0x116F, 
0x116E, 0x1166, 0x1170, 
0x116E, 0x1168, 0x118C, 
0x116E, 0x116E, 0x118D, 
0x116E, 0x1175, 0x1171, 
0x116E, 0x117C, 0x118B, 
0x116F, 0x1173, 0x118B, 
0x116F, 0x1175, 0x1170, 
0x1172, 0x1161, 0x118E, 
0x1172, 0x1165, 0x118F, 
0x1172, 0x1166, 0x1190, 
0x1172, 0x1167, 0x1191, 
0x1172, 0x1168, 0x1192, 
0x1172, 0x116E, 0x1193, 
0x1172, 0x1175, 0x1194, 
0x1173, 0x116E, 0x1195, 
0x1173, 0x1173, 0x1196, 
0x1173, 0x1175, 0x1174, 
0x1174, 0x116E, 0x1197, 
0x1175, 0x1161, 0x1198, 
0x1175, 0x1163, 0x1199, 
0x1175, 0x1169, 0x119A, 
0x1175, 0x116E, 0x119B, 
0x1175, 0x1173, 0x119C, 
0x1175, 0x119E, 0x119D, 
0x119E, 0x1165, 0x119F, 
0x119E, 0x116E, 0x11A0, 
0x119E, 0x1175, 0x11A1, 
0x119E, 0x119E, 0x11A2, 
0x11A8, 0x11A8, 0x11A9, 
0x11A8, 0x11AF, 0x11C3, 
0x11A8, 0x11BA, 0x11AA, 
0x11A8, 0x11E7, 0x11C4, 
0x11AA, 0x11A8, 0x11C4, 
0x11AB, 0x11A8, 0x11C5, 
0x11AB, 0x11AE, 0x11C6, 
0x11AB, 0x11BA, 0x11C7, 
0x11AB, 0x11BD, 0x11AC, 
0x11AB, 0x11C0, 0x11C9, 
0x11AB, 0x11C2, 0x11AD, 
0x11AB, 0x11EB, 0x11C8, 
0x11AE, 0x11A8, 0x11CA, 
0x11AE, 0x11AF, 0x11CB, 
0x11AF, 0x11A8, 0x11B0, 
0x11AF, 0x11AA, 0x11CC, 
0x11AF, 0x11AB, 0x11CD, 
0x11AF, 0x11AE, 0x11CE, 
0x11AF, 0x11AF, 0x11D0, 
0x11AF, 0x11B7, 0x11B1, 
0x11AF, 0x11B8, 0x11B2, 
0x11AF, 0x11B9, 0x11D3, 
0x11AF, 0x11BA, 0x11B3, 
0x11AF, 0x11BB, 0x11D6, 
0x11AF, 0x11BF, 0x11D8, 
0x11AF, 0x11C0, 0x11B4, 
0x11AF, 0x11C1, 0x11B5, 
0x11AF, 0x11C2, 0x11B6, 
0x11AF, 0x11DA, 0x11D1, 
0x11AF, 0x11DD, 0x11D2, 
0x11AF, 0x11E5, 0x11D4, 
0x11AF, 0x11E6, 0x11D5, 
0x11AF, 0x11EB, 0x11D7, 
0x11AF, 0x11F9, 0x11D9, 
0x11B0, 0x11BA, 0x11CC, 
0x11B1, 0x11A8, 0x11D1, 
0x11B1, 0x11BA, 0x11D2, 
0x11B2, 0x11BA, 0x11D3, 
0x11B2, 0x11BC, 0x11D5, 
0x11B2, 0x11C2, 0x11D4, 
0x11B3, 0x11BA, 0x11D6, 
0x11B7, 0x11A8, 0x11DA, 
0x11B7, 0x11AF, 0x11DB, 
0x11B7, 0x11B8, 0x11DC, 
0x11B7, 0x11BA, 0x11DD, 
0x11B7, 0x11BB, 0x11DE, 
0x11B7, 0x11BC, 0x11E2, 
0x11B7, 0x11BE, 0x11E0, 
0x11B7, 0x11C2, 0x11E1, 
0x11B7, 0x11EB, 0x11DF, 
0x11B8, 0x11AF, 0x11E3, 
0x11B8, 0x11BA, 0x11B9, 
0x11B8, 0x11BC, 0x11E6, 
0x11B8, 0x11C1, 0x11E4, 
0x11B8, 0x11C2, 0x11E5, 
0x11BA, 0x11A8, 0x11E7, 
0x11BA, 0x11AE, 0x11E8, 
0x11BA, 0x11AF, 0x11E9, 
0x11BA, 0x11B8, 0x11EA, 
0x11BA, 0x11BA, 0x11BB, 
0x11BC, 0x11A8, 0x11EC, 
0x11BC, 0x11A9, 0x11ED, 
0x11BC, 0x11BC, 0x11EE, 
0x11BC, 0x11BF, 0x11EF, 
0x11C1, 0x11B8, 0x11F3, 
0x11C1, 0x11BC, 0x11F4, 
0x11C2, 0x11AB, 0x11F5, 
0x11C2, 0x11AF, 0x11F6, 
0x11C2, 0x11B7, 0x11F7, 
0x11C2, 0x11B8, 0x11F8, 
0x11CE, 0x11C2, 0x11CF, 
0x11DD, 0x11BA, 0x11DE, 
0x11EC, 0x11A8, 0x11ED, 
0x11F0, 0x11BA, 0x11F1, 
0x11F0, 0x11EB, 0x11F2, 
0x1FBF, GDK_KEY_apostrophe, 0x1FCE, 
0x1FBF, GDK_KEY_grave, 0x1FCD, 
0x1FBF, GDK_KEY_asciitilde, 0x1FCF, 
0x1FBF, GDK_KEY_acute, 0x1FCE, 
0x1FBF, GDK_KEY_dead_grave, 0x1FCD, 
0x1FBF, GDK_KEY_dead_acute, 0x1FCE, 
0x1FBF, GDK_KEY_dead_tilde, 0x1FCF, 
0x1FFE, GDK_KEY_apostrophe, 0x1FDE, 
0x1FFE, GDK_KEY_grave, 0x1FDD, 
0x1FFE, GDK_KEY_asciitilde, 0x1FDF, 
0x1FFE, GDK_KEY_acute, 0x1FDE, 
0x1FFE, GDK_KEY_dead_grave, 0x1FDD, 
0x1FFE, GDK_KEY_dead_acute, 0x1FDE, 
0x1FFE, GDK_KEY_dead_tilde, 0x1FDF, 
0x2190, GDK_KEY_bar, 0x2345, 
0x2190, 0x2395, 0x2347, 
0x2191, GDK_KEY_minus, 0x234F, 
0x2191, 0x2395, 0x2350, 
0x2192, GDK_KEY_bar, 0x2346, 
0x2192, 0x2395, 0x2348, 
0x2193, GDK_KEY_minus, 0x2356, 
0x2193, 0x2395, 0x2357, 
0x2203, 0x0338, 0x2204, 
0x2206, GDK_KEY_underscore, 0x2359, 
0x2206, GDK_KEY_bar, 0x234B, 
0x2206, 0x2395, 0x234D, 
0x2207, GDK_KEY_bar, 0x2352, 
0x2207, GDK_KEY_asciitilde, 0x236B, 
0x2207, GDK_KEY_diaeresis, 0x2362, 
0x2207, 0x2395, 0x2354, 
0x2208, 0x0338, 0x2209, 
0x220A, GDK_KEY_underscore, 0x2377, 
0x220B, 0x0338, 0x220C, 
0x2218, GDK_KEY_underscore, 0x235B, 
0x2218, GDK_KEY_diaeresis, 0x2364, 
0x2218, 0x2229, 0x235D, 
0x2218, GDK_KEY_uptack, 0x2355, 
0x2218, GDK_KEY_downtack, 0x234E, 
0x2218, 0x2395, 0x233B, 
0x2218, GDK_KEY_emopencircle, 0x233E, 
0x2223, 0x0338, 0x2224, 
0x2225, 0x0338, 0x2226, 
0x2227, GDK_KEY_asciitilde, 0x2372, 
0x2227, 0x2228, 0x22C4, 
0x2227, 0x2395, 0x2353, 
0x2228, GDK_KEY_asciitilde, 0x2371, 
0x2228, 0x2227, 0x22C4, 
0x2228, 0x2395, 0x234C, 
0x2229, 0x2218, 0x235D, 
GDK_KEY_union, GDK_KEY_bar, 0x2366, 
0x223C, 0x0338, 0x2241, 
0x2243, 0x0338, 0x2244, 
0x2248, 0x0338, 0x2249, 
0x224D, 0x0338, 0x226D, 
0x2260, GDK_KEY_underscore, 0x2262, 
0x2260, 0x2395, 0x236F, 
0x2272, 0x0338, 0x2274, 
0x2273, 0x0338, 0x2275, 
0x2276, 0x0338, 0x2278, 
0x2277, 0x0338, 0x2279, 
0x227A, 0x0338, 0x2280, 
0x227B, 0x0338, 0x2281, 
0x227C, 0x0338, 0x22E0, 
0x227D, 0x0338, 0x22E1, 
0x2282, GDK_KEY_underscore, 0x2286, 
0x2282, GDK_KEY_bar, 0x2367, 
0x2283, GDK_KEY_underscore, 0x2287, 
0x2286, 0x0338, 0x2288, 
0x2287, 0x0338, 0x2289, 
0x2291, 0x0338, 0x22E2, 
0x2292, 0x0338, 0x22E3, 
GDK_KEY_uptack, GDK_KEY_diaeresis, 0x2361, 
GDK_KEY_uptack, GDK_KEY_macron, 0x2351, 
GDK_KEY_uptack, 0x2218, 0x2355, 
GDK_KEY_uptack, GDK_KEY_downtack, 0x2336, 
GDK_KEY_downtack, GDK_KEY_underscore, 0x234A, 
GDK_KEY_downtack, 0x2218, 0x234E, 
GDK_KEY_downtack, GDK_KEY_uptack, 0x2336, 
0x22A8, 0x0338, 0x22AD, 
0x22A9, 0x0338, 0x22AE, 
0x22AB, 0x0338, 0x22AF, 
0x22B2, 0x0338, 0x22EA, 
0x22B3, 0x0338, 0x22EB, 
0x22B4, 0x0338, 0x22EC, 
0x22B5, 0x0338, 0x22ED, 
0x22C4, GDK_KEY_underscore, 0x235A, 
0x22C4, 0x2395, 0x233A, 
0x2373, GDK_KEY_underscore, 0x2378, 
0x2375, GDK_KEY_underscore, 0x2379, 
0x237A, GDK_KEY_underscore, 0x2376, 
0x2395, GDK_KEY_apostrophe, 0x235E, 
0x2395, GDK_KEY_slash, 0x2341, 
0x2395, GDK_KEY_colon, 0x2360, 
0x2395, GDK_KEY_less, 0x2343, 
0x2395, GDK_KEY_equal, 0x2338, 
0x2395, GDK_KEY_greater, 0x2344, 
0x2395, GDK_KEY_question, 0x2370, 
0x2395, GDK_KEY_backslash, 0x2342, 
0x2395, GDK_KEY_division, 0x2339, 
0x2395, 0x2190, 0x2347, 
0x2395, 0x2191, 0x2350, 
0x2395, 0x2192, 0x2348, 
0x2395, 0x2193, 0x2357, 
0x2395, 0x2206, 0x234D, 
0x2395, 0x2207, 0x2354, 
0x2395, 0x2218, 0x233B, 
0x2395, 0x2227, 0x2353, 
0x2395, 0x2228, 0x234C, 
0x2395, 0x2260, 0x236F, 
0x2395, 0x22C4, 0x233A, 
0x2395, GDK_KEY_emopencircle, 0x233C, 
GDK_KEY_emopencircle, GDK_KEY_asterisk, 0x235F, 
GDK_KEY_emopencircle, GDK_KEY_minus, 0x2296, 
GDK_KEY_emopencircle, GDK_KEY_period, 0x2299, 
GDK_KEY_emopencircle, GDK_KEY_backslash, 0x2349, 
GDK_KEY_emopencircle, GDK_KEY_underscore, 0x235C, 
GDK_KEY_emopencircle, GDK_KEY_bar, 0x233D, 
GDK_KEY_emopencircle, GDK_KEY_diaeresis, 0x2365, 
GDK_KEY_emopencircle, 0x2218, 0x233E, 
GDK_KEY_emopencircle, 0x2395, 0x233C, 
0x2ADD, 0x0338, 0x2ADC, 
GDK_KEY_KP_Divide, GDK_KEY_D, 0x0110, 
GDK_KEY_KP_Divide, GDK_KEY_G, 0x01E4, 
GDK_KEY_KP_Divide, GDK_KEY_H, 0x0126, 
GDK_KEY_KP_Divide, GDK_KEY_I, 0x0197, 
GDK_KEY_KP_Divide, GDK_KEY_L, 0x0141, 
GDK_KEY_KP_Divide, GDK_KEY_O, 0x00D8, 
GDK_KEY_KP_Divide, GDK_KEY_T, 0x0166, 
GDK_KEY_KP_Divide, GDK_KEY_Z, 0x01B5, 
GDK_KEY_KP_Divide, GDK_KEY_b, 0x0180, 
GDK_KEY_KP_Divide, GDK_KEY_d, 0x0111, 
GDK_KEY_KP_Divide, GDK_KEY_g, 0x01E5, 
GDK_KEY_KP_Divide, GDK_KEY_h, 0x0127, 
GDK_KEY_KP_Divide, GDK_KEY_i, 0x0268, 
GDK_KEY_KP_Divide, GDK_KEY_l, 0x0142, 
GDK_KEY_KP_Divide, GDK_KEY_o, 0x00F8, 
GDK_KEY_KP_Divide, GDK_KEY_t, 0x0167, 
GDK_KEY_KP_Divide, GDK_KEY_z, 0x01B6, 
GDK_KEY_KP_Divide, 0x0294, 0x02A1, 
GDK_KEY_KP_Divide, 0x04AE, 0x04B0, 
GDK_KEY_KP_Divide, 0x04AF, 0x04B1, 
GDK_KEY_KP_Divide, GDK_KEY_Cyrillic_ghe, 0x0493, 
GDK_KEY_KP_Divide, GDK_KEY_Cyrillic_ka, 0x049F, 
GDK_KEY_KP_Divide, GDK_KEY_Cyrillic_GHE, 0x0492, 
GDK_KEY_KP_Divide, GDK_KEY_Cyrillic_KA, 0x049E, 
GDK_KEY_KP_Divide, GDK_KEY_leftarrow, 0x219A, 
GDK_KEY_KP_Divide, GDK_KEY_rightarrow, 0x219B, 
GDK_KEY_KP_Divide, 0x2194, 0x21AE, 
GDK_KEY_KP_Equal, 0x0338, 0x2260, 
GDK_KEY_exclam, GDK_KEY_plus, GDK_KEY_O, 0x1EE2, 
GDK_KEY_exclam, GDK_KEY_plus, GDK_KEY_U, 0x1EF0, 
GDK_KEY_exclam, GDK_KEY_plus, GDK_KEY_o, 0x1EE3, 
GDK_KEY_exclam, GDK_KEY_plus, GDK_KEY_u, 0x1EF1, 
GDK_KEY_exclam, GDK_KEY_dead_horn, GDK_KEY_O, 0x1EE2, 
GDK_KEY_exclam, GDK_KEY_dead_horn, GDK_KEY_U, 0x1EF0, 
GDK_KEY_exclam, GDK_KEY_dead_horn, GDK_KEY_o, 0x1EE3, 
GDK_KEY_exclam, GDK_KEY_dead_horn, GDK_KEY_u, 0x1EF1, 
GDK_KEY_quotedbl, GDK_KEY_underscore, GDK_KEY_U, 0x1E7A, 
GDK_KEY_quotedbl, GDK_KEY_underscore, GDK_KEY_u, 0x1E7B, 
GDK_KEY_quotedbl, GDK_KEY_asciitilde, GDK_KEY_O, 0x1E4E, 
GDK_KEY_quotedbl, GDK_KEY_asciitilde, GDK_KEY_o, 0x1E4F, 
GDK_KEY_quotedbl, GDK_KEY_macron, GDK_KEY_U, 0x1E7A, 
GDK_KEY_quotedbl, GDK_KEY_macron, GDK_KEY_u, 0x1E7B, 
GDK_KEY_quotedbl, GDK_KEY_dead_tilde, GDK_KEY_O, 0x1E4E, 
GDK_KEY_quotedbl, GDK_KEY_dead_tilde, GDK_KEY_o, 0x1E4F, 
GDK_KEY_quotedbl, GDK_KEY_dead_macron, GDK_KEY_U, 0x1E7A, 
GDK_KEY_quotedbl, GDK_KEY_dead_macron, GDK_KEY_u, 0x1E7B, 
GDK_KEY_apostrophe, GDK_KEY_quotedbl, GDK_KEY_space, 0x0385, 
GDK_KEY_apostrophe, GDK_KEY_quotedbl, GDK_KEY_I, 0x1E2E, 
GDK_KEY_apostrophe, GDK_KEY_quotedbl, GDK_KEY_U, 0x01D7, 
GDK_KEY_apostrophe, GDK_KEY_quotedbl, GDK_KEY_i, 0x1E2F, 
GDK_KEY_apostrophe, GDK_KEY_quotedbl, GDK_KEY_u, 0x01D8, 
GDK_KEY_apostrophe, GDK_KEY_quotedbl, GDK_KEY_Greek_iota, 0x0390, 
GDK_KEY_apostrophe, GDK_KEY_quotedbl, GDK_KEY_Greek_upsilon, 0x03B0, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F0D, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_EPSILON, 0x1F1D, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F2D, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_IOTA, 0x1F3D, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_OMICRON, 0x1F4D, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_UPSILON, 0x1F5D, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1F6D, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F05, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_epsilon, 0x1F15, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F25, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_iota, 0x1F35, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_omicron, 0x1F45, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_upsilon, 0x1F55, 
GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1F65, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F0C, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_EPSILON, 0x1F1C, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F2C, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_IOTA, 0x1F3C, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_OMICRON, 0x1F4C, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1F6C, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F04, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_epsilon, 0x1F14, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F24, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_iota, 0x1F34, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_omicron, 0x1F44, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_upsilon, 0x1F54, 
GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1F64, 
GDK_KEY_apostrophe, GDK_KEY_plus, GDK_KEY_O, 0x1EDA, 
GDK_KEY_apostrophe, GDK_KEY_plus, GDK_KEY_U, 0x1EE8, 
GDK_KEY_apostrophe, GDK_KEY_plus, GDK_KEY_o, 0x1EDB, 
GDK_KEY_apostrophe, GDK_KEY_plus, GDK_KEY_u, 0x1EE9, 
GDK_KEY_apostrophe, GDK_KEY_slash, GDK_KEY_O, 0x01FE, 
GDK_KEY_apostrophe, GDK_KEY_slash, GDK_KEY_o, 0x01FF, 
GDK_KEY_apostrophe, GDK_KEY_asciicircum, GDK_KEY_A, 0x1EA4, 
GDK_KEY_apostrophe, GDK_KEY_asciicircum, GDK_KEY_E, 0x1EBE, 
GDK_KEY_apostrophe, GDK_KEY_asciicircum, GDK_KEY_O, 0x1ED0, 
GDK_KEY_apostrophe, GDK_KEY_asciicircum, GDK_KEY_a, 0x1EA5, 
GDK_KEY_apostrophe, GDK_KEY_asciicircum, GDK_KEY_e, 0x1EBF, 
GDK_KEY_apostrophe, GDK_KEY_asciicircum, GDK_KEY_o, 0x1ED1, 
GDK_KEY_apostrophe, GDK_KEY_underscore, GDK_KEY_E, 0x1E16, 
GDK_KEY_apostrophe, GDK_KEY_underscore, GDK_KEY_O, 0x1E52, 
GDK_KEY_apostrophe, GDK_KEY_underscore, GDK_KEY_e, 0x1E17, 
GDK_KEY_apostrophe, GDK_KEY_underscore, GDK_KEY_o, 0x1E53, 
GDK_KEY_apostrophe, GDK_KEY_b, GDK_KEY_A, 0x1EAE, 
GDK_KEY_apostrophe, GDK_KEY_b, GDK_KEY_a, 0x1EAF, 
GDK_KEY_apostrophe, GDK_KEY_asciitilde, GDK_KEY_O, 0x1E4C, 
GDK_KEY_apostrophe, GDK_KEY_asciitilde, GDK_KEY_U, 0x1E78, 
GDK_KEY_apostrophe, GDK_KEY_asciitilde, GDK_KEY_o, 0x1E4D, 
GDK_KEY_apostrophe, GDK_KEY_asciitilde, GDK_KEY_u, 0x1E79, 
GDK_KEY_apostrophe, GDK_KEY_macron, GDK_KEY_E, 0x1E16, 
GDK_KEY_apostrophe, GDK_KEY_macron, GDK_KEY_O, 0x1E52, 
GDK_KEY_apostrophe, GDK_KEY_macron, GDK_KEY_e, 0x1E17, 
GDK_KEY_apostrophe, GDK_KEY_macron, GDK_KEY_o, 0x1E53, 
GDK_KEY_apostrophe, GDK_KEY_cedilla, GDK_KEY_C, 0x1E08, 
GDK_KEY_apostrophe, GDK_KEY_cedilla, GDK_KEY_c, 0x1E09, 
GDK_KEY_apostrophe, GDK_KEY_dead_circumflex, GDK_KEY_A, 0x1EA4, 
GDK_KEY_apostrophe, GDK_KEY_dead_circumflex, GDK_KEY_E, 0x1EBE, 
GDK_KEY_apostrophe, GDK_KEY_dead_circumflex, GDK_KEY_O, 0x1ED0, 
GDK_KEY_apostrophe, GDK_KEY_dead_circumflex, GDK_KEY_a, 0x1EA5, 
GDK_KEY_apostrophe, GDK_KEY_dead_circumflex, GDK_KEY_e, 0x1EBF, 
GDK_KEY_apostrophe, GDK_KEY_dead_circumflex, GDK_KEY_o, 0x1ED1, 
GDK_KEY_apostrophe, GDK_KEY_dead_tilde, GDK_KEY_O, 0x1E4C, 
GDK_KEY_apostrophe, GDK_KEY_dead_tilde, GDK_KEY_U, 0x1E78, 
GDK_KEY_apostrophe, GDK_KEY_dead_tilde, GDK_KEY_o, 0x1E4D, 
GDK_KEY_apostrophe, GDK_KEY_dead_tilde, GDK_KEY_u, 0x1E79, 
GDK_KEY_apostrophe, GDK_KEY_dead_macron, GDK_KEY_E, 0x1E16, 
GDK_KEY_apostrophe, GDK_KEY_dead_macron, GDK_KEY_O, 0x1E52, 
GDK_KEY_apostrophe, GDK_KEY_dead_macron, GDK_KEY_e, 0x1E17, 
GDK_KEY_apostrophe, GDK_KEY_dead_macron, GDK_KEY_o, 0x1E53, 
GDK_KEY_apostrophe, GDK_KEY_dead_breve, GDK_KEY_A, 0x1EAE, 
GDK_KEY_apostrophe, GDK_KEY_dead_breve, GDK_KEY_a, 0x1EAF, 
GDK_KEY_apostrophe, GDK_KEY_dead_diaeresis, GDK_KEY_I, 0x1E2E, 
GDK_KEY_apostrophe, GDK_KEY_dead_diaeresis, GDK_KEY_U, 0x01D7, 
GDK_KEY_apostrophe, GDK_KEY_dead_diaeresis, GDK_KEY_i, 0x1E2F, 
GDK_KEY_apostrophe, GDK_KEY_dead_diaeresis, GDK_KEY_u, 0x01D8, 
GDK_KEY_apostrophe, GDK_KEY_dead_diaeresis, GDK_KEY_Greek_iota, 0x0390, 
GDK_KEY_apostrophe, GDK_KEY_dead_diaeresis, GDK_KEY_Greek_upsilon, 0x03B0, 
GDK_KEY_apostrophe, GDK_KEY_dead_abovering, GDK_KEY_A, 0x01FA, 
GDK_KEY_apostrophe, GDK_KEY_dead_abovering, GDK_KEY_a, 0x01FB, 
GDK_KEY_apostrophe, GDK_KEY_dead_cedilla, GDK_KEY_C, 0x1E08, 
GDK_KEY_apostrophe, GDK_KEY_dead_cedilla, GDK_KEY_c, 0x1E09, 
GDK_KEY_apostrophe, GDK_KEY_dead_horn, GDK_KEY_O, 0x1EDA, 
GDK_KEY_apostrophe, GDK_KEY_dead_horn, GDK_KEY_U, 0x1EE8, 
GDK_KEY_apostrophe, GDK_KEY_dead_horn, GDK_KEY_o, 0x1EDB, 
GDK_KEY_apostrophe, GDK_KEY_dead_horn, GDK_KEY_u, 0x1EE9, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F0C, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_EPSILON, 0x1F1C, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F2C, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_IOTA, 0x1F3C, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_OMICRON, 0x1F4C, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1F6C, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F04, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_epsilon, 0x1F14, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F24, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_iota, 0x1F34, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_omicron, 0x1F44, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_upsilon, 0x1F54, 
GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1F64, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F0D, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_EPSILON, 0x1F1D, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F2D, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_IOTA, 0x1F3D, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMICRON, 0x1F4D, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_UPSILON, 0x1F5D, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1F6D, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F05, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_epsilon, 0x1F15, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F25, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_iota, 0x1F35, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_omicron, 0x1F45, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_upsilon, 0x1F55, 
GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1F65, 
GDK_KEY_apostrophe, GDK_KEY_KP_Divide, GDK_KEY_O, 0x01FE, 
GDK_KEY_apostrophe, GDK_KEY_KP_Divide, GDK_KEY_o, 0x01FF, 
GDK_KEY_parenleft, GDK_KEY_0, GDK_KEY_parenright, 0x24EA, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_parenright, 0x2460, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_parenright, 0x2461, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_parenright, 0x2462, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_parenright, 0x2463, 
GDK_KEY_parenleft, GDK_KEY_5, GDK_KEY_parenright, 0x2464, 
GDK_KEY_parenleft, GDK_KEY_6, GDK_KEY_parenright, 0x2465, 
GDK_KEY_parenleft, GDK_KEY_7, GDK_KEY_parenright, 0x2466, 
GDK_KEY_parenleft, GDK_KEY_8, GDK_KEY_parenright, 0x2467, 
GDK_KEY_parenleft, GDK_KEY_9, GDK_KEY_parenright, 0x2468, 
GDK_KEY_parenleft, GDK_KEY_A, GDK_KEY_parenright, 0x24B6, 
GDK_KEY_parenleft, GDK_KEY_B, GDK_KEY_parenright, 0x24B7, 
GDK_KEY_parenleft, GDK_KEY_C, GDK_KEY_parenright, 0x24B8, 
GDK_KEY_parenleft, GDK_KEY_D, GDK_KEY_parenright, 0x24B9, 
GDK_KEY_parenleft, GDK_KEY_E, GDK_KEY_parenright, 0x24BA, 
GDK_KEY_parenleft, GDK_KEY_F, GDK_KEY_parenright, 0x24BB, 
GDK_KEY_parenleft, GDK_KEY_G, GDK_KEY_parenright, 0x24BC, 
GDK_KEY_parenleft, GDK_KEY_H, GDK_KEY_parenright, 0x24BD, 
GDK_KEY_parenleft, GDK_KEY_I, GDK_KEY_parenright, 0x24BE, 
GDK_KEY_parenleft, GDK_KEY_J, GDK_KEY_parenright, 0x24BF, 
GDK_KEY_parenleft, GDK_KEY_K, GDK_KEY_parenright, 0x24C0, 
GDK_KEY_parenleft, GDK_KEY_L, GDK_KEY_parenright, 0x24C1, 
GDK_KEY_parenleft, GDK_KEY_M, GDK_KEY_parenright, 0x24C2, 
GDK_KEY_parenleft, GDK_KEY_N, GDK_KEY_parenright, 0x24C3, 
GDK_KEY_parenleft, GDK_KEY_O, GDK_KEY_parenright, 0x24C4, 
GDK_KEY_parenleft, GDK_KEY_P, GDK_KEY_parenright, 0x24C5, 
GDK_KEY_parenleft, GDK_KEY_Q, GDK_KEY_parenright, 0x24C6, 
GDK_KEY_parenleft, GDK_KEY_R, GDK_KEY_parenright, 0x24C7, 
GDK_KEY_parenleft, GDK_KEY_S, GDK_KEY_parenright, 0x24C8, 
GDK_KEY_parenleft, GDK_KEY_T, GDK_KEY_parenright, 0x24C9, 
GDK_KEY_parenleft, GDK_KEY_U, GDK_KEY_parenright, 0x24CA, 
GDK_KEY_parenleft, GDK_KEY_V, GDK_KEY_parenright, 0x24CB, 
GDK_KEY_parenleft, GDK_KEY_W, GDK_KEY_parenright, 0x24CC, 
GDK_KEY_parenleft, GDK_KEY_X, GDK_KEY_parenright, 0x24CD, 
GDK_KEY_parenleft, GDK_KEY_Y, GDK_KEY_parenright, 0x24CE, 
GDK_KEY_parenleft, GDK_KEY_Z, GDK_KEY_parenright, 0x24CF, 
GDK_KEY_parenleft, GDK_KEY_a, GDK_KEY_parenright, 0x24D0, 
GDK_KEY_parenleft, GDK_KEY_b, GDK_KEY_parenright, 0x24D1, 
GDK_KEY_parenleft, GDK_KEY_c, GDK_KEY_parenright, 0x24D2, 
GDK_KEY_parenleft, GDK_KEY_d, GDK_KEY_parenright, 0x24D3, 
GDK_KEY_parenleft, GDK_KEY_e, GDK_KEY_parenright, 0x24D4, 
GDK_KEY_parenleft, GDK_KEY_f, GDK_KEY_parenright, 0x24D5, 
GDK_KEY_parenleft, GDK_KEY_g, GDK_KEY_parenright, 0x24D6, 
GDK_KEY_parenleft, GDK_KEY_h, GDK_KEY_parenright, 0x24D7, 
GDK_KEY_parenleft, GDK_KEY_i, GDK_KEY_parenright, 0x24D8, 
GDK_KEY_parenleft, GDK_KEY_j, GDK_KEY_parenright, 0x24D9, 
GDK_KEY_parenleft, GDK_KEY_k, GDK_KEY_parenright, 0x24DA, 
GDK_KEY_parenleft, GDK_KEY_l, GDK_KEY_parenright, 0x24DB, 
GDK_KEY_parenleft, GDK_KEY_m, GDK_KEY_parenright, 0x24DC, 
GDK_KEY_parenleft, GDK_KEY_n, GDK_KEY_parenright, 0x24DD, 
GDK_KEY_parenleft, GDK_KEY_o, GDK_KEY_parenright, 0x24DE, 
GDK_KEY_parenleft, GDK_KEY_p, GDK_KEY_parenright, 0x24DF, 
GDK_KEY_parenleft, GDK_KEY_q, GDK_KEY_parenright, 0x24E0, 
GDK_KEY_parenleft, GDK_KEY_r, GDK_KEY_parenright, 0x24E1, 
GDK_KEY_parenleft, GDK_KEY_s, GDK_KEY_parenright, 0x24E2, 
GDK_KEY_parenleft, GDK_KEY_t, GDK_KEY_parenright, 0x24E3, 
GDK_KEY_parenleft, GDK_KEY_u, GDK_KEY_parenright, 0x24E4, 
GDK_KEY_parenleft, GDK_KEY_v, GDK_KEY_parenright, 0x24E5, 
GDK_KEY_parenleft, GDK_KEY_w, GDK_KEY_parenright, 0x24E6, 
GDK_KEY_parenleft, GDK_KEY_x, GDK_KEY_parenright, 0x24E7, 
GDK_KEY_parenleft, GDK_KEY_y, GDK_KEY_parenright, 0x24E8, 
GDK_KEY_parenleft, GDK_KEY_z, GDK_KEY_parenright, 0x24E9, 
GDK_KEY_parenleft, GDK_KEY_kana_WO, GDK_KEY_parenright, 0x32FE, 
GDK_KEY_parenleft, GDK_KEY_kana_A, GDK_KEY_parenright, 0x32D0, 
GDK_KEY_parenleft, GDK_KEY_kana_I, GDK_KEY_parenright, 0x32D1, 
GDK_KEY_parenleft, GDK_KEY_kana_U, GDK_KEY_parenright, 0x32D2, 
GDK_KEY_parenleft, GDK_KEY_kana_E, GDK_KEY_parenright, 0x32D3, 
GDK_KEY_parenleft, GDK_KEY_kana_O, GDK_KEY_parenright, 0x32D4, 
GDK_KEY_parenleft, GDK_KEY_kana_KA, GDK_KEY_parenright, 0x32D5, 
GDK_KEY_parenleft, GDK_KEY_kana_KI, GDK_KEY_parenright, 0x32D6, 
GDK_KEY_parenleft, GDK_KEY_kana_KU, GDK_KEY_parenright, 0x32D7, 
GDK_KEY_parenleft, GDK_KEY_kana_KE, GDK_KEY_parenright, 0x32D8, 
GDK_KEY_parenleft, GDK_KEY_kana_KO, GDK_KEY_parenright, 0x32D9, 
GDK_KEY_parenleft, GDK_KEY_kana_SA, GDK_KEY_parenright, 0x32DA, 
GDK_KEY_parenleft, GDK_KEY_kana_SHI, GDK_KEY_parenright, 0x32DB, 
GDK_KEY_parenleft, GDK_KEY_kana_SU, GDK_KEY_parenright, 0x32DC, 
GDK_KEY_parenleft, GDK_KEY_kana_SE, GDK_KEY_parenright, 0x32DD, 
GDK_KEY_parenleft, GDK_KEY_kana_SO, GDK_KEY_parenright, 0x32DE, 
GDK_KEY_parenleft, GDK_KEY_kana_TA, GDK_KEY_parenright, 0x32DF, 
GDK_KEY_parenleft, GDK_KEY_kana_CHI, GDK_KEY_parenright, 0x32E0, 
GDK_KEY_parenleft, GDK_KEY_kana_TSU, GDK_KEY_parenright, 0x32E1, 
GDK_KEY_parenleft, GDK_KEY_kana_TE, GDK_KEY_parenright, 0x32E2, 
GDK_KEY_parenleft, GDK_KEY_kana_TO, GDK_KEY_parenright, 0x32E3, 
GDK_KEY_parenleft, GDK_KEY_kana_NA, GDK_KEY_parenright, 0x32E4, 
GDK_KEY_parenleft, GDK_KEY_kana_NI, GDK_KEY_parenright, 0x32E5, 
GDK_KEY_parenleft, GDK_KEY_kana_NU, GDK_KEY_parenright, 0x32E6, 
GDK_KEY_parenleft, GDK_KEY_kana_NE, GDK_KEY_parenright, 0x32E7, 
GDK_KEY_parenleft, GDK_KEY_kana_NO, GDK_KEY_parenright, 0x32E8, 
GDK_KEY_parenleft, GDK_KEY_kana_HA, GDK_KEY_parenright, 0x32E9, 
GDK_KEY_parenleft, GDK_KEY_kana_HI, GDK_KEY_parenright, 0x32EA, 
GDK_KEY_parenleft, GDK_KEY_kana_FU, GDK_KEY_parenright, 0x32EB, 
GDK_KEY_parenleft, GDK_KEY_kana_HE, GDK_KEY_parenright, 0x32EC, 
GDK_KEY_parenleft, GDK_KEY_kana_HO, GDK_KEY_parenright, 0x32ED, 
GDK_KEY_parenleft, GDK_KEY_kana_MA, GDK_KEY_parenright, 0x32EE, 
GDK_KEY_parenleft, GDK_KEY_kana_MI, GDK_KEY_parenright, 0x32EF, 
GDK_KEY_parenleft, GDK_KEY_kana_MU, GDK_KEY_parenright, 0x32F0, 
GDK_KEY_parenleft, GDK_KEY_kana_ME, GDK_KEY_parenright, 0x32F1, 
GDK_KEY_parenleft, GDK_KEY_kana_MO, GDK_KEY_parenright, 0x32F2, 
GDK_KEY_parenleft, GDK_KEY_kana_YA, GDK_KEY_parenright, 0x32F3, 
GDK_KEY_parenleft, GDK_KEY_kana_YU, GDK_KEY_parenright, 0x32F4, 
GDK_KEY_parenleft, GDK_KEY_kana_YO, GDK_KEY_parenright, 0x32F5, 
GDK_KEY_parenleft, GDK_KEY_kana_RA, GDK_KEY_parenright, 0x32F6, 
GDK_KEY_parenleft, GDK_KEY_kana_RI, GDK_KEY_parenright, 0x32F7, 
GDK_KEY_parenleft, GDK_KEY_kana_RU, GDK_KEY_parenright, 0x32F8, 
GDK_KEY_parenleft, GDK_KEY_kana_RE, GDK_KEY_parenright, 0x32F9, 
GDK_KEY_parenleft, GDK_KEY_kana_RO, GDK_KEY_parenright, 0x32FA, 
GDK_KEY_parenleft, GDK_KEY_kana_WA, GDK_KEY_parenright, 0x32FB, 
GDK_KEY_parenleft, 0x1100, GDK_KEY_parenright, 0x3260, 
GDK_KEY_parenleft, 0x1102, GDK_KEY_parenright, 0x3261, 
GDK_KEY_parenleft, 0x1103, GDK_KEY_parenright, 0x3262, 
GDK_KEY_parenleft, 0x1105, GDK_KEY_parenright, 0x3263, 
GDK_KEY_parenleft, 0x1106, GDK_KEY_parenright, 0x3264, 
GDK_KEY_parenleft, 0x1107, GDK_KEY_parenright, 0x3265, 
GDK_KEY_parenleft, 0x1109, GDK_KEY_parenright, 0x3266, 
GDK_KEY_parenleft, 0x110B, GDK_KEY_parenright, 0x3267, 
GDK_KEY_parenleft, 0x110C, GDK_KEY_parenright, 0x3268, 
GDK_KEY_parenleft, 0x110E, GDK_KEY_parenright, 0x3269, 
GDK_KEY_parenleft, 0x110F, GDK_KEY_parenright, 0x326A, 
GDK_KEY_parenleft, 0x1110, GDK_KEY_parenright, 0x326B, 
GDK_KEY_parenleft, 0x1111, GDK_KEY_parenright, 0x326C, 
GDK_KEY_parenleft, 0x1112, GDK_KEY_parenright, 0x326D, 
GDK_KEY_parenleft, 0x30F0, GDK_KEY_parenright, 0x32FC, 
GDK_KEY_parenleft, 0x30F1, GDK_KEY_parenright, 0x32FD, 
GDK_KEY_parenleft, 0x4E00, GDK_KEY_parenright, 0x3280, 
GDK_KEY_parenleft, 0x4E03, GDK_KEY_parenright, 0x3286, 
GDK_KEY_parenleft, 0x4E09, GDK_KEY_parenright, 0x3282, 
GDK_KEY_parenleft, 0x4E0A, GDK_KEY_parenright, 0x32A4, 
GDK_KEY_parenleft, 0x4E0B, GDK_KEY_parenright, 0x32A6, 
GDK_KEY_parenleft, 0x4E2D, GDK_KEY_parenright, 0x32A5, 
GDK_KEY_parenleft, 0x4E5D, GDK_KEY_parenright, 0x3288, 
GDK_KEY_parenleft, 0x4E8C, GDK_KEY_parenright, 0x3281, 
GDK_KEY_parenleft, 0x4E94, GDK_KEY_parenright, 0x3284, 
GDK_KEY_parenleft, 0x4F01, GDK_KEY_parenright, 0x32AD, 
GDK_KEY_parenleft, 0x4F11, GDK_KEY_parenright, 0x32A1, 
GDK_KEY_parenleft, 0x512A, GDK_KEY_parenright, 0x329D, 
GDK_KEY_parenleft, 0x516B, GDK_KEY_parenright, 0x3287, 
GDK_KEY_parenleft, 0x516D, GDK_KEY_parenright, 0x3285, 
GDK_KEY_parenleft, 0x5199, GDK_KEY_parenright, 0x32A2, 
GDK_KEY_parenleft, 0x52B4, GDK_KEY_parenright, 0x3298, 
GDK_KEY_parenleft, 0x533B, GDK_KEY_parenright, 0x32A9, 
GDK_KEY_parenleft, 0x5341, GDK_KEY_parenright, 0x3289, 
GDK_KEY_parenleft, 0x5354, GDK_KEY_parenright, 0x32AF, 
GDK_KEY_parenleft, 0x5370, GDK_KEY_parenright, 0x329E, 
GDK_KEY_parenleft, 0x53F3, GDK_KEY_parenright, 0x32A8, 
GDK_KEY_parenleft, 0x540D, GDK_KEY_parenright, 0x3294, 
GDK_KEY_parenleft, 0x56DB, GDK_KEY_parenright, 0x3283, 
GDK_KEY_parenleft, 0x571F, GDK_KEY_parenright, 0x328F, 
GDK_KEY_parenleft, 0x591C, GDK_KEY_parenright, 0x32B0, 
GDK_KEY_parenleft, 0x5973, GDK_KEY_parenright, 0x329B, 
GDK_KEY_parenleft, 0x5B66, GDK_KEY_parenright, 0x32AB, 
GDK_KEY_parenleft, 0x5B97, GDK_KEY_parenright, 0x32AA, 
GDK_KEY_parenleft, 0x5DE6, GDK_KEY_parenright, 0x32A7, 
GDK_KEY_parenleft, 0x65E5, GDK_KEY_parenright, 0x3290, 
GDK_KEY_parenleft, 0x6708, GDK_KEY_parenright, 0x328A, 
GDK_KEY_parenleft, 0x6709, GDK_KEY_parenright, 0x3292, 
GDK_KEY_parenleft, 0x6728, GDK_KEY_parenright, 0x328D, 
GDK_KEY_parenleft, 0x682A, GDK_KEY_parenright, 0x3291, 
GDK_KEY_parenleft, 0x6B63, GDK_KEY_parenright, 0x32A3, 
GDK_KEY_parenleft, 0x6C34, GDK_KEY_parenright, 0x328C, 
GDK_KEY_parenleft, 0x6CE8, GDK_KEY_parenright, 0x329F, 
GDK_KEY_parenleft, 0x706B, GDK_KEY_parenright, 0x328B, 
GDK_KEY_parenleft, 0x7279, GDK_KEY_parenright, 0x3295, 
GDK_KEY_parenleft, 0x7537, GDK_KEY_parenright, 0x329A, 
GDK_KEY_parenleft, 0x76E3, GDK_KEY_parenright, 0x32AC, 
GDK_KEY_parenleft, 0x793E, GDK_KEY_parenright, 0x3293, 
GDK_KEY_parenleft, 0x795D, GDK_KEY_parenright, 0x3297, 
GDK_KEY_parenleft, 0x79D8, GDK_KEY_parenright, 0x3299, 
GDK_KEY_parenleft, 0x8CA1, GDK_KEY_parenright, 0x3296, 
GDK_KEY_parenleft, 0x8CC7, GDK_KEY_parenright, 0x32AE, 
GDK_KEY_parenleft, 0x9069, GDK_KEY_parenright, 0x329C, 
GDK_KEY_parenleft, 0x91D1, GDK_KEY_parenright, 0x328E, 
GDK_KEY_parenleft, 0x9805, GDK_KEY_parenright, 0x32A0, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_parenright, 0x2461, 
GDK_KEY_parenleft, GDK_KEY_KP_0, GDK_KEY_parenright, 0x24EA, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_parenright, 0x2460, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_parenright, 0x2461, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_parenright, 0x2462, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_parenright, 0x2463, 
GDK_KEY_parenleft, GDK_KEY_KP_5, GDK_KEY_parenright, 0x2464, 
GDK_KEY_parenleft, GDK_KEY_KP_6, GDK_KEY_parenright, 0x2465, 
GDK_KEY_parenleft, GDK_KEY_KP_7, GDK_KEY_parenright, 0x2466, 
GDK_KEY_parenleft, GDK_KEY_KP_8, GDK_KEY_parenright, 0x2467, 
GDK_KEY_parenleft, GDK_KEY_KP_9, GDK_KEY_parenright, 0x2468, 
GDK_KEY_asterisk, GDK_KEY_apostrophe, GDK_KEY_A, 0x01FA, 
GDK_KEY_asterisk, GDK_KEY_apostrophe, GDK_KEY_a, 0x01FB, 
GDK_KEY_minus, GDK_KEY_minus, GDK_KEY_space, 0x00AD, 
GDK_KEY_minus, GDK_KEY_minus, GDK_KEY_minus, 0x2014, 
GDK_KEY_minus, GDK_KEY_minus, GDK_KEY_period, 0x2013, 
GDK_KEY_period, GDK_KEY_exclam, GDK_KEY_S, 0x1E68, 
GDK_KEY_period, GDK_KEY_exclam, GDK_KEY_s, 0x1E69, 
GDK_KEY_period, GDK_KEY_apostrophe, GDK_KEY_S, 0x1E64, 
GDK_KEY_period, GDK_KEY_apostrophe, GDK_KEY_s, 0x1E65, 
GDK_KEY_period, GDK_KEY_acute, GDK_KEY_S, 0x1E64, 
GDK_KEY_period, GDK_KEY_acute, GDK_KEY_s, 0x1E65, 
GDK_KEY_period, GDK_KEY_dead_acute, GDK_KEY_S, 0x1E64, 
GDK_KEY_period, GDK_KEY_dead_acute, GDK_KEY_s, 0x1E65, 
GDK_KEY_period, GDK_KEY_dead_caron, GDK_KEY_S, 0x1E66, 
GDK_KEY_period, GDK_KEY_dead_caron, GDK_KEY_s, 0x1E67, 
GDK_KEY_period, GDK_KEY_dead_belowdot, GDK_KEY_S, 0x1E68, 
GDK_KEY_period, GDK_KEY_dead_belowdot, GDK_KEY_s, 0x1E69, 
GDK_KEY_1, GDK_KEY_1, GDK_KEY_0, 0x2152, 
GDK_KEY_question, GDK_KEY_plus, GDK_KEY_O, 0x1EDE, 
GDK_KEY_question, GDK_KEY_plus, GDK_KEY_U, 0x1EEC, 
GDK_KEY_question, GDK_KEY_plus, GDK_KEY_o, 0x1EDF, 
GDK_KEY_question, GDK_KEY_plus, GDK_KEY_u, 0x1EED, 
GDK_KEY_question, GDK_KEY_asciicircum, GDK_KEY_A, 0x1EA8, 
GDK_KEY_question, GDK_KEY_asciicircum, GDK_KEY_E, 0x1EC2, 
GDK_KEY_question, GDK_KEY_asciicircum, GDK_KEY_O, 0x1ED4, 
GDK_KEY_question, GDK_KEY_asciicircum, GDK_KEY_a, 0x1EA9, 
GDK_KEY_question, GDK_KEY_asciicircum, GDK_KEY_e, 0x1EC3, 
GDK_KEY_question, GDK_KEY_asciicircum, GDK_KEY_o, 0x1ED5, 
GDK_KEY_question, GDK_KEY_b, GDK_KEY_A, 0x1EB2, 
GDK_KEY_question, GDK_KEY_b, GDK_KEY_a, 0x1EB3, 
GDK_KEY_question, GDK_KEY_dead_circumflex, GDK_KEY_A, 0x1EA8, 
GDK_KEY_question, GDK_KEY_dead_circumflex, GDK_KEY_E, 0x1EC2, 
GDK_KEY_question, GDK_KEY_dead_circumflex, GDK_KEY_O, 0x1ED4, 
GDK_KEY_question, GDK_KEY_dead_circumflex, GDK_KEY_a, 0x1EA9, 
GDK_KEY_question, GDK_KEY_dead_circumflex, GDK_KEY_e, 0x1EC3, 
GDK_KEY_question, GDK_KEY_dead_circumflex, GDK_KEY_o, 0x1ED5, 
GDK_KEY_question, GDK_KEY_dead_breve, GDK_KEY_A, 0x1EB2, 
GDK_KEY_question, GDK_KEY_dead_breve, GDK_KEY_a, 0x1EB3, 
GDK_KEY_question, GDK_KEY_dead_horn, GDK_KEY_O, 0x1EDE, 
GDK_KEY_question, GDK_KEY_dead_horn, GDK_KEY_U, 0x1EEC, 
GDK_KEY_question, GDK_KEY_dead_horn, GDK_KEY_o, 0x1EDF, 
GDK_KEY_question, GDK_KEY_dead_horn, GDK_KEY_u, 0x1EED, 
GDK_KEY_U, GDK_KEY_exclam, GDK_KEY_A, 0x1EB6, 
GDK_KEY_U, GDK_KEY_exclam, GDK_KEY_a, 0x1EB7, 
GDK_KEY_U, GDK_KEY_cedilla, GDK_KEY_E, 0x1E1C, 
GDK_KEY_U, GDK_KEY_cedilla, GDK_KEY_e, 0x1E1D, 
GDK_KEY_U, GDK_KEY_dead_cedilla, GDK_KEY_E, 0x1E1C, 
GDK_KEY_U, GDK_KEY_dead_cedilla, GDK_KEY_e, 0x1E1D, 
GDK_KEY_U, GDK_KEY_dead_belowdot, GDK_KEY_A, 0x1EB6, 
GDK_KEY_U, GDK_KEY_dead_belowdot, GDK_KEY_a, 0x1EB7, 
GDK_KEY_asciicircum, GDK_KEY_exclam, GDK_KEY_A, 0x1EAC, 
GDK_KEY_asciicircum, GDK_KEY_exclam, GDK_KEY_E, 0x1EC6, 
GDK_KEY_asciicircum, GDK_KEY_exclam, GDK_KEY_O, 0x1ED8, 
GDK_KEY_asciicircum, GDK_KEY_exclam, GDK_KEY_a, 0x1EAD, 
GDK_KEY_asciicircum, GDK_KEY_exclam, GDK_KEY_e, 0x1EC7, 
GDK_KEY_asciicircum, GDK_KEY_exclam, GDK_KEY_o, 0x1ED9, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_a, 0x00AA, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_h, 0x02B0, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_i, 0x2071, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_j, 0x02B2, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_l, 0x02E1, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_n, 0x207F, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_o, 0x00BA, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_r, 0x02B3, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_s, 0x02E2, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_w, 0x02B7, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_x, 0x02E3, 
GDK_KEY_asciicircum, GDK_KEY_underscore, GDK_KEY_y, 0x02B8, 
GDK_KEY_asciicircum, GDK_KEY_underscore, 0x0263, 0x02E0, 
GDK_KEY_asciicircum, GDK_KEY_underscore, 0x0266, 0x02B1, 
GDK_KEY_asciicircum, GDK_KEY_underscore, 0x0279, 0x02B4, 
GDK_KEY_asciicircum, GDK_KEY_underscore, 0x027B, 0x02B5, 
GDK_KEY_asciicircum, GDK_KEY_underscore, 0x0281, 0x02B6, 
GDK_KEY_asciicircum, GDK_KEY_underscore, 0x0295, 0x02E4, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_a, 0x00AA, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_h, 0x02B0, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_i, 0x2071, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_j, 0x02B2, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_l, 0x02E1, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_n, 0x207F, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_o, 0x00BA, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_r, 0x02B3, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_s, 0x02E2, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_w, 0x02B7, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_x, 0x02E3, 
GDK_KEY_asciicircum, GDK_KEY_underbar, GDK_KEY_y, 0x02B8, 
GDK_KEY_asciicircum, GDK_KEY_underbar, 0x0263, 0x02E0, 
GDK_KEY_asciicircum, GDK_KEY_underbar, 0x0266, 0x02B1, 
GDK_KEY_asciicircum, GDK_KEY_underbar, 0x0279, 0x02B4, 
GDK_KEY_asciicircum, GDK_KEY_underbar, 0x027B, 0x02B5, 
GDK_KEY_asciicircum, GDK_KEY_underbar, 0x0281, 0x02B6, 
GDK_KEY_asciicircum, GDK_KEY_underbar, 0x0295, 0x02E4, 
GDK_KEY_asciicircum, GDK_KEY_dead_belowdot, GDK_KEY_A, 0x1EAC, 
GDK_KEY_asciicircum, GDK_KEY_dead_belowdot, GDK_KEY_E, 0x1EC6, 
GDK_KEY_asciicircum, GDK_KEY_dead_belowdot, GDK_KEY_O, 0x1ED8, 
GDK_KEY_asciicircum, GDK_KEY_dead_belowdot, GDK_KEY_a, 0x1EAD, 
GDK_KEY_asciicircum, GDK_KEY_dead_belowdot, GDK_KEY_e, 0x1EC7, 
GDK_KEY_asciicircum, GDK_KEY_dead_belowdot, GDK_KEY_o, 0x1ED9, 
GDK_KEY_underscore, GDK_KEY_exclam, GDK_KEY_L, 0x1E38, 
GDK_KEY_underscore, GDK_KEY_exclam, GDK_KEY_R, 0x1E5C, 
GDK_KEY_underscore, GDK_KEY_exclam, GDK_KEY_l, 0x1E39, 
GDK_KEY_underscore, GDK_KEY_exclam, GDK_KEY_r, 0x1E5D, 
GDK_KEY_underscore, GDK_KEY_quotedbl, GDK_KEY_A, 0x01DE, 
GDK_KEY_underscore, GDK_KEY_quotedbl, GDK_KEY_O, 0x022A, 
GDK_KEY_underscore, GDK_KEY_quotedbl, GDK_KEY_U, 0x01D5, 
GDK_KEY_underscore, GDK_KEY_quotedbl, GDK_KEY_a, 0x01DF, 
GDK_KEY_underscore, GDK_KEY_quotedbl, GDK_KEY_o, 0x022B, 
GDK_KEY_underscore, GDK_KEY_quotedbl, GDK_KEY_u, 0x01D6, 
GDK_KEY_underscore, GDK_KEY_period, GDK_KEY_A, 0x01E0, 
GDK_KEY_underscore, GDK_KEY_period, GDK_KEY_O, 0x0230, 
GDK_KEY_underscore, GDK_KEY_period, GDK_KEY_a, 0x01E1, 
GDK_KEY_underscore, GDK_KEY_period, GDK_KEY_o, 0x0231, 
GDK_KEY_underscore, GDK_KEY_semicolon, GDK_KEY_O, 0x01EC, 
GDK_KEY_underscore, GDK_KEY_semicolon, GDK_KEY_o, 0x01ED, 
GDK_KEY_underscore, GDK_KEY_asciitilde, GDK_KEY_O, 0x022C, 
GDK_KEY_underscore, GDK_KEY_asciitilde, GDK_KEY_o, 0x022D, 
GDK_KEY_underscore, GDK_KEY_dead_tilde, GDK_KEY_O, 0x022C, 
GDK_KEY_underscore, GDK_KEY_dead_tilde, GDK_KEY_o, 0x022D, 
GDK_KEY_underscore, GDK_KEY_dead_abovedot, GDK_KEY_A, 0x01E0, 
GDK_KEY_underscore, GDK_KEY_dead_abovedot, GDK_KEY_O, 0x0230, 
GDK_KEY_underscore, GDK_KEY_dead_abovedot, GDK_KEY_a, 0x01E1, 
GDK_KEY_underscore, GDK_KEY_dead_abovedot, GDK_KEY_o, 0x0231, 
GDK_KEY_underscore, GDK_KEY_dead_diaeresis, GDK_KEY_A, 0x01DE, 
GDK_KEY_underscore, GDK_KEY_dead_diaeresis, GDK_KEY_O, 0x022A, 
GDK_KEY_underscore, GDK_KEY_dead_diaeresis, GDK_KEY_U, 0x01D5, 
GDK_KEY_underscore, GDK_KEY_dead_diaeresis, GDK_KEY_a, 0x01DF, 
GDK_KEY_underscore, GDK_KEY_dead_diaeresis, GDK_KEY_o, 0x022B, 
GDK_KEY_underscore, GDK_KEY_dead_diaeresis, GDK_KEY_u, 0x01D6, 
GDK_KEY_underscore, GDK_KEY_dead_ogonek, GDK_KEY_O, 0x01EC, 
GDK_KEY_underscore, GDK_KEY_dead_ogonek, GDK_KEY_o, 0x01ED, 
GDK_KEY_underscore, GDK_KEY_dead_belowdot, GDK_KEY_L, 0x1E38, 
GDK_KEY_underscore, GDK_KEY_dead_belowdot, GDK_KEY_R, 0x1E5C, 
GDK_KEY_underscore, GDK_KEY_dead_belowdot, GDK_KEY_l, 0x1E39, 
GDK_KEY_underscore, GDK_KEY_dead_belowdot, GDK_KEY_r, 0x1E5D, 
GDK_KEY_grave, GDK_KEY_quotedbl, GDK_KEY_U, 0x01DB, 
GDK_KEY_grave, GDK_KEY_quotedbl, GDK_KEY_u, 0x01DC, 
GDK_KEY_grave, GDK_KEY_quotedbl, GDK_KEY_Greek_iota, 0x1FD2, 
GDK_KEY_grave, GDK_KEY_quotedbl, GDK_KEY_Greek_upsilon, 0x1FE2, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F0B, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_EPSILON, 0x1F1B, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F2B, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_IOTA, 0x1F3B, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_OMICRON, 0x1F4B, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_UPSILON, 0x1F5B, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1F6B, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F03, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_epsilon, 0x1F13, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F23, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_iota, 0x1F33, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_omicron, 0x1F43, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_upsilon, 0x1F53, 
GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1F63, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F0A, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_EPSILON, 0x1F1A, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F2A, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_IOTA, 0x1F3A, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_OMICRON, 0x1F4A, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1F6A, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F02, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_epsilon, 0x1F12, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F22, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_iota, 0x1F32, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_omicron, 0x1F42, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_upsilon, 0x1F52, 
GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1F62, 
GDK_KEY_grave, GDK_KEY_plus, GDK_KEY_O, 0x1EDC, 
GDK_KEY_grave, GDK_KEY_plus, GDK_KEY_U, 0x1EEA, 
GDK_KEY_grave, GDK_KEY_plus, GDK_KEY_o, 0x1EDD, 
GDK_KEY_grave, GDK_KEY_plus, GDK_KEY_u, 0x1EEB, 
GDK_KEY_grave, GDK_KEY_asciicircum, GDK_KEY_A, 0x1EA6, 
GDK_KEY_grave, GDK_KEY_asciicircum, GDK_KEY_E, 0x1EC0, 
GDK_KEY_grave, GDK_KEY_asciicircum, GDK_KEY_O, 0x1ED2, 
GDK_KEY_grave, GDK_KEY_asciicircum, GDK_KEY_a, 0x1EA7, 
GDK_KEY_grave, GDK_KEY_asciicircum, GDK_KEY_e, 0x1EC1, 
GDK_KEY_grave, GDK_KEY_asciicircum, GDK_KEY_o, 0x1ED3, 
GDK_KEY_grave, GDK_KEY_underscore, GDK_KEY_E, 0x1E14, 
GDK_KEY_grave, GDK_KEY_underscore, GDK_KEY_O, 0x1E50, 
GDK_KEY_grave, GDK_KEY_underscore, GDK_KEY_e, 0x1E15, 
GDK_KEY_grave, GDK_KEY_underscore, GDK_KEY_o, 0x1E51, 
GDK_KEY_grave, GDK_KEY_b, GDK_KEY_A, 0x1EB0, 
GDK_KEY_grave, GDK_KEY_b, GDK_KEY_a, 0x1EB1, 
GDK_KEY_grave, GDK_KEY_macron, GDK_KEY_E, 0x1E14, 
GDK_KEY_grave, GDK_KEY_macron, GDK_KEY_O, 0x1E50, 
GDK_KEY_grave, GDK_KEY_macron, GDK_KEY_e, 0x1E15, 
GDK_KEY_grave, GDK_KEY_macron, GDK_KEY_o, 0x1E51, 
GDK_KEY_grave, GDK_KEY_dead_circumflex, GDK_KEY_A, 0x1EA6, 
GDK_KEY_grave, GDK_KEY_dead_circumflex, GDK_KEY_E, 0x1EC0, 
GDK_KEY_grave, GDK_KEY_dead_circumflex, GDK_KEY_O, 0x1ED2, 
GDK_KEY_grave, GDK_KEY_dead_circumflex, GDK_KEY_a, 0x1EA7, 
GDK_KEY_grave, GDK_KEY_dead_circumflex, GDK_KEY_e, 0x1EC1, 
GDK_KEY_grave, GDK_KEY_dead_circumflex, GDK_KEY_o, 0x1ED3, 
GDK_KEY_grave, GDK_KEY_dead_macron, GDK_KEY_E, 0x1E14, 
GDK_KEY_grave, GDK_KEY_dead_macron, GDK_KEY_O, 0x1E50, 
GDK_KEY_grave, GDK_KEY_dead_macron, GDK_KEY_e, 0x1E15, 
GDK_KEY_grave, GDK_KEY_dead_macron, GDK_KEY_o, 0x1E51, 
GDK_KEY_grave, GDK_KEY_dead_breve, GDK_KEY_A, 0x1EB0, 
GDK_KEY_grave, GDK_KEY_dead_breve, GDK_KEY_a, 0x1EB1, 
GDK_KEY_grave, GDK_KEY_dead_diaeresis, GDK_KEY_U, 0x01DB, 
GDK_KEY_grave, GDK_KEY_dead_diaeresis, GDK_KEY_u, 0x01DC, 
GDK_KEY_grave, GDK_KEY_dead_diaeresis, GDK_KEY_Greek_iota, 0x1FD2, 
GDK_KEY_grave, GDK_KEY_dead_diaeresis, GDK_KEY_Greek_upsilon, 0x1FE2, 
GDK_KEY_grave, GDK_KEY_dead_horn, GDK_KEY_O, 0x1EDC, 
GDK_KEY_grave, GDK_KEY_dead_horn, GDK_KEY_U, 0x1EEA, 
GDK_KEY_grave, GDK_KEY_dead_horn, GDK_KEY_o, 0x1EDD, 
GDK_KEY_grave, GDK_KEY_dead_horn, GDK_KEY_u, 0x1EEB, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F0A, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_EPSILON, 0x1F1A, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F2A, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_IOTA, 0x1F3A, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_OMICRON, 0x1F4A, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1F6A, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F02, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_epsilon, 0x1F12, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F22, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_iota, 0x1F32, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_omicron, 0x1F42, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_upsilon, 0x1F52, 
GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1F62, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F0B, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_EPSILON, 0x1F1B, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F2B, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_IOTA, 0x1F3B, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMICRON, 0x1F4B, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_UPSILON, 0x1F5B, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1F6B, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F03, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_epsilon, 0x1F13, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F23, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_iota, 0x1F33, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_omicron, 0x1F43, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_upsilon, 0x1F53, 
GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1F63, 
GDK_KEY_b, GDK_KEY_exclam, GDK_KEY_A, 0x1EB6, 
GDK_KEY_b, GDK_KEY_exclam, GDK_KEY_a, 0x1EB7, 
GDK_KEY_b, GDK_KEY_comma, GDK_KEY_E, 0x1E1C, 
GDK_KEY_b, GDK_KEY_comma, GDK_KEY_e, 0x1E1D, 
GDK_KEY_b, GDK_KEY_cedilla, GDK_KEY_E, 0x1E1C, 
GDK_KEY_b, GDK_KEY_cedilla, GDK_KEY_e, 0x1E1D, 
GDK_KEY_b, GDK_KEY_dead_cedilla, GDK_KEY_E, 0x1E1C, 
GDK_KEY_b, GDK_KEY_dead_cedilla, GDK_KEY_e, 0x1E1D, 
GDK_KEY_b, GDK_KEY_dead_belowdot, GDK_KEY_A, 0x1EB6, 
GDK_KEY_b, GDK_KEY_dead_belowdot, GDK_KEY_a, 0x1EB7, 
GDK_KEY_c, GDK_KEY_quotedbl, GDK_KEY_U, 0x01D9, 
GDK_KEY_c, GDK_KEY_quotedbl, GDK_KEY_u, 0x01DA, 
GDK_KEY_c, GDK_KEY_dead_diaeresis, GDK_KEY_U, 0x01D9, 
GDK_KEY_c, GDK_KEY_dead_diaeresis, GDK_KEY_u, 0x01DA, 
GDK_KEY_asciitilde, GDK_KEY_quotedbl, GDK_KEY_Greek_iota, 0x1FD7, 
GDK_KEY_asciitilde, GDK_KEY_quotedbl, GDK_KEY_Greek_upsilon, 0x1FE7, 
GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F0F, 
GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F2F, 
GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_IOTA, 0x1F3F, 
GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_UPSILON, 0x1F5F, 
GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1F6F, 
GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F07, 
GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F27, 
GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_iota, 0x1F37, 
GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_upsilon, 0x1F57, 
GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1F67, 
GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F0E, 
GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F2E, 
GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_IOTA, 0x1F3E, 
GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1F6E, 
GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F06, 
GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F26, 
GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_iota, 0x1F36, 
GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_upsilon, 0x1F56, 
GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1F66, 
GDK_KEY_asciitilde, GDK_KEY_plus, GDK_KEY_O, 0x1EE0, 
GDK_KEY_asciitilde, GDK_KEY_plus, GDK_KEY_U, 0x1EEE, 
GDK_KEY_asciitilde, GDK_KEY_plus, GDK_KEY_o, 0x1EE1, 
GDK_KEY_asciitilde, GDK_KEY_plus, GDK_KEY_u, 0x1EEF, 
GDK_KEY_asciitilde, GDK_KEY_asciicircum, GDK_KEY_A, 0x1EAA, 
GDK_KEY_asciitilde, GDK_KEY_asciicircum, GDK_KEY_E, 0x1EC4, 
GDK_KEY_asciitilde, GDK_KEY_asciicircum, GDK_KEY_O, 0x1ED6, 
GDK_KEY_asciitilde, GDK_KEY_asciicircum, GDK_KEY_a, 0x1EAB, 
GDK_KEY_asciitilde, GDK_KEY_asciicircum, GDK_KEY_e, 0x1EC5, 
GDK_KEY_asciitilde, GDK_KEY_asciicircum, GDK_KEY_o, 0x1ED7, 
GDK_KEY_asciitilde, GDK_KEY_b, GDK_KEY_A, 0x1EB4, 
GDK_KEY_asciitilde, GDK_KEY_b, GDK_KEY_a, 0x1EB5, 
GDK_KEY_asciitilde, GDK_KEY_dead_circumflex, GDK_KEY_A, 0x1EAA, 
GDK_KEY_asciitilde, GDK_KEY_dead_circumflex, GDK_KEY_E, 0x1EC4, 
GDK_KEY_asciitilde, GDK_KEY_dead_circumflex, GDK_KEY_O, 0x1ED6, 
GDK_KEY_asciitilde, GDK_KEY_dead_circumflex, GDK_KEY_a, 0x1EAB, 
GDK_KEY_asciitilde, GDK_KEY_dead_circumflex, GDK_KEY_e, 0x1EC5, 
GDK_KEY_asciitilde, GDK_KEY_dead_circumflex, GDK_KEY_o, 0x1ED7, 
GDK_KEY_asciitilde, GDK_KEY_dead_breve, GDK_KEY_A, 0x1EB4, 
GDK_KEY_asciitilde, GDK_KEY_dead_breve, GDK_KEY_a, 0x1EB5, 
GDK_KEY_asciitilde, GDK_KEY_dead_diaeresis, GDK_KEY_Greek_iota, 0x1FD7, 
GDK_KEY_asciitilde, GDK_KEY_dead_diaeresis, GDK_KEY_Greek_upsilon, 0x1FE7, 
GDK_KEY_asciitilde, GDK_KEY_dead_horn, GDK_KEY_O, 0x1EE0, 
GDK_KEY_asciitilde, GDK_KEY_dead_horn, GDK_KEY_U, 0x1EEE, 
GDK_KEY_asciitilde, GDK_KEY_dead_horn, GDK_KEY_o, 0x1EE1, 
GDK_KEY_asciitilde, GDK_KEY_dead_horn, GDK_KEY_u, 0x1EEF, 
GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F0E, 
GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F2E, 
GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_IOTA, 0x1F3E, 
GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1F6E, 
GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F06, 
GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F26, 
GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_iota, 0x1F36, 
GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_upsilon, 0x1F56, 
GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1F66, 
GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F0F, 
GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F2F, 
GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_IOTA, 0x1F3F, 
GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_UPSILON, 0x1F5F, 
GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1F6F, 
GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F07, 
GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F27, 
GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_iota, 0x1F37, 
GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_upsilon, 0x1F57, 
GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1F67, 
GDK_KEY_macron, GDK_KEY_exclam, GDK_KEY_L, 0x1E38, 
GDK_KEY_macron, GDK_KEY_exclam, GDK_KEY_R, 0x1E5C, 
GDK_KEY_macron, GDK_KEY_exclam, GDK_KEY_l, 0x1E39, 
GDK_KEY_macron, GDK_KEY_exclam, GDK_KEY_r, 0x1E5D, 
GDK_KEY_macron, GDK_KEY_quotedbl, GDK_KEY_A, 0x01DE, 
GDK_KEY_macron, GDK_KEY_quotedbl, GDK_KEY_O, 0x022A, 
GDK_KEY_macron, GDK_KEY_quotedbl, GDK_KEY_U, 0x01D5, 
GDK_KEY_macron, GDK_KEY_quotedbl, GDK_KEY_a, 0x01DF, 
GDK_KEY_macron, GDK_KEY_quotedbl, GDK_KEY_o, 0x022B, 
GDK_KEY_macron, GDK_KEY_quotedbl, GDK_KEY_u, 0x01D6, 
GDK_KEY_macron, GDK_KEY_period, GDK_KEY_A, 0x01E0, 
GDK_KEY_macron, GDK_KEY_period, GDK_KEY_O, 0x0230, 
GDK_KEY_macron, GDK_KEY_period, GDK_KEY_a, 0x01E1, 
GDK_KEY_macron, GDK_KEY_period, GDK_KEY_o, 0x0231, 
GDK_KEY_macron, GDK_KEY_semicolon, GDK_KEY_O, 0x01EC, 
GDK_KEY_macron, GDK_KEY_semicolon, GDK_KEY_o, 0x01ED, 
GDK_KEY_macron, GDK_KEY_asciitilde, GDK_KEY_O, 0x022C, 
GDK_KEY_macron, GDK_KEY_asciitilde, GDK_KEY_o, 0x022D, 
GDK_KEY_macron, GDK_KEY_dead_tilde, GDK_KEY_O, 0x022C, 
GDK_KEY_macron, GDK_KEY_dead_tilde, GDK_KEY_o, 0x022D, 
GDK_KEY_macron, GDK_KEY_dead_abovedot, GDK_KEY_A, 0x01E0, 
GDK_KEY_macron, GDK_KEY_dead_abovedot, GDK_KEY_O, 0x0230, 
GDK_KEY_macron, GDK_KEY_dead_abovedot, GDK_KEY_a, 0x01E1, 
GDK_KEY_macron, GDK_KEY_dead_abovedot, GDK_KEY_o, 0x0231, 
GDK_KEY_macron, GDK_KEY_dead_diaeresis, GDK_KEY_A, 0x01DE, 
GDK_KEY_macron, GDK_KEY_dead_diaeresis, GDK_KEY_O, 0x022A, 
GDK_KEY_macron, GDK_KEY_dead_diaeresis, GDK_KEY_U, 0x01D5, 
GDK_KEY_macron, GDK_KEY_dead_diaeresis, GDK_KEY_a, 0x01DF, 
GDK_KEY_macron, GDK_KEY_dead_diaeresis, GDK_KEY_o, 0x022B, 
GDK_KEY_macron, GDK_KEY_dead_diaeresis, GDK_KEY_u, 0x01D6, 
GDK_KEY_macron, GDK_KEY_dead_ogonek, GDK_KEY_O, 0x01EC, 
GDK_KEY_macron, GDK_KEY_dead_ogonek, GDK_KEY_o, 0x01ED, 
GDK_KEY_macron, GDK_KEY_dead_belowdot, GDK_KEY_L, 0x1E38, 
GDK_KEY_macron, GDK_KEY_dead_belowdot, GDK_KEY_R, 0x1E5C, 
GDK_KEY_macron, GDK_KEY_dead_belowdot, GDK_KEY_l, 0x1E39, 
GDK_KEY_macron, GDK_KEY_dead_belowdot, GDK_KEY_r, 0x1E5D, 
GDK_KEY_acute, GDK_KEY_quotedbl, GDK_KEY_I, 0x1E2E, 
GDK_KEY_acute, GDK_KEY_quotedbl, GDK_KEY_U, 0x01D7, 
GDK_KEY_acute, GDK_KEY_quotedbl, GDK_KEY_i, 0x1E2F, 
GDK_KEY_acute, GDK_KEY_quotedbl, GDK_KEY_u, 0x01D8, 
GDK_KEY_acute, GDK_KEY_quotedbl, GDK_KEY_Greek_iota, 0x0390, 
GDK_KEY_acute, GDK_KEY_quotedbl, GDK_KEY_Greek_upsilon, 0x03B0, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F0D, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_EPSILON, 0x1F1D, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F2D, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_IOTA, 0x1F3D, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_OMICRON, 0x1F4D, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_UPSILON, 0x1F5D, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1F6D, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F05, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_epsilon, 0x1F15, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F25, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_iota, 0x1F35, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_omicron, 0x1F45, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_upsilon, 0x1F55, 
GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1F65, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F0C, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_EPSILON, 0x1F1C, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F2C, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_IOTA, 0x1F3C, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_OMICRON, 0x1F4C, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1F6C, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F04, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_epsilon, 0x1F14, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F24, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_iota, 0x1F34, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_omicron, 0x1F44, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_upsilon, 0x1F54, 
GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1F64, 
GDK_KEY_acute, GDK_KEY_plus, GDK_KEY_O, 0x1EDA, 
GDK_KEY_acute, GDK_KEY_plus, GDK_KEY_U, 0x1EE8, 
GDK_KEY_acute, GDK_KEY_plus, GDK_KEY_o, 0x1EDB, 
GDK_KEY_acute, GDK_KEY_plus, GDK_KEY_u, 0x1EE9, 
GDK_KEY_acute, GDK_KEY_comma, GDK_KEY_C, 0x1E08, 
GDK_KEY_acute, GDK_KEY_comma, GDK_KEY_c, 0x1E09, 
GDK_KEY_acute, GDK_KEY_slash, GDK_KEY_O, 0x01FE, 
GDK_KEY_acute, GDK_KEY_slash, GDK_KEY_o, 0x01FF, 
GDK_KEY_acute, GDK_KEY_asciicircum, GDK_KEY_A, 0x1EA4, 
GDK_KEY_acute, GDK_KEY_asciicircum, GDK_KEY_E, 0x1EBE, 
GDK_KEY_acute, GDK_KEY_asciicircum, GDK_KEY_O, 0x1ED0, 
GDK_KEY_acute, GDK_KEY_asciicircum, GDK_KEY_a, 0x1EA5, 
GDK_KEY_acute, GDK_KEY_asciicircum, GDK_KEY_e, 0x1EBF, 
GDK_KEY_acute, GDK_KEY_asciicircum, GDK_KEY_o, 0x1ED1, 
GDK_KEY_acute, GDK_KEY_underscore, GDK_KEY_E, 0x1E16, 
GDK_KEY_acute, GDK_KEY_underscore, GDK_KEY_O, 0x1E52, 
GDK_KEY_acute, GDK_KEY_underscore, GDK_KEY_e, 0x1E17, 
GDK_KEY_acute, GDK_KEY_underscore, GDK_KEY_o, 0x1E53, 
GDK_KEY_acute, GDK_KEY_b, GDK_KEY_A, 0x1EAE, 
GDK_KEY_acute, GDK_KEY_b, GDK_KEY_a, 0x1EAF, 
GDK_KEY_acute, GDK_KEY_asciitilde, GDK_KEY_O, 0x1E4C, 
GDK_KEY_acute, GDK_KEY_asciitilde, GDK_KEY_U, 0x1E78, 
GDK_KEY_acute, GDK_KEY_asciitilde, GDK_KEY_o, 0x1E4D, 
GDK_KEY_acute, GDK_KEY_asciitilde, GDK_KEY_u, 0x1E79, 
GDK_KEY_acute, GDK_KEY_macron, GDK_KEY_E, 0x1E16, 
GDK_KEY_acute, GDK_KEY_macron, GDK_KEY_O, 0x1E52, 
GDK_KEY_acute, GDK_KEY_macron, GDK_KEY_e, 0x1E17, 
GDK_KEY_acute, GDK_KEY_macron, GDK_KEY_o, 0x1E53, 
GDK_KEY_acute, GDK_KEY_cedilla, GDK_KEY_C, 0x1E08, 
GDK_KEY_acute, GDK_KEY_cedilla, GDK_KEY_c, 0x1E09, 
GDK_KEY_acute, GDK_KEY_dead_circumflex, GDK_KEY_A, 0x1EA4, 
GDK_KEY_acute, GDK_KEY_dead_circumflex, GDK_KEY_E, 0x1EBE, 
GDK_KEY_acute, GDK_KEY_dead_circumflex, GDK_KEY_O, 0x1ED0, 
GDK_KEY_acute, GDK_KEY_dead_circumflex, GDK_KEY_a, 0x1EA5, 
GDK_KEY_acute, GDK_KEY_dead_circumflex, GDK_KEY_e, 0x1EBF, 
GDK_KEY_acute, GDK_KEY_dead_circumflex, GDK_KEY_o, 0x1ED1, 
GDK_KEY_acute, GDK_KEY_dead_tilde, GDK_KEY_O, 0x1E4C, 
GDK_KEY_acute, GDK_KEY_dead_tilde, GDK_KEY_U, 0x1E78, 
GDK_KEY_acute, GDK_KEY_dead_tilde, GDK_KEY_o, 0x1E4D, 
GDK_KEY_acute, GDK_KEY_dead_tilde, GDK_KEY_u, 0x1E79, 
GDK_KEY_acute, GDK_KEY_dead_macron, GDK_KEY_E, 0x1E16, 
GDK_KEY_acute, GDK_KEY_dead_macron, GDK_KEY_O, 0x1E52, 
GDK_KEY_acute, GDK_KEY_dead_macron, GDK_KEY_e, 0x1E17, 
GDK_KEY_acute, GDK_KEY_dead_macron, GDK_KEY_o, 0x1E53, 
GDK_KEY_acute, GDK_KEY_dead_breve, GDK_KEY_A, 0x1EAE, 
GDK_KEY_acute, GDK_KEY_dead_breve, GDK_KEY_a, 0x1EAF, 
GDK_KEY_acute, GDK_KEY_dead_diaeresis, GDK_KEY_I, 0x1E2E, 
GDK_KEY_acute, GDK_KEY_dead_diaeresis, GDK_KEY_U, 0x01D7, 
GDK_KEY_acute, GDK_KEY_dead_diaeresis, GDK_KEY_i, 0x1E2F, 
GDK_KEY_acute, GDK_KEY_dead_diaeresis, GDK_KEY_u, 0x01D8, 
GDK_KEY_acute, GDK_KEY_dead_diaeresis, GDK_KEY_Greek_iota, 0x0390, 
GDK_KEY_acute, GDK_KEY_dead_diaeresis, GDK_KEY_Greek_upsilon, 0x03B0, 
GDK_KEY_acute, GDK_KEY_dead_abovering, GDK_KEY_A, 0x01FA, 
GDK_KEY_acute, GDK_KEY_dead_abovering, GDK_KEY_a, 0x01FB, 
GDK_KEY_acute, GDK_KEY_dead_cedilla, GDK_KEY_C, 0x1E08, 
GDK_KEY_acute, GDK_KEY_dead_cedilla, GDK_KEY_c, 0x1E09, 
GDK_KEY_acute, GDK_KEY_dead_horn, GDK_KEY_O, 0x1EDA, 
GDK_KEY_acute, GDK_KEY_dead_horn, GDK_KEY_U, 0x1EE8, 
GDK_KEY_acute, GDK_KEY_dead_horn, GDK_KEY_o, 0x1EDB, 
GDK_KEY_acute, GDK_KEY_dead_horn, GDK_KEY_u, 0x1EE9, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F0C, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_EPSILON, 0x1F1C, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F2C, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_IOTA, 0x1F3C, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_OMICRON, 0x1F4C, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1F6C, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F04, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_epsilon, 0x1F14, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F24, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_iota, 0x1F34, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_omicron, 0x1F44, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_upsilon, 0x1F54, 
GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1F64, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F0D, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_EPSILON, 0x1F1D, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F2D, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_IOTA, 0x1F3D, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMICRON, 0x1F4D, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_UPSILON, 0x1F5D, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1F6D, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F05, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_epsilon, 0x1F15, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F25, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_iota, 0x1F35, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_omicron, 0x1F45, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_upsilon, 0x1F55, 
GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1F65, 
GDK_KEY_acute, GDK_KEY_KP_Divide, GDK_KEY_O, 0x01FE, 
GDK_KEY_acute, GDK_KEY_KP_Divide, GDK_KEY_o, 0x01FF, 
0x05C1, 0x05BC, GDK_KEY_hebrew_shin, 0xFB2C, 
0x05C2, 0x05BC, GDK_KEY_hebrew_shin, 0xFB2D, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_Greek_alpha, 0x1FB4, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_Greek_eta, 0x1FC4, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_Greek_omega, 0x1FF4, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F00, 0x1F84, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F01, 0x1F85, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F08, 0x1F8C, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F09, 0x1F8D, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F20, 0x1F94, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F21, 0x1F95, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F28, 0x1F9C, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F29, 0x1F9D, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F60, 0x1FA4, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F61, 0x1FA5, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F68, 0x1FAC, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, 0x1F69, 0x1FAD, 
GDK_KEY_Greek_iota, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F89, 
GDK_KEY_Greek_iota, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F99, 
GDK_KEY_Greek_iota, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FA9, 
GDK_KEY_Greek_iota, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F81, 
GDK_KEY_Greek_iota, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F91, 
GDK_KEY_Greek_iota, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA1, 
GDK_KEY_Greek_iota, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F88, 
GDK_KEY_Greek_iota, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F98, 
GDK_KEY_Greek_iota, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FA8, 
GDK_KEY_Greek_iota, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F80, 
GDK_KEY_Greek_iota, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F90, 
GDK_KEY_Greek_iota, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA0, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_Greek_alpha, 0x1FB2, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_Greek_eta, 0x1FC2, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_Greek_omega, 0x1FF2, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F00, 0x1F82, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F01, 0x1F83, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F08, 0x1F8A, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F09, 0x1F8B, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F20, 0x1F92, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F21, 0x1F93, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F28, 0x1F9A, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F29, 0x1F9B, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F60, 0x1FA2, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F61, 0x1FA3, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F68, 0x1FAA, 
GDK_KEY_Greek_iota, GDK_KEY_grave, 0x1F69, 0x1FAB, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_Greek_alpha, 0x1FB7, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_Greek_eta, 0x1FC7, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_Greek_omega, 0x1FF7, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F00, 0x1F86, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F01, 0x1F87, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F08, 0x1F8E, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F09, 0x1F8F, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F20, 0x1F96, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F21, 0x1F97, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F28, 0x1F9E, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F29, 0x1F9F, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F60, 0x1FA6, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F61, 0x1FA7, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F68, 0x1FAE, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, 0x1F69, 0x1FAF, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_Greek_alpha, 0x1FB4, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_Greek_eta, 0x1FC4, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_Greek_omega, 0x1FF4, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F00, 0x1F84, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F01, 0x1F85, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F08, 0x1F8C, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F09, 0x1F8D, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F20, 0x1F94, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F21, 0x1F95, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F28, 0x1F9C, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F29, 0x1F9D, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F60, 0x1FA4, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F61, 0x1FA5, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F68, 0x1FAC, 
GDK_KEY_Greek_iota, GDK_KEY_acute, 0x1F69, 0x1FAD, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_Greek_alpha, 0x1FB2, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_Greek_eta, 0x1FC2, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_Greek_omega, 0x1FF2, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F00, 0x1F82, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F01, 0x1F83, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F08, 0x1F8A, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F09, 0x1F8B, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F20, 0x1F92, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F21, 0x1F93, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F28, 0x1F9A, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F29, 0x1F9B, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F60, 0x1FA2, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F61, 0x1FA3, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F68, 0x1FAA, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, 0x1F69, 0x1FAB, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_Greek_alpha, 0x1FB4, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_Greek_eta, 0x1FC4, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_Greek_omega, 0x1FF4, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F00, 0x1F84, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F01, 0x1F85, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F08, 0x1F8C, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F09, 0x1F8D, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F20, 0x1F94, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F21, 0x1F95, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F28, 0x1F9C, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F29, 0x1F9D, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F60, 0x1FA4, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F61, 0x1FA5, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F68, 0x1FAC, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, 0x1F69, 0x1FAD, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_Greek_alpha, 0x1FB7, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_Greek_eta, 0x1FC7, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_Greek_omega, 0x1FF7, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F00, 0x1F86, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F01, 0x1F87, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F08, 0x1F8E, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F09, 0x1F8F, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F20, 0x1F96, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F21, 0x1F97, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F28, 0x1F9E, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F29, 0x1F9F, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F60, 0x1FA6, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F61, 0x1FA7, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F68, 0x1FAE, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, 0x1F69, 0x1FAF, 
GDK_KEY_Greek_iota, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F88, 
GDK_KEY_Greek_iota, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F98, 
GDK_KEY_Greek_iota, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FA8, 
GDK_KEY_Greek_iota, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F80, 
GDK_KEY_Greek_iota, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F90, 
GDK_KEY_Greek_iota, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA0, 
GDK_KEY_Greek_iota, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F89, 
GDK_KEY_Greek_iota, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F99, 
GDK_KEY_Greek_iota, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FA9, 
GDK_KEY_Greek_iota, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F81, 
GDK_KEY_Greek_iota, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F91, 
GDK_KEY_Greek_iota, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA1, 
GDK_KEY_dead_abovedot, GDK_KEY_f, GDK_KEY_s, 0x1E9B, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_0, GDK_KEY_parenright, 0x2469, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_1, GDK_KEY_parenright, 0x246A, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_2, GDK_KEY_parenright, 0x246B, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_3, GDK_KEY_parenright, 0x246C, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_4, GDK_KEY_parenright, 0x246D, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_5, GDK_KEY_parenright, 0x246E, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_6, GDK_KEY_parenright, 0x246F, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_7, GDK_KEY_parenright, 0x2470, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_8, GDK_KEY_parenright, 0x2471, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_9, GDK_KEY_parenright, 0x2472, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_Space, GDK_KEY_parenright, 0x246B, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_0, GDK_KEY_parenright, 0x2469, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_1, GDK_KEY_parenright, 0x246A, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_2, GDK_KEY_parenright, 0x246B, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_3, GDK_KEY_parenright, 0x246C, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_4, GDK_KEY_parenright, 0x246D, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_5, GDK_KEY_parenright, 0x246E, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_6, GDK_KEY_parenright, 0x246F, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_7, GDK_KEY_parenright, 0x2470, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_8, GDK_KEY_parenright, 0x2471, 
GDK_KEY_parenleft, GDK_KEY_1, GDK_KEY_KP_9, GDK_KEY_parenright, 0x2472, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_0, GDK_KEY_parenright, 0x2473, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_1, GDK_KEY_parenright, 0x3251, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_2, GDK_KEY_parenright, 0x3252, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_3, GDK_KEY_parenright, 0x3253, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_4, GDK_KEY_parenright, 0x3254, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_5, GDK_KEY_parenright, 0x3255, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_6, GDK_KEY_parenright, 0x3256, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_7, GDK_KEY_parenright, 0x3257, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_8, GDK_KEY_parenright, 0x3258, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_9, GDK_KEY_parenright, 0x3259, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_Space, GDK_KEY_parenright, 0x3252, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_0, GDK_KEY_parenright, 0x2473, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_1, GDK_KEY_parenright, 0x3251, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_2, GDK_KEY_parenright, 0x3252, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_3, GDK_KEY_parenright, 0x3253, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_4, GDK_KEY_parenright, 0x3254, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_5, GDK_KEY_parenright, 0x3255, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_6, GDK_KEY_parenright, 0x3256, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_7, GDK_KEY_parenright, 0x3257, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_8, GDK_KEY_parenright, 0x3258, 
GDK_KEY_parenleft, GDK_KEY_2, GDK_KEY_KP_9, GDK_KEY_parenright, 0x3259, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_0, GDK_KEY_parenright, 0x325A, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_1, GDK_KEY_parenright, 0x325B, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_2, GDK_KEY_parenright, 0x325C, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_3, GDK_KEY_parenright, 0x325D, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_4, GDK_KEY_parenright, 0x325E, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_5, GDK_KEY_parenright, 0x325F, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_6, GDK_KEY_parenright, 0x32B1, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_7, GDK_KEY_parenright, 0x32B2, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_8, GDK_KEY_parenright, 0x32B3, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_9, GDK_KEY_parenright, 0x32B4, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_Space, GDK_KEY_parenright, 0x325C, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_0, GDK_KEY_parenright, 0x325A, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_1, GDK_KEY_parenright, 0x325B, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_2, GDK_KEY_parenright, 0x325C, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_3, GDK_KEY_parenright, 0x325D, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_4, GDK_KEY_parenright, 0x325E, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_5, GDK_KEY_parenright, 0x325F, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_6, GDK_KEY_parenright, 0x32B1, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_7, GDK_KEY_parenright, 0x32B2, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_8, GDK_KEY_parenright, 0x32B3, 
GDK_KEY_parenleft, GDK_KEY_3, GDK_KEY_KP_9, GDK_KEY_parenright, 0x32B4, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_0, GDK_KEY_parenright, 0x32B5, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_1, GDK_KEY_parenright, 0x32B6, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_2, GDK_KEY_parenright, 0x32B7, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_3, GDK_KEY_parenright, 0x32B8, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_4, GDK_KEY_parenright, 0x32B9, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_5, GDK_KEY_parenright, 0x32BA, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_6, GDK_KEY_parenright, 0x32BB, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_7, GDK_KEY_parenright, 0x32BC, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_8, GDK_KEY_parenright, 0x32BD, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_9, GDK_KEY_parenright, 0x32BE, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_Space, GDK_KEY_parenright, 0x32B7, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_0, GDK_KEY_parenright, 0x32B5, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_1, GDK_KEY_parenright, 0x32B6, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_2, GDK_KEY_parenright, 0x32B7, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_3, GDK_KEY_parenright, 0x32B8, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_4, GDK_KEY_parenright, 0x32B9, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_5, GDK_KEY_parenright, 0x32BA, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_6, GDK_KEY_parenright, 0x32BB, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_7, GDK_KEY_parenright, 0x32BC, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_8, GDK_KEY_parenright, 0x32BD, 
GDK_KEY_parenleft, GDK_KEY_4, GDK_KEY_KP_9, GDK_KEY_parenright, 0x32BE, 
GDK_KEY_parenleft, GDK_KEY_5, GDK_KEY_KP_0, GDK_KEY_parenright, 0x32BF, 
GDK_KEY_parenleft, 0x1100, 0x1161, GDK_KEY_parenright, 0x326E, 
GDK_KEY_parenleft, 0x1102, 0x1161, GDK_KEY_parenright, 0x326F, 
GDK_KEY_parenleft, 0x1103, 0x1161, GDK_KEY_parenright, 0x3270, 
GDK_KEY_parenleft, 0x1105, 0x1161, GDK_KEY_parenright, 0x3271, 
GDK_KEY_parenleft, 0x1106, 0x1161, GDK_KEY_parenright, 0x3272, 
GDK_KEY_parenleft, 0x1107, 0x1161, GDK_KEY_parenright, 0x3273, 
GDK_KEY_parenleft, 0x1109, 0x1161, GDK_KEY_parenright, 0x3274, 
GDK_KEY_parenleft, 0x110B, 0x1161, GDK_KEY_parenright, 0x3275, 
GDK_KEY_parenleft, 0x110C, 0x1161, GDK_KEY_parenright, 0x3276, 
GDK_KEY_parenleft, 0x110E, 0x1161, GDK_KEY_parenright, 0x3277, 
GDK_KEY_parenleft, 0x110F, 0x1161, GDK_KEY_parenright, 0x3278, 
GDK_KEY_parenleft, 0x1110, 0x1161, GDK_KEY_parenright, 0x3279, 
GDK_KEY_parenleft, 0x1111, 0x1161, GDK_KEY_parenright, 0x327A, 
GDK_KEY_parenleft, 0x1112, 0x1161, GDK_KEY_parenright, 0x327B, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_0, GDK_KEY_parenright, 0x2473, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_1, GDK_KEY_parenright, 0x3251, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_2, GDK_KEY_parenright, 0x3252, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_3, GDK_KEY_parenright, 0x3253, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_4, GDK_KEY_parenright, 0x3254, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_5, GDK_KEY_parenright, 0x3255, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_6, GDK_KEY_parenright, 0x3256, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_7, GDK_KEY_parenright, 0x3257, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_8, GDK_KEY_parenright, 0x3258, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_9, GDK_KEY_parenright, 0x3259, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_Space, GDK_KEY_parenright, 0x3252, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_0, GDK_KEY_parenright, 0x2473, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_1, GDK_KEY_parenright, 0x3251, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_2, GDK_KEY_parenright, 0x3252, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_3, GDK_KEY_parenright, 0x3253, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_4, GDK_KEY_parenright, 0x3254, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_5, GDK_KEY_parenright, 0x3255, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_6, GDK_KEY_parenright, 0x3256, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_7, GDK_KEY_parenright, 0x3257, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_8, GDK_KEY_parenright, 0x3258, 
GDK_KEY_parenleft, GDK_KEY_KP_Space, GDK_KEY_KP_9, GDK_KEY_parenright, 0x3259, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_0, GDK_KEY_parenright, 0x2469, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_1, GDK_KEY_parenright, 0x246A, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_2, GDK_KEY_parenright, 0x246B, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_3, GDK_KEY_parenright, 0x246C, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_4, GDK_KEY_parenright, 0x246D, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_5, GDK_KEY_parenright, 0x246E, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_6, GDK_KEY_parenright, 0x246F, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_7, GDK_KEY_parenright, 0x2470, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_8, GDK_KEY_parenright, 0x2471, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_9, GDK_KEY_parenright, 0x2472, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_Space, GDK_KEY_parenright, 0x246B, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_0, GDK_KEY_parenright, 0x2469, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_1, GDK_KEY_parenright, 0x246A, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_2, GDK_KEY_parenright, 0x246B, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_3, GDK_KEY_parenright, 0x246C, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_4, GDK_KEY_parenright, 0x246D, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_5, GDK_KEY_parenright, 0x246E, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_6, GDK_KEY_parenright, 0x246F, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_7, GDK_KEY_parenright, 0x2470, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_8, GDK_KEY_parenright, 0x2471, 
GDK_KEY_parenleft, GDK_KEY_KP_1, GDK_KEY_KP_9, GDK_KEY_parenright, 0x2472, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_0, GDK_KEY_parenright, 0x2473, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_1, GDK_KEY_parenright, 0x3251, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_2, GDK_KEY_parenright, 0x3252, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_3, GDK_KEY_parenright, 0x3253, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_4, GDK_KEY_parenright, 0x3254, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_5, GDK_KEY_parenright, 0x3255, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_6, GDK_KEY_parenright, 0x3256, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_7, GDK_KEY_parenright, 0x3257, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_8, GDK_KEY_parenright, 0x3258, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_9, GDK_KEY_parenright, 0x3259, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_Space, GDK_KEY_parenright, 0x3252, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_0, GDK_KEY_parenright, 0x2473, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_1, GDK_KEY_parenright, 0x3251, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_2, GDK_KEY_parenright, 0x3252, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_3, GDK_KEY_parenright, 0x3253, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_4, GDK_KEY_parenright, 0x3254, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_5, GDK_KEY_parenright, 0x3255, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_6, GDK_KEY_parenright, 0x3256, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_7, GDK_KEY_parenright, 0x3257, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_8, GDK_KEY_parenright, 0x3258, 
GDK_KEY_parenleft, GDK_KEY_KP_2, GDK_KEY_KP_9, GDK_KEY_parenright, 0x3259, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_0, GDK_KEY_parenright, 0x325A, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_1, GDK_KEY_parenright, 0x325B, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_2, GDK_KEY_parenright, 0x325C, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_3, GDK_KEY_parenright, 0x325D, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_4, GDK_KEY_parenright, 0x325E, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_5, GDK_KEY_parenright, 0x325F, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_6, GDK_KEY_parenright, 0x32B1, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_7, GDK_KEY_parenright, 0x32B2, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_8, GDK_KEY_parenright, 0x32B3, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_9, GDK_KEY_parenright, 0x32B4, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_Space, GDK_KEY_parenright, 0x325C, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_0, GDK_KEY_parenright, 0x325A, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_1, GDK_KEY_parenright, 0x325B, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_2, GDK_KEY_parenright, 0x325C, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_3, GDK_KEY_parenright, 0x325D, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_4, GDK_KEY_parenright, 0x325E, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_5, GDK_KEY_parenright, 0x325F, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_6, GDK_KEY_parenright, 0x32B1, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_7, GDK_KEY_parenright, 0x32B2, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_8, GDK_KEY_parenright, 0x32B3, 
GDK_KEY_parenleft, GDK_KEY_KP_3, GDK_KEY_KP_9, GDK_KEY_parenright, 0x32B4, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_0, GDK_KEY_parenright, 0x32B5, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_1, GDK_KEY_parenright, 0x32B6, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_2, GDK_KEY_parenright, 0x32B7, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_3, GDK_KEY_parenright, 0x32B8, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_4, GDK_KEY_parenright, 0x32B9, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_5, GDK_KEY_parenright, 0x32BA, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_6, GDK_KEY_parenright, 0x32BB, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_7, GDK_KEY_parenright, 0x32BC, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_8, GDK_KEY_parenright, 0x32BD, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_9, GDK_KEY_parenright, 0x32BE, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_Space, GDK_KEY_parenright, 0x32B7, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_0, GDK_KEY_parenright, 0x32B5, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_1, GDK_KEY_parenright, 0x32B6, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_2, GDK_KEY_parenright, 0x32B7, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_3, GDK_KEY_parenright, 0x32B8, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_4, GDK_KEY_parenright, 0x32B9, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_5, GDK_KEY_parenright, 0x32BA, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_6, GDK_KEY_parenright, 0x32BB, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_7, GDK_KEY_parenright, 0x32BC, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_8, GDK_KEY_parenright, 0x32BD, 
GDK_KEY_parenleft, GDK_KEY_KP_4, GDK_KEY_KP_9, GDK_KEY_parenright, 0x32BE, 
GDK_KEY_parenleft, GDK_KEY_KP_5, GDK_KEY_KP_0, GDK_KEY_parenright, 0x32BF, 
GDK_KEY_C, GDK_KEY_C, GDK_KEY_C, GDK_KEY_P, 0x262D, 
GDK_KEY_U, GDK_KEY_space, GDK_KEY_comma, GDK_KEY_E, 0x1E1C, 
GDK_KEY_U, GDK_KEY_space, GDK_KEY_comma, GDK_KEY_e, 0x1E1D, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_Greek_iota, GDK_KEY_apostrophe, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8B, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9B, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAB, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F83, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F93, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA3, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8A, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9A, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAA, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F82, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F92, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA2, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8A, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9A, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAA, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F82, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F92, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA2, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8B, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9B, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAB, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F83, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F93, 
GDK_KEY_Greek_iota, GDK_KEY_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA3, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8F, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9F, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAF, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F87, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F97, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA7, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8E, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9E, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAE, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F86, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F96, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA6, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8E, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9E, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAE, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F86, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F96, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA6, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8F, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9F, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAF, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F87, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F97, 
GDK_KEY_Greek_iota, GDK_KEY_asciitilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA7, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_Greek_iota, GDK_KEY_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8B, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9B, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAB, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F83, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F93, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA3, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8A, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9A, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAA, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F82, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F92, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA2, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8A, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9A, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAA, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F82, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F92, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA2, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8B, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9B, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAB, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F83, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F93, 
GDK_KEY_Greek_iota, GDK_KEY_dead_grave, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA3, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8C, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9C, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAC, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F84, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F94, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA4, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8D, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9D, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAD, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F85, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F95, 
GDK_KEY_Greek_iota, GDK_KEY_dead_acute, GDK_KEY_dead_dasia, GDK_KEY_Greek_omega, 0x1FA5, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenleft, GDK_KEY_Greek_ALPHA, 0x1F8F, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenleft, GDK_KEY_Greek_ETA, 0x1F9F, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenleft, GDK_KEY_Greek_OMEGA, 0x1FAF, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenleft, GDK_KEY_Greek_alpha, 0x1F87, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenleft, GDK_KEY_Greek_eta, 0x1F97, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenleft, GDK_KEY_Greek_omega, 0x1FA7, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenright, GDK_KEY_Greek_ALPHA, 0x1F8E, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenright, GDK_KEY_Greek_ETA, 0x1F9E, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenright, GDK_KEY_Greek_OMEGA, 0x1FAE, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenright, GDK_KEY_Greek_alpha, 0x1F86, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenright, GDK_KEY_Greek_eta, 0x1F96, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_parenright, GDK_KEY_Greek_omega, 0x1FA6, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_ALPHA, 0x1F8E, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_ETA, 0x1F9E, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_OMEGA, 0x1FAE, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_alpha, 0x1F86, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_eta, 0x1F96, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_psili, GDK_KEY_Greek_omega, 0x1FA6, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_ALPHA, 0x1F8F, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_ETA, 0x1F9F, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_OMEGA, 0x1FAF, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_alpha, 0x1F87, 
GDK_KEY_Greek_iota, GDK_KEY_dead_tilde, GDK_KEY_dead_dasia, GDK_KEY_Greek_eta, 0x1F97, 
};

#endif /* __GTK_IM_CONTEXT_SIMPLE_SEQS_H__ */

